/*****************************************************************************
 * File: MultiSignalReader.cpp
 * Author: Jonathan Gips, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2003 MIT Media Lab
 *
 * Adapted from OpineViewerEngine.cpp
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/timeb.h>
#include <signal.h>

const int TRUE = 1;
const int FALSE = 0;

#include "MultiSignalReader.hpp"

typedef void *(*ThreadProcType)(void *);

/* GLOBAL DECLARATIONS */
MultiSignalReader *reader = NULL;

MultiSignalReader::MultiSignalReader(const char *sUNL, FILE *oArg, bool tArg) 
  : subUNL(strdup(sUNL)),
    outFD(oArg), terse(tArg)
{
  recs = ArrayNew(sizeof(MultiSignalRec), 4, NULL);

  pthread_mutex_init(&printLock, NULL);
}

MultiSignalReader::~MultiSignalReader()
{
  int i;
  MultiSignalRec *rec = NULL;

  for(i = 0; i < ArrayLength(recs); i++)
    {
      rec = *((MultiSignalRec**) ArrayNth(recs, i));
      
      pthread_mutex_lock(&rec->recLock);

      if(rec->sigName) free(rec->sigName);
      if(rec->sig) SignalClose(rec->sig);

      pthread_mutex_unlock(&rec->recLock);
      free(rec);
    }

  if(recs) ArrayFree(recs);
  if(subUNL) free(subUNL);
}

/*
 * When an update of the subscribed UNL is received, this function is called
 * to reconcile the applications's state with the state of the tree to which
 * it is subscribed.  This means that some new signals must be read and some
 * old ones aborted.
 */

void MultiSignalReader::handleUpdate(NodeType *updateNode, const char *unl, 
				     const char *publisher) 
{
  for( int i = 0; i < ArrayLength(recs); i++)
    {
      MultiSignalRec *rec = *((MultiSignalRec**) ArrayNth(recs, i));
      rec->foundInUpdate = false;
    }

  if(updateNode) TraverseTree(updateNode, NULL, this, NULL, ProcessUpdateNode, subUNL);

  for( int i = 0; i < ArrayLength(recs); i++)
    {
        MultiSignalRec *rec = *((MultiSignalRec**) ArrayNth(recs, i));
      if(!rec->foundInUpdate) rec->destroy = true;
    }
  
  if(updateNode && updateNode->type != T_SIGNAL) FreeNode(updateNode);
}

/*
 * This function is passed to TraverseTree and called on every branch node.
 * It will remove the stored state for dead signals and also create threads
 * for new nodes.
 */

int MultiSignalReader::ProcessUpdateNode(void *data, NodeType *processNode, 
		      NodeType *parentNode, const char *UNL)
{
  int i;
  MultiSignalReader *reader = (MultiSignalReader *) data;
  MultiSignalRec *rec = NULL, *newRec = NULL;
  ThreadRec *thrRec = NULL;
  bool inserted(false);
  pthread_t thread;

  if(processNode->type != T_SIGNAL) return TRUE;
  
  inserted = false;
  //fprintf(stdout, "Processing Node with UNL: %s\n", UNL);

  // Linear search for the matching record.  This isn't too bad
  // because updates should be somewhat rare.  We could change this
  // to do some smarter search/sorting...
  for(i = 0; i < ArrayLength(reader->recs); i++)
    {
      rec = *((MultiSignalRec**) ArrayNth(reader->recs, i));
      
      // The signal threads null out the sigName field if the signal
      // is dead and needs to be removed.  These threads then
      // exit, so the record needs to be deleted.
      if(!rec->sigName)
	{
	  ArrayDeleteAt(reader->recs, i);
	  
	  i--;
	  continue;
	}
      
      // If the current node in the update matches the rec
      // then this node has "inserted" status, and the
      // loop can be exited.
      else if(!strcmp(rec->sigName, UNL))
	{
	  inserted = true;
	  rec->foundInUpdate = true;
	  break;
	}
    }
  
  // If the node in the update is not in the record array, 
  // then a new rec needs to be allocated, inserted, and a 
  // new thread needs to be inserted.
  if(!inserted)
    {
      // New sigName signal, so add a rec
      newRec = (MultiSignalRec*) malloc(sizeof(MultiSignalRec));
      thrRec = (ThreadRec*) malloc(sizeof(ThreadRec));
      
      ArrayAppend(reader->recs, &newRec);
      
      pthread_mutex_init(&newRec->recLock, NULL);
      newRec->sigName = strdup(UNL);
      
      // processNode is handled from here on by signal code
      if(parentNode) RemoveNode(parentNode, processNode);
      
      newRec->sig = NewSignalNode(processNode);
      newRec->destroy = false;
      newRec->foundInUpdate = true;

      thrRec->rec = newRec;
      thrRec->reader = reader;
      
      // printf("%s is about to get a new thread\n", newRec->sigName);
      pthread_create(&thread, NULL, receiveSignal, thrRec);
    }  

  return TRUE;
}

/*
 * Threads that handle new signal nodes are passed into this
 * thread for processing.
 */

void *MultiSignalReader::receiveSignal(void *data)
{
  ThreadRec *thrRec = (ThreadRec*) data;
  MultiSignalRec *rec = thrRec->rec;
  MultiSignalReader *reader = thrRec->reader;
  char featureBuff[kFeatureBufferLen];
  struct timeval sigTime, localTime;
  
  while(rec->sig && 
    SignalRecv(rec->sig, &sigTime, featureBuff))
    {
      if(rec->destroy) 
	{
	  SignalClose(rec->sig);
	  rec->sig = NULL;
	  break;
	}

      gettimeofday(&localTime,NULL);
      
      pthread_mutex_lock(&reader->printLock);
      
      if (reader->terse) {     
	fprintf(reader->outFD, "%s ", strrchr(rec->sigName,'/')+1);
	TerselyPrintSignalElements(reader->outFD, rec->sig, sigTime, featureBuff);
      } else {
	fprintf(reader->outFD, "%s ", rec->sigName);
	PrintSignalElements(reader->outFD, rec->sig, sigTime, featureBuff);
      }

      pthread_mutex_unlock(&reader->printLock);
    }
  
  if(rec->sig) {
    SignalClose(rec->sig);
    rec->sig = NULL;
  }
  free(rec->sigName);
  rec->sigName = NULL; // Main thread uses null rec->sigName to know it is safe to delete rec
  free(thrRec); // free the ThreadRec, but the MultiSignalRec is still in use
  
  //printf("exiting signal thread.\n");
  
  pthread_exit(0);

  return NULL;
}

int MultiSignalReader::lockFile()
{
  if(lockMethod) return 0;

  pthread_mutex_lock(&printLock);
  lockMethod = true;

  return 1;
}

int MultiSignalReader::unlockFile()
{
  if(!lockMethod) return 0;

  pthread_mutex_unlock(&printLock);
  lockMethod = false;
  
  return 1;
}


void SigHandler(int sig) {
  reader->lockFile();
  exit(0);
}

void PrintUsage(char *arg0)
{
  fprintf(stderr, "Usage: %s [OPTION]...\n", arg0);
  fprintf(stderr, 
	  "Connect to multiple signals at and below a specified server and path.  "
	  "Output is kept synchronized with the nodes on the whiteboard, if they "
	  "change.\n\n");
  fprintf(stderr, "  -s [server_addr]	server to read from, defaults to 'localhost'\n");
  fprintf(stderr, "  -p [path]	        path to read from, defaults to '/'\n");
  fprintf(stderr, "  -w [FILE]	        write to [FILE], defaults to stdout\n");
  fprintf(stderr, "  -a [FILE]	        append to [FILE]\n");
  fprintf(stderr, "  -t                 results in terse, space separated output\n");
  //fprintf(stderr, 
  //	  "  -d		when writing to a file, put all the info with each\n");
  //fprintf(stderr, 
  //	  "		datagram, so the file can be split easily\n");
  //fprintf(stderr, "  -c [AUTH]	hold data if there is no connected reader;\n");
  //fprintf(stderr, "           	when reading, send receipts after the data\n");
  //fprintf(stderr, "           	is committed.\n");
  //fprintf(stderr, "  -t [FILE]	use [FILE] for temporary storage if needed\n");
  //fprintf(stderr, "  -s [SVR]	make [SVR] the default enchantment server\n");
  //fprintf(stderr, "  -f       	fast; ignore original speed\n");
  //fprintf(stderr, "  -p         proxy through the server\n");
  //fprintf(stderr, "  -n       	shift the timestamps to start now.\n");
  fprintf(stderr, "\n");
    
}


int main(int argc, char *argv[]) 
{ 
  char hostName[128];
  gethostname(hostName, 128);
  char *server = NULL, *path = NULL;
  bool serverSet(false), pathSet(false), terse(false);
  bool writeSet(false), appendSet(false);
  char *writeFilename, *appendFilename;
  FILE *outFD = NULL;

  for(int i = 1; i < argc; i++)
    {
      if(argv[i][0] == '-')
	{
	  if((argv[i][1] == 's'))
	    {
	      i++;
	      server = argv[i];
	      serverSet = true;
	    }
	  else if((argv[i][1] == 'p'))
	    {
	      i++;
	      path = argv[i];
	      pathSet = true;
	    }
	  else if((argv[i][1] == 'w'))
	    {
	      i++;
	      writeFilename = argv[i];
	      fprintf(stderr, "Writing output to %s.\n", argv[i]);
	      writeSet = true;
	    }
	  else if((argv[i][1] == 'a'))
	    {
	      i++;
	      appendFilename = argv[i];
	      fprintf(stderr, "Appending output to %s.\n", argv[i]);
	      appendSet = true;
	    }
	  
	  else if((argv[i][1] == 't'))
            {
              terse = true;
              fprintf(stderr, "Terse mode.\n");
            }
	  else if((argv[i][1] == 'h'))
	    {
	      PrintUsage(*argv);	      
	      exit(1);
	    }
	} 
    }

  if(!pathSet)
    {
      path = "/";
    }
  
  if(!serverSet) 
    {
      server = "localhost";
    }

  if(writeSet)
    {
      outFD = fopen(writeFilename, "w");
    }
  else if(appendSet)
    {
      outFD = fopen(appendFilename, "a");
    }
  else
    {
      outFD = stdout;
    }

  const char *subUNL = CreateUNL(server, path);
  fprintf(stderr, "Reading from %s\n", subUNL);
  

  reader = new MultiSignalReader(subUNL, outFD, terse);
    
  SubscribeUNLInfo subInfo;
  subInfo.common.operationFlags = ENCHANT_SUBSCRIBE_INITIAL;
  
  SubscribeUNL(subUNL, &subInfo);

  //Die nicely
  signal(SIGINT,SigHandler);

  while(1) 
    {
      GetUpdateInfo getInfo;

      WaitForUpdate(&getInfo, NULL);
      
      reader->handleUpdate(getInfo.node, getInfo.UNL, getInfo.publisher);
    }
    
  return 0;
}



