/*****************************************************************************
*
* Author: Ron Caneel
* http://www.media.mit.edu/wearables/
* Copyright (C) 2004 MIT Media Lab
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************
* Modified by Mark Blum (mark@media.mit.edu) in July 2005
*
* This code computes the five voice features: energy, spectal entropy, formant
* frequency, maximal and number of autocorrelation peaks.
* From those features several statistics are calculated (not verified) as well
* as voiced/non-voiced and speaking/non-speaking states.
* In addition the means and variances of these features are computed over
* BIG_WINDOWSIZE samples. These are the features used for Speech and Location
* Classification.
* Anmol Madan will be further developping this code
*****************************************************************************/

// comments for improvment:
// entropy: 3.2 matlab vs 3.0 in C
// 1)funny behavior: linear relation between spec.entropy matlab and c
// 2)shifting hanning separatly from vals: to improve kiss_fft with low numbers
// 3)check kiss_fft with floats instead of short
// autocorrelation:
// 4) check shifting of values; they don't quite add up somewhere.
// 5) fundamental freq.: confidence & spectral entropy enough for autocor??
// 6) in not 5) optimization in fft due to symetrics

#define VERBOSE_DEBUG

extern "C" {
#include <enchantment/whiteboard.h>
#include <enchantment/unl.h>
#include <enchantment/signals.h>
#include <getopt.h>
#include <tools/kiss_fftr.h>
}

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <signal.h>
#include <math.h>
#include <assert.h>


#define pi 3.1415926536f

#define SAMPLE_RATE 8000
#define VITERBI_PATH_LENGTH_VOICED 30
#define VITERBI_PATH_LENGTH_SPEAKING 140
#define NORMALIZE_AUDIO 32768  // = 2^15
#define frameSizeBits 8
#define frameSize (1 << frameSizeBits)
#define stepSize 128

#define SAMPLES_PER_MINUTE 3480 // 58*60
#define NUM_MINUTES 1
#define BIG_WINDOWSIZE 300 // about 5 seconds (features at 62.5 Hz)

#define maxEntropy 6.0
#define x_log_x_scale (65535.0 / maxEntropy)

//These offset correction are added by Mark after comparing the outputs in Matlab
#define OFFSET_F0 20.48
#define SpecEnt_A 0.091167 
#define SpecEnt_B -0.83876
#define SpecEnt_C 1.93907

int usec_difference(struct timeval *t1, struct timeval *t2)
{
    return (t1->tv_sec - t2->tv_sec) * 1000000 + (t1->tv_usec - t2->tv_usec);
}


void fast_find_acorr_peaks(int *acorr_slide, int len, float *maxpeak, int *numpeaks);
void fast_find_acorr_peaks_float(float *acorr_slide, int len, float *maxpeak, int *numpeaks);

int viterbi_voiced(float obser_prob[2], float trans_probs[2][2]);
void viterbi_speaking(int states[], float init_state_distro[2],float obser_prob[][2], float trans_probs[][2],int start_position);

void SigHandler(int sig);



// Returns floor(log_2(x))
// msb_32(1) == 0
// msb_32(3) == msb_32(2) == 1
// msb_32(1 << n) == n
int msb_32(int x)
{
    assert(x > 0);

    int shift =0;
    int max_shift = 0;
    for (int i = 1; i <= (1 << 30) && i > 0; i = i << 1){
        if (x & i){
            max_shift = shift;
        }
        shift++;
    }
    return max_shift;
}

// Returns floor(log_2(x))
int msb_16(int x)
{
    assert(x > 0);
    if (x >= 65536)
        printf("oh no!\n");
    assert(x < 65536);

    if (x >= 256) {
        if (x >= 4096) {
            if (x >= 16384) {
                if (x >= 32768)
                    return 15;
                else
                    return 14;
            } else {
                if (x >= 8192)
                    return 13;
                else
                    return 12;
            }
        } else {
            if (x >= 1024) {
                if (x >= 2048)
                    return 11;
                else
                    return 10;
            } else {
                if (x >= 512)
                    return 9;
                else
                    return 8;
            }
        }
    } else {
        if (x >= 16) {
            if (x >= 64) {
                if (x >= 128)
                    return 7;
                else
                    return 6;
            } else {
                if (x >= 32)
                    return 5;
                else
                    return 4;
            }
        } else {
            if (x >= 4) {
                if (x >= 8)
                    return 3;
                else
                    return 2;
            } else {
                if (x >= 2)
                    return 1;
                else
                    return 0;
            }
        }
    }
}

inline float CalcMeanColumn(float inMat[][5], int windowSize, int columnIndex) {
    int i;
    float sum = 0.0;
    for (i=0; i < windowSize; i++){
        sum += inMat[i][columnIndex];
    }
    return (float)sum/((float)windowSize);
}

inline float CalcVarColumn(float inMat[][5], int windowSize, int columnIndex, float mean) {
    int i;
    float sum = 0.0;
    for (i=0; i < windowSize; i++){
        sum += inMat[i][columnIndex]*inMat[i][columnIndex];
    }
    return sum/((float)windowSize) - mean*mean;            // var(X) = E(X - u)^2 = E(X^2) - u^2
}

float init_state[2];
volatile int runflag = 1;
volatile int killflag = 0;


int main(int argc, char *argv[]) {


    int fileInput =0;
    FILE *fin;

// signal variables
    char *sigUNL;
    const char *outputUNL;
    char *outUNL;
    const char *chunkUNL;

    Signal *in;
    Signal *features_sig;
    Signal *bigFeatures_sig;
    Signal *speaking_sig;
    Signal *voiced_sig;

    short *buffer;
    short *input;

    int length;


    float *features;
    short *voicedvalues;
    short *speakingvalues;
    float output[100][5];
    float features_cache[30][5];
//  short voiced_cache[20];
//  float feat_cache[50][5];

    //by Mark 8/3/2005
    float bigBuffer[BIG_WINDOWSIZE][5];
    int bigCounter = 0;
    float averagedFeatures[10];
    float bigFeaturesNormalization[] = {8.540838e-001, 1.451333e+001, 4.480513e+000, 3.008648e-002, 4.305218e+002, 1.336457e-001, 1.157824e+002, 3.351075e-001, 6.731011e-003, 2.713941e+004};
    bool lightMode = 0;
    
// variables are used to synchronise different signals
    short synch_voiced_count = 0;
    short synch_speak_count = 0;
    short synch_voiced = 29 ;
    short synch_speak = 49;

    short send_counter = 0;


//hmm variables
    FILE *fp_hmm_params;

    float feature_prob_state[2];
    float total_log_coeffs[2];
    float trans_prob[2][2] =  {{0.97654,0.02346},{0.09106,0.90984}};
    float mus[2][3] = {{0.0827,14.979,4.5668},{0.5875,7.5494,4.1535}};
    float	sigmas[2][3] = {{0.14274,11.284,0.1311},{0.24858,3.6685,0.29464}};
    init_state[0] = 0.5;//{-0.693147180559945,-0.693147180559945};
    init_state[1] = 0.5;
    float trans_probs_speaking[2][2] = {{0.9999, 0.0001},{0.0001,0.9999}};
    float prob_voiced[2][2] =  {{0.99, 0.5},{0.01,0.5}};
    float prob_state_voiced[VITERBI_PATH_LENGTH_SPEAKING][2];

//  int viterbicounter_voiced = 0;
    int viterbicounter_speaking = 0;
    int state = 1;
    int states[VITERBI_PATH_LENGTH_SPEAKING];
    int state_cache[VITERBI_PATH_LENGTH_VOICED];
    int state_counter = 0;

// auto correlation variables
    int comp[128];
    int acorr[128];
    float maxpeak;
    int numpeaks;
    float total;
    int avg;

// general control variables
    int internal_loops;
    int index, i,loop,c;

    int sum;
    int shift;
    int max_vals, min_vals;
    int vals_shift,spec_shift;
    int max_inverse, max_inv_shift;
    int tmp_inv[frameSize];



// calculating variables
    unsigned short log_x_times_x[32768];
    int sqrt_shift = 7; // sqrt(32768) * sqrt_shift < 32768
    short sqrt_x[32768];


    struct timeval tpre, tall, tstart;
    struct timeval inTimestamp, outTimestamp;

// fft variables
    void *cfg, *cfg_inv, *cfg_auto, *cfg_big_fft;
    kiss_fft_scalar vals_spec[frameSize];
    kiss_fft_cpx vals_inv[2*frameSize];
    kiss_fft_scalar *vals;
    kiss_fft_scalar vals_shifted[2*frameSize];
    kiss_fft_cpx	outvals[frameSize];

    kiss_fft_scalar big_fft_in[frameSize*2];	// only use frameSize/2, but need to be same length as fft resolution
    kiss_fft_cpx big_fft[frameSize*2];

    short min_freq = 50, max_freq = 500;
    short sampling_rate = 7403;
    int max_index = (frameSize*2*sampling_rate)/(min_freq * frameSize);
    int min_index = (frameSize*2*sampling_rate)/(max_freq * frameSize);

// energy + spec variables
    int stdDevSet=0, permanentSet =0;
    float stdDev, stdDev_cache;
    int stdDev_int;

    int hanning[frameSize];
    int spec_sq[frameSize/2 +1];
    int spec[frameSize/2 +1];
    int normspec[frameSize];
    int sumspec;
    int energy;
    int energy_s;
    float cum_energy;
    unsigned short features_ent = 0;

    short voiced;
    short speak;

// statistical variables
    float formantSum, formantSqSum, specSum, specSqSum, energySum, energySqSum;
    int voicedSum, speakSum;
    int voicedSegCount, speakSegCount;
    bool lastVoiced, lastSpeak;

    float formantMean, specMean,energyMean = 0.0;
    float formantStd, specStd, energyStd =0.0;
    float fracSpeak, voicedRate = 0.0;

    short sample_counter =0;


    EnchantData data_formantMean, data_specMean, data_energyMean, data_formantStd, data_specStd, data_energyStd,data_fracSpeak, data_voicedRate;
    NodeType *node, *formantMeanNode, *specMeanNode,*energyMeanNode, *formantStdNode,*specStdNode,*energyStdNode,*fracSpeakNode, *voicedRateNode;


//int tot,cum_val,cum_init,cum_entA,cum_entB,cum_entC,cum_auto0,cum_autoA,cum_autoB1,cum_autoB2,cum_autoC,cum_peak,cum_hmm1,cum_hmm2,cum_hmm3,cum_sig;
//tot=cum_val=cum_init=cum_entA=cum_entB=cum_entC=cum_auto0=cum_autoA=cum_autoB1=cum_autoB2=cum_autoC=cum_peak=cum_hmm1=cum_hmm2=cum_hmm3=cum_sig=0;
//long iteration_count=0;

    stdDev = 0;
        
    int opt;  
    char **cmdargv = argv;
    int cmdargc = argc;

        
    do{
        opt = getopt(cmdargc,cmdargv,"ps:fhl");
        if (opt == 's'){
            stdDevSet = 1;
            stdDev = atof(optarg);
        }
        else if (opt == 'p'){
            permanentSet = 1;
        }
        else if (opt == 'f'){
            fileInput = 1;
        }
        else if (opt == 'l'){
            lightMode = 1;
        }
        else if (opt == 'h') {
            printf("\nUsage:\tAudioFeatures [-p] [-s std_dev] [-f] [-l] inputUNL outputUNL\n");
            printf("\twithout any additional options stdDev will be adjusted over time\n");
            printf("\twithout passing -s the first ~ 7 sec are used for callibration \n");
            printf("use [-s std_dev] to set initial value for energy threshold \n");
            printf("use [-p] to use calculated or set stdDev permanently (no rolling update)\n");
            printf("use [-f] to read input from file (enter filename instead of inputUNL)\n\n");
            printf("use [-l] for light mode, voiced and speaking calculations ommited\n\n");
            exit(0);
        }
    }	
    while (opt > 0);
        
    if (cmdargc - optind < 2){
        printf("\nUsage:\tAudioFeatures [-p] [-s std_dev] [-f] [-l] inputUNL outputUNL\n");
        printf("\twithout any additional options stdDev will be adjusted over time\n");
        printf("\twithout passing -s the first ~ 7 sec are used for callibration \n");
        printf("use [-s std_dev] to set initial value for energy threshold \n");
        printf("use [-p] to use calculated or set stdDev permanently (no rolling update)\n");
        printf("use [-f] to read input from file (enter filename instead of inputUNL)\n\n");
        printf("use [-l] for light mode, voiced and speaking calculations ommited\n\n");
        exit(0);
    }

        
    sigUNL = argv[optind++];
    outUNL = argv[optind++];

    printf("\ninput : %s\noutput: %s\n", sigUNL, outUNL);
    printf("\tstates : %s/states \n\tstatistics: %s/statistics\n\n", outUNL,outUNL);

/*  if (fileInput == 1 && stdDevSet != 1){
    printf("\n currently -f only works in combination with -s\n\n");
    exit(0);
}
*/
    fp_hmm_params = fopen("hmm_params_audio_features.cfg","r");
    if (fp_hmm_params != NULL){
        fscanf(fp_hmm_params,"trans_prob %f,%f,%f,%f\n",&trans_prob[0][0],&trans_prob[0][1],&trans_prob[1][0],&trans_prob[1][1]);
        fscanf(fp_hmm_params,"mus %f,%f,%f,%f,%f,%f\n",&mus[0][0],&mus[0][1],&mus[0][2],&mus[1][0],&mus[1][1],&mus[1][2]);
        fscanf(fp_hmm_params,"sigmas %f,%f,%f,%f,%f,%f\n",&sigmas[0][0],&sigmas[0][1],&sigmas[0][2],&sigmas[1][0],&sigmas[1][1],&sigmas[1][2]);
        fclose(fp_hmm_params);
    }
    else{
        printf("Could not find hmm_params.txt. Taking hardcoded values instead\n");
    }
        
        
        // Initialize Sighandlers
    signal(SIGINT,SigHandler);
    signal(SIGUSR1,SigHandler);



// init loop for stdDev
// guess stdDev grom first samples
    
    if (!stdDevSet){
        if (permanentSet  && fileInput){
            fin = fopen(sigUNL, "rb");
            int ch;
            float norm = 0;
            cum_energy = 0;
            while (fscanf(fin,"%d", &ch) != EOF){
                cum_energy += pow((float)ch/NORMALIZE_AUDIO,2);
                norm = norm +1;
            }
                
            stdDev = (cum_energy/norm)/25;
            fclose(fin);

            printf("stdDev for whole file is: %f\n",stdDev);
        }
        if (fileInput && !permanentSet){
            fin = fopen(sigUNL, "rb");
            int ch;
            float norm = 0;
            cum_energy = 0;
            for (c=1; c< 50 * 1024;c++){
                if (fscanf(fin,"%d", &ch) != EOF){
                    cum_energy += pow((float)ch/NORMALIZE_AUDIO,2);
                    norm = norm +1;
                }
            }
            stdDev = (cum_energy/norm)/25;
                
            fclose(fin);
        
        }
        else if(!fileInput) {
            in = NewSignalUNL(sigUNL);
            if (!in)
            {
                printf("Error allocating signal\n");
                exit(1);
            }

            SignalConnectWait(in);
            if (SignalGetType(in) != T_INT16)
            {
                printf("Input signal type error");
                exit(1);
            }

            buffer = (short *) malloc(SignalBufferSize(in));
            length = SignalGetUnitSamples(in);


            printf("start calibration\n");
            for(c=1; c< 50; c++){
                SignalRecv(in, &inTimestamp, (void *) buffer);
                cum_energy = 0;
                for (i= 0; i < length; i++) {
                    cum_energy += pow((float)buffer[i]/NORMALIZE_AUDIO,2);
                }
                stdDev += cum_energy/length;
            }
            stdDev = (stdDev/50) /25;
            free(buffer);
            SignalClose(in);
        }    
        printf("finished calibration stdDev = %.8f\n", stdDev);

    }


    stdDev_int = (int)(stdDev * (1 << 30));
    stdDev_cache = 0;

    features =(float *) malloc(sizeof(float) * 5);
    voicedvalues = (short *) malloc(sizeof(short) * 100);
    speakingvalues = (short *) malloc(sizeof(short) * 100);
//feat_cache =(float *) malloc(sizeof(float) * 5 * 50);
//voiced_cache = (short *) malloc(sizeof(short) * 20);
//output =(float *) malloc(sizeof(float) * 700);

// converting from probs to logs
    for (i = 0; i<2 ; i++){
        for (int j=0; j<2;j++){
            trans_prob[i][j] = log(trans_prob[i][j]);
            trans_probs_speaking[i][j]=log(trans_probs_speaking[i][j]);
            prob_voiced[i][j]= log(prob_voiced[i][j]);
        }
        init_state[i] = log(init_state[i]);
    }

/*  for (i = 0; i<100 ; i++){
    for (int j=0; j<7;j++){
    output[i][j] = 0.0;
}
}
*/

// init viterbi variables
    total_log_coeffs[0]=0;
    total_log_coeffs[1]=0;
    for (i = 0;i <3;i++){
        total_log_coeffs[0] -= log(sigmas[0][i]* sqrt(2*pi));
        total_log_coeffs[1] -= log(sigmas[1][i]* sqrt(2*pi));
    }
    feature_prob_state[0] = 0;
    feature_prob_state[1] = 0;
    for (i = 0;i < VITERBI_PATH_LENGTH_VOICED;i++){
        state_cache[i]= 0;
    }

// config fft.

    cfg = kiss_fftr_alloc( frameSize ,0,0,0 );
    cfg_auto = kiss_fftr_alloc( 2*frameSize ,0,0,0 );
    cfg_inv = kiss_fft_alloc( 2*frameSize ,1,0,0 );
    cfg_big_fft = kiss_fftr_alloc( 2*frameSize ,0,0,0 );

// init hanning window
    for (i = 0; i < frameSize; i++) {
        hanning[i]= (int)((0.5-0.5*cos(2.0*pi*((float)(i+1)/(frameSize+1)))) * 32767);
        vals_spec[i] = 0;
    }

// init compensate variable
    for (c= frameSize,i = 0;c > stepSize;c--, i++){
        comp[i] = (int)((1/((float)c/frameSize)) * 16384 );
        acorr[i] = 0;
    }

    /* precaculate log values since log() is done in float and to slow on arm*/
    log_x_times_x[0] = 0;
    for (i= 1; i < 32768 ; i++){
        log_x_times_x[i] = (short)(-x_log_x_scale * ((float)i/32768.0)*log((float)i/32768.0)+0.5);
    }

    sqrt_x[0] = 0;
    for (i= 1; i < 32768 ; i++){
        sqrt_x[i] = (int)(sqrt(i) * (1 << sqrt_shift));
    }

    chunkUNL = AppendUNLPath(outUNL,"statistics");
    outputUNL = AppendUNLPath(outUNL,"states/features");
    features_sig = NewSignalGenUNL(outputUNL);
    if (!features_sig)
    {
        printf("Error allocating signalGen\n");
        exit(1);
    }
    outputUNL = AppendUNLPath(outUNL,"states/bigFeatures");
    bigFeatures_sig = NewSignalGenUNL(outputUNL);
    if (!bigFeatures_sig)
    {
        printf("Error allocating signalGen\n");
        exit(1);
    }
    
//  SignalSetType(in, 2, SIGNAL_TYPE_INT);
    SignalSetType(features_sig, T_FLOAT);
    SignalSetType(bigFeatures_sig, T_FLOAT);
    SignalSetDimension(features_sig, 1, 5, "Audio Features",NULL);
    SignalSetDimension(bigFeatures_sig, 1, 10, "Averaged Audio Features",NULL);
    SignalSetSampleRate(features_sig, ((float)SAMPLE_RATE)/stepSize, 100);
    SignalSetSampleRate(bigFeatures_sig, ((float)BIG_WINDOWSIZE)/(((float)SAMPLE_RATE)/stepSize), 1);  //BIG_WINDOWSIZE/(8000/128)
   
    if (!lightMode) {
        //ExtractParentUNL(&voicedUNL,speakingUNL);
        outputUNL = AppendUNLPath(outUNL,"states/speaking");
        speaking_sig = NewSignalGenUNL(outputUNL);
        if (!speaking_sig)
        {
            printf("Error allocating signalGen\n");
            exit(1);
        }
        outputUNL = AppendUNLPath(outUNL,"states/voiced");
        voiced_sig = NewSignalGenUNL(outputUNL);
        if (!voiced_sig)
        {
            printf("Error allocating signalGen\n");
            exit(1);
        }
        SignalSetType(speaking_sig, T_INT16);
        SignalSetType(voiced_sig, T_INT16);
        SignalSetDimension(speaking_sig, 1, 1, "Speaking Features",NULL);
        SignalSetDimension(voiced_sig, 1, 1, "Voiced",NULL);
        SignalSetSampleRate(speaking_sig, ((float)SAMPLE_RATE)/stepSize ,100);
        SignalSetSampleRate(voiced_sig, ((float)SAMPLE_RATE)/stepSize , 100);
        // actual rate is 57.421875 (with smapled rate 1/3 of 22050)    // legacy??? corrected by mark to 8000/128 = 62.5
        
        data_formantMean.float_val = 0.0;
        data_specMean.float_val = 0.0;
        data_energyMean.float_val = 0.0;
        data_formantStd.float_val = 0.0;
        data_specStd.float_val = 0.0;
        data_energyStd.float_val = 0.0;
        data_fracSpeak.float_val = 0.0;
        data_voicedRate.float_val = 0.0;
        
        node = CreateBranchNode("ChunkFeatures",T_BRANCH);
        formantMeanNode = CreateTypedNode("formantMean",T_FLOAT,&data_formantMean);
        specMeanNode = CreateTypedNode("specMean",T_FLOAT,&data_specMean);
        energyMeanNode = CreateTypedNode("energyMean",T_FLOAT,&data_energyMean);
        formantStdNode = CreateTypedNode("formantStd",T_FLOAT,&data_formantStd);
        specStdNode = CreateTypedNode("specStd",T_FLOAT,&data_specStd);
        energyStdNode = CreateTypedNode("energyStd",T_FLOAT,&data_energyStd);
        fracSpeakNode = CreateTypedNode("fracSpeak",T_FLOAT,&data_fracSpeak);
        voicedRateNode = CreateTypedNode("voicedRate",T_FLOAT,&data_voicedRate);
        
        
        InsertNode(node, formantMeanNode);
        InsertNode(node, specMeanNode);
        InsertNode(node, energyMeanNode);
        InsertNode(node, formantStdNode);
        InsertNode(node, specStdNode);
        InsertNode(node, energyStdNode);
        InsertNode(node, fracSpeakNode);
        InsertNode(node, voicedRateNode);
        
        
        PublishNode(chunkUNL, node, NULL);
    }

    index = 0;


    for (i = 0; i < frameSize; i++){
        vals_spec[i] = 0;
    }

    printf("starting now ...\n");

    if (fileInput){
        length = 1024;
        internal_loops = length / stepSize;

        buffer = (short *)malloc(sizeof(short)*1024);
        input = (short *)malloc(sizeof(short) * (length + stepSize));
                
        fin = fopen(sigUNL, "rb");
        int ch;
        for (i= 0; i < 1024; i++) {
            fscanf(fin,"%d", &ch);
            buffer[i] =  ch;
        }


    }
    else{

        in = NewSignalUNL(sigUNL);

        if (!in)
        {
            printf("Error allocating signal\n");
            exit(1);
        }
        SignalConnectWait(in);

        length = SignalGetUnitSamples(in);
        internal_loops = length / stepSize;


        buffer = (short *) malloc(SignalBufferSize(in));
        input = (short *) malloc(sizeof(short) * (length + stepSize));



        for (i = 0; i < frameSize; i++){
            input[length - stepSize+ i] = 0;
        }

        SignalRecv(in, &inTimestamp, (void *) buffer);
    }

// 7 first frames are dropped input is filled up at
// this part will be copied to head of input in while loop
    for (i= length; i < length +stepSize; i++) {
        input[i] = (short)buffer[i-stepSize];
    }

    gettimeofday(&tall, NULL);
    gettimeofday(&tstart, NULL);

    formantSum = 0;
    formantSqSum = 0;
    specSum = 0;
    specSqSum = 0;
    energySum = 0;
    energySqSum = 0;
    voicedSum = 0;
    speakSum = 0;
    voicedSegCount = 0;
    speakSegCount = 0;
    lastVoiced = false;
    lastSpeak = false;


// main loop reads receives next singal chunk (1024)
    while (runflag)
    {

        if (fileInput){
            for (i= 0; i < 1024; i++) {
                int ch;
                if (fscanf(fin,"%d", &ch) != EOF)
                    buffer[i] =  ch;
                else{
                    runflag = 0;
                    continue;
                }
            }
        }
        else{
            SignalRecv(in, &inTimestamp, (void *) buffer);
        }

    // fills beginning of buffer with end of previous
        for (i= 0 ; i < stepSize; i++) {
            input[i] = input[length + i];
        }
        for (i= 0; i < length; i++) {
            input[i + stepSize] = (short)buffer[i];
        }

    // process the input with 256 frame size and 128 step size
        for (loop = 0; loop < internal_loops; loop++){
            gettimeofday(&tpre, NULL);

            sum =0;
            max_vals = 0;
            min_vals = 0;
            vals_shift = 0;
            spec_shift = 0;
            shift = 0;
            avg = 0;
            energy=0;

            vals = &input[loop*stepSize];
            for (c=0; c < frameSize; c++){
                // avg < 32768 * frameSize
                avg += vals[c];
                // energy < 32768^2 * frameSize / frameSize = 32768^2;
                energy += (vals[c]*vals[c] + frameSize/2) >> frameSizeBits;
                if (vals[c] > max_vals) max_vals = vals[c];
                if (vals[c] < min_vals) min_vals = vals[c];
            }

            features[3] = (float)energy / ((float)32768 * 32767) ;
            stdDev_cache+= features[3];

                // Find the largest absolute value of the new, zero mean vals.
            max_vals -= (avg + frameSize/2) >> frameSizeBits;
            min_vals -= (avg + frameSize/2) >> frameSizeBits;
            if (-min_vals > max_vals) max_vals = -min_vals;

            vals_shift = 14 - msb_16(max_vals);

                // avg < 32768
            shift = frameSizeBits - vals_shift;
            if (shift > 0)
                avg = (avg + (1 << (shift-1))) >> shift;
            else
                avg = avg << -shift;

                // Subtract mean and shift vals into vals_shifted;
                // multiply by hanning window into vals_spec,
                // put largest vals_spec in max_spc.
            if (vals_shift < 0) {
                for (i = 0; i < frameSize; i++) {
                    vals_shifted[i] = (vals[i] >> -vals_shift) - avg;
                    vals_spec[i] = (vals_shifted[i] * hanning[i]) >> 15;
                }
            } else {
                for (i = 0; i < frameSize; i++) {
                    vals_shifted[i] = (vals[i] << vals_shift) - avg;
                    vals_spec[i] = (vals_shifted[i] * hanning[i]) >> 15;
                }
            }


            max_inverse =0;

            /* calculate spec entropy*/

                // NOTE: this trashes vals_spec!
            kiss_fftr(cfg,vals_spec,outvals); // compare matlab: outvals * 256 >> vals_shift

            for (i = 0; i < frameSize/2; i++) {
                spec_sq[i] = (outvals[i].r * outvals[i].r + outvals[i].i * outvals[i].i); // =(abs(complex))^2
                if (spec_sq[i]> max_inverse) max_inverse =spec_sq[i];
            }

                // max_inverse == 1 << 15 -> max_inv_shift = 1
            max_inv_shift = msb_32(max_inverse) - 14;
            if (max_inv_shift < 0)
                max_inv_shift = 0;


                // sumspec < 32768/sqrt(2) * frameSize
                // i.e. sumspec bits: 14.5 + frameSizeBits = 22.5
            sumspec = 0;
            for (i = 0; i < frameSize/2; i++) {
                spec[i] = sqrt_x[spec_sq[i] >> max_inv_shift];
                sumspec += spec[i];
            }

            for (i = 0; i < frameSize/2; i++) {
                // spec[i] < 32768
                int ns_tmp = ((spec[i] << 16)/ sumspec) >> 1;
                if( 32768 < ns_tmp){
                    printf("warning wrong shifting 3\n");
                }
                assert(ns_tmp < 32768);

                normspec[i] = ns_tmp;
            }

            features_ent = 0;

            for (i = 0; i < frameSize/2; i++) {
                // index to log_x_times_x is value * 32768, where 0 < value < 1
                features_ent += log_x_times_x[normspec[i]] ;
            }


            features[2] = features_ent/x_log_x_scale; 
            // This quadratic correction is added by Mark to correct an error pattern observed by comparing outputs in Matlab
            features[2] = SpecEnt_A * features[2]*features[2] + (SpecEnt_B + 1 ) * features[2] + SpecEnt_C;
            

            /* find formant frequency (f_0) */

                // take sqrt, as per matlab code
                // only go up to fs/2.  [fs/2,fs] is redundant...
            for( i = 0; i < frameSize/2; i++ ) {
                big_fft_in[i] = sqrt_x[normspec[i]];	// big_fft_in shifted up by sqrt_shift
            }
                // pad w/ 0s
            for( i = frameSize/2; i < 2*frameSize; i++ ) {
                big_fft_in[i] = 0;
            }

                // big_fft_in: shifted up by sqrt_shift
                // NOTE: this trashes big_fft_in!
            kiss_fftr(cfg_big_fft, big_fft_in, big_fft);
                // big_fft: shifted up by ????????? sqrt_shift - (frameSizeBits + 1)/2

                // find max within interesting values
            int maxloc = min_index;
            for( i = min_index; i <= max_index; i++ ) {
                if( big_fft[i].r > big_fft[maxloc].r ) {
                    maxloc = i;
                }
            }

            features[4] = (float)(frameSize*2*sampling_rate)/(float)(maxloc*frameSize) + OFFSET_F0;



            /* compute auto correlation*/



                // energy = sum(vals^2) / 256
                // let vs = vals_shifted, v = vals, S = 1 << vals_shift
                // vs = v*S - avg
                // therefore v = (vs + avg) / S
                // so energy = sum((vs + avg)^2/S^2) / 256
                // = 1/S^2 * sum(vs^2 + 2*vs*avg + avg^2) / 256
                // = 1/S^2 * (sum(vs^2) + 2*avg*sum(vs) + N*avg^2) / 256
            //
                // Because we subtracted the average, sum(vs) = 0.
            //
                // So e * 256 * S^2 - 256*avg^2 = sum(vs^2)
                // sum(vs^2)/256 = e * S^2 - avg^2
            //

                //vals_shifted < 32768

            for (i = frameSize; i < 2*frameSize; i++) {
                vals_shifted[i] = 0;
            }

                // vals_shifted: shifted up by vals_shift
                // NOTE: this trashes vals_shifted!
            kiss_fftr(cfg_auto,vals_shifted,outvals);
                // outvals: shifted up by vals_shift - (frameSizeBits + 1)/2


                // tmp_inv: shifted up by 2*vals_shift - (frameSizeBits + 1)
            max_inverse = 0;
            for (i = 0; i <= frameSize; i++) {
                tmp_inv[i] = (int)(outvals[i].r * outvals[i].r + outvals[i].i * outvals[i].i);
                max_inverse =  (tmp_inv[i]> max_inverse) ? tmp_inv[i] : max_inverse;
            }

            max_inv_shift = msb_32(max_inverse) - 14;
            if (max_inv_shift < 0)
                max_inv_shift = 0;

                // vals_inv: shifted up by 2*vals_shift - (frameSizeBits + 1) - max_inv_shift
            for (i = 0; i < frameSize; i++) {
                vals_inv[i].r = (short)(tmp_inv[i] >> max_inv_shift);
                vals_inv[i].i = 0;
                vals_inv[2*frameSize -i].r = vals_inv[i].r;
                vals_inv[2*frameSize -i].i = 0;
            }

            i = frameSize;
            vals_inv[i].r = (short)(tmp_inv[i] >> max_inv_shift);
            vals_inv[i].i = 0;

                // NOTE: this trashes vals_inv!
            kiss_fft(cfg_inv, vals_inv, outvals);
                // outvals: shifted up by 2*vals_shift - 3*(frameSizeBits+1)/2 - max_inv_shift

            /* Multiply by comp to compensate for frame effects */

            for (i = 0; i < frameSize/2 ; i++) {
                acorr[i]= outvals[i].r * comp[i];
            }

            fast_find_acorr_peaks(acorr, stepSize, &maxpeak, &numpeaks);

                // energy_s is shifted up by 2*vals_shift
            if (vals_shift >= 0)
                energy_s = (energy << (2*vals_shift)) - avg*avg;
            else
                energy_s = (energy >> (-2*vals_shift)) - avg*avg;

                // stdDev is shifted up by -30
                // stdDev_int is unshifted

                // outvals: shifted up by 2*vals_shift - 3*(frameSizeBits+1)/2 - max_inv_shift
                // energy_s is shifted up by 2*vals_shift

                // outvals[0].r << (10 + max_inv_shift) == energy_s
                // acorr[0] << (max_inv_shift - 4) == energy_s


                // stdDev is shifted up by -30
                // stdDev_int is unshifted
                // energy_s is shifted up by 2*vals_shift
                // maxpeak is shifted up by 2*vals_shift - max_inv_shift + 4
            if (max_inv_shift >= 4)
                // This can't overflow since biggest maxpeak can be
                // after this shifting is energy_s, which fits in an int.
                maxpeak = maxpeak * (1 << (max_inv_shift - 4));
            else
                maxpeak = maxpeak / (1 << (4 - max_inv_shift));

            if (vals_shift > 0)
                maxpeak = maxpeak / (energy_s + stdDev_int * (float)(1 << (2*vals_shift)));
            else
                maxpeak = maxpeak / (energy_s + stdDev_int / (float)(1 << (-2*vals_shift)));

            features[0] = maxpeak;
            features[1] = numpeaks;
            
            if (!lightMode){
                // init values for voiced viterbi

            for (i=0;i<2;i++){
                float tmp = 0.0;

                tmp = ((features[0] - mus[i][0])/sigmas[i][0]);
                total = - (tmp*tmp);
                tmp = ((features[1] - mus[i][1])/sigmas[i][1]);
                total -= (tmp*tmp);
                tmp = ((features[2] - mus[i][2])/sigmas[i][2]);
                total -= (tmp*tmp);

                feature_prob_state[i] = total/2 + total_log_coeffs[i];
            }

            state = viterbi_voiced(feature_prob_state, trans_prob);
            state_cache[state_counter] = state;


                // init values for speaking viterbi
            prob_state_voiced[state_counter][0] = prob_voiced[state][0];
            prob_state_voiced[state_counter][1] = prob_voiced[state][1];
            state_counter = (state_counter+1) % VITERBI_PATH_LENGTH_SPEAKING;


                // don't do anything as long as voiced viterbi doesn't return proper values
            if (state != -1) {

                viterbicounter_speaking = (viterbicounter_speaking +1) % 700;

                if ((viterbicounter_speaking % 100) == 0){


                        // viterbi array is longer than samples read from it for accuracy
                    viterbi_speaking(states, init_state, prob_state_voiced, trans_probs_speaking,viterbicounter_speaking%VITERBI_PATH_LENGTH_SPEAKING);

                    for(i = 20; i<120;i++){

                        if (synch_speak_count <= synch_speak){
                            speakingvalues[i-20] = -1;
                            synch_speak_count++;
                        }
                        else{
                            speakingvalues[i-20] = states[i];
                            speak = states[i];
                            speakSum += speak;
                            if( !lastSpeak && speak ) // this is new speak segment
                                speakSegCount++;
                            lastSpeak = (speak ? true : false );

                        }
                    }

                    SignalGenAccept(speaking_sig);
                    gettimeofday(&outTimestamp, NULL);
                    SignalSend(speaking_sig, &outTimestamp, (short *)speakingvalues);
                        
                
                }
                        
            }


                // cache values untel voiced is synchronized with features
            if (synch_voiced_count <= synch_voiced)
            {
                memcpy(&features_cache[synch_voiced_count][0],&features[0],sizeof(float)*5);
                voicedvalues[send_counter] = -1;
                synch_voiced_count++;
            }
            else
            {
                voicedvalues[send_counter] = state;

                voiced = state;
                voicedSum += voiced;

                // only calculate statistics if segment is voiced
                if( state )
                {
                    formantSum += features_cache[0][4];
                    formantSqSum += features_cache[0][4]*features_cache[0][4];
                    specSum += features_cache[0][2];
                    specSqSum += features_cache[0][2]*features_cache[0][2];
                    energySum += features_cache[0][3];
                    energySqSum += features_cache[0][3]*features_cache[0][3];
                }
                // shift all values to front and new features to end
                memmove(&features_cache[0][0],&features_cache[1][0],sizeof(float)*5*29);
                memcpy(&features_cache[29][0],&features[0],sizeof(float)*5);

                // count number of new voiced segments
                if( !lastVoiced && voiced )
                    voicedSegCount++;

                // update last values
                lastVoiced = (voiced ? true : false );
            }
                
            }
            
            memcpy(&output[send_counter][0],&features[0],sizeof(float)*5);
            send_counter++;
            
            memcpy(&bigBuffer[bigCounter][0],&features[0],sizeof(float)*5);
            bigCounter++;
            

            if (send_counter >= 100){
                if (!lightMode){
                SignalGenAccept(voiced_sig);
                gettimeofday(&outTimestamp, NULL);
                SignalSend(voiced_sig, &outTimestamp, (short *)voicedvalues);
                }
                SignalGenAccept(features_sig);
                gettimeofday(&outTimestamp, NULL);
                SignalSend(features_sig, &outTimestamp, (float *)output);
                        
                if (send_counter >= 99){
                    if (!permanentSet && speakSum > 80 ) {
                        stdDev = 0.95 * stdDev + 0.05* stdDev_cache/(send_counter*25);					
                        stdDev_int = (int)(stdDev * (float) (1 << 30));
                        stdDev_cache = 0;

                    }
                }



                send_counter = 0;
            }

            if (bigCounter >= BIG_WINDOWSIZE){
                
                for (i=0; i<5; i++) {
                    averagedFeatures[i] = CalcMeanColumn(bigBuffer, BIG_WINDOWSIZE, i);// bigFeaturesNormalization[i];
                    averagedFeatures[i+5] = CalcVarColumn(bigBuffer, BIG_WINDOWSIZE, i, averagedFeatures[i]) / bigFeaturesNormalization[i+5];
                    averagedFeatures[i] = averagedFeatures[i] / bigFeaturesNormalization[i];

                }
                
                SignalGenAccept(bigFeatures_sig);
                gettimeofday(&outTimestamp, NULL);
                SignalSend(bigFeatures_sig, &outTimestamp, averagedFeatures);
                bigCounter = 0;
            }
            
            sample_counter++;
            
                // post statistics if true
            if ((sample_counter > SAMPLES_PER_MINUTE*NUM_MINUTES) && !lightMode)
            {
                formantMean = formantSum / (float)voicedSum;
                specMean = specSum / (float)voicedSum;
                energyMean = energySum / (float)voicedSum;

                formantStd = sqrt((formantSqSum / (float)voicedSum) - formantMean*formantMean)/(float)formantMean;
                specStd = sqrt((specSqSum / (float)voicedSum) - specMean*specMean);
                energyStd = sqrt((energySqSum / (float)voicedSum) - energyMean*energyMean)/(float)energyMean;

                //	  float voicedSegLenMean = (float)voicedSum / voicedSegCount;
                //	  float speakSegLenMean = (float)speakSum / speakSegCount;

                fracSpeak = (float)speakSum / (float)(SAMPLES_PER_MINUTE*NUM_MINUTES);
                voicedRate = (float)voicedSum / (60.*NUM_MINUTES);

                formantMeanNode->data.float_val = formantMean;
                specMeanNode->data.float_val = specMean;
                energyMeanNode->data.float_val = energyMean;
                formantStdNode->data.float_val = formantStd;
                specStdNode->data.float_val = specStd;
                energyStdNode->data.float_val = energyStd;
                fracSpeakNode->data.float_val = fracSpeak;
                voicedRateNode->data.float_val = voicedRate;

                if(!PublishNode(chunkUNL, node, NULL))
                {
                    printf("\n Publish Err");
                    return(-1);
                }

                formantSum = 0;
                formantSqSum = 0;
                specSum = 0;
                specSqSum = 0;
                energySum = 0;
                energySqSum = 0;
                voicedSum = 0;
                speakSum = 0;
                voicedSegCount = 0;
                speakSegCount = 0;
                lastVoiced = false;
                lastSpeak = false;
                sample_counter = 0;
            }
        } // end for
    } // end while


    free(buffer);
    free(input) ;
    free(features);
    free(voicedvalues);
    free(speakingvalues);
    free(cfg);
        
    PublishNodeInfo info;
    PublishNode(NULL, NULL, &info);
    if (!PublishNode(chunkUNL, NULL, &info))
    {
        fprintf(stderr, "Error unpublishing %s\n", chunkUNL);
        return 2;
    }

    return 0;
}



// finds max peak and num peak for a given array in acorr_slide
void fast_find_acorr_peaks(int *acorr_slide, int len, float *pmaxpeak, int *pnumpeaks){
    int i=0;
    int maxpeak = 0;
    int numpeaks = 0;
    int local_best = 1;


    while ( i< len && acorr_slide[i] >= 0 ) {
        i++;
    }

    while ( i < len){
        while ( i < len && acorr_slide[i] < 0) {i++;}
        local_best = -1;
        while (  i < len &&acorr_slide[i] >= 0) {
            if (acorr_slide[i] > 	local_best){
                local_best = 		acorr_slide[i];
            }
            i++;
        }
        if (local_best > 0){
            if (local_best > maxpeak) {
                maxpeak = local_best;
            }
            numpeaks++;
        }
    }

    if (local_best == 1) numpeaks = 1;

    *pnumpeaks = numpeaks;
    *pmaxpeak = (float)maxpeak;
    return;
}


// -1 - not enough data yet
// 0 - not voiced
// 1 - voiced
int viterbi_voiced(float obser_probs[2], float trans_prob[2][2])
{
    static bool first_time = true;
    static float delta[2];
    float mat[2][2];
    static int xi[VITERBI_PATH_LENGTH_VOICED][2];
    static int first_xi = VITERBI_PATH_LENGTH_VOICED-1;
    static int next_xi = VITERBI_PATH_LENGTH_VOICED-1;
    int tmp,i;
    int state = -1;

    next_xi = (next_xi + 1) % VITERBI_PATH_LENGTH_VOICED;

    if (first_time) {
        delta[0] = init_state[0] + obser_probs[0];
        delta[1] = init_state[1] + obser_probs[1];
    }
    first_time = false;


    mat[0][0] = delta[0] + trans_prob[0][0];
    mat[0][1] = delta[0] + trans_prob[0][1];
    mat[1][0] = delta[1] + trans_prob[1][0];
    mat[1][1] = delta[1] + trans_prob[1][1];

    if (mat[0][0] > mat[1][0]) {
        xi[next_xi][0] = 0;
        tmp = 0;
    }
    else {
        xi[next_xi][0] = 1;
        tmp = 1;
    }
    delta[0] = mat[tmp][0] + obser_probs[0];

    if (mat[0][1] > mat[1][1]) {
        xi[next_xi][1] = 0;
        tmp = 0;
    }
    else {
        xi[next_xi][1] = 1;
        tmp = 1;
    }
    delta[1] = mat[tmp][1] + obser_probs[1];


    if (delta[0] > delta[1]) {
        delta[1] = delta[1] - delta[0];
        delta[0] = delta[0] - delta[0];
    } else {
        delta[0] = delta[0] - delta[1];
        delta[1] = delta[1] - delta[1];
    }

// allow viterbi array to fill up. until then return state = -1
    if (next_xi == first_xi) {
        if (delta[0] > delta[1]) {
            state = 0;
        }
        else {
            state = 1;
        }

        i = next_xi;
        do {
            state = xi[i][state];
            i = i - 1;
            if (i < 0)
                i += VITERBI_PATH_LENGTH_VOICED;
        } while (i != first_xi);

        // when viterbi array is loaded next_xi == first_xi should always be kept true;
        first_xi = (first_xi +1) % VITERBI_PATH_LENGTH_VOICED;
    }

    return state;
}



void viterbi_speaking(int states[], float init_state[2],float obser_probs[][2], float trans_prob[][2],int start_position){

    float delta[2];
    float mat[2][2];
    int xi[VITERBI_PATH_LENGTH_SPEAKING][2];
    int prev_state,tmp;
    int i,c;

    delta[0] = init_state[0] + obser_probs[start_position%VITERBI_PATH_LENGTH_SPEAKING][0];
    delta[1] = init_state[1] + obser_probs[start_position%VITERBI_PATH_LENGTH_SPEAKING][1];

    for (i= start_position +1, c = 0; i< VITERBI_PATH_LENGTH_SPEAKING + start_position + 1; i++, c++){

        mat[0][0] = delta[0] + trans_prob[0][0];
        mat[0][1] = delta[0] + trans_prob[0][1];
        mat[1][0] = delta[1] + trans_prob[1][0];
        mat[1][1] = delta[1] + trans_prob[1][1];

        if (mat[0][0] > mat[1][0]) {
            xi[c][0] = 0;
            tmp = 0;
        }
        else {
            xi[c][0] = 1;
            tmp = 1;
        }
        delta[0] = mat[tmp][0] + obser_probs[i%VITERBI_PATH_LENGTH_SPEAKING][0];

        if (mat[0][1] > mat[1][1]) {
            xi[c][1] = 0;
            tmp = 0;
        }
        else {
            xi[c][1] = 1;
            tmp = 1;
        }
        delta[1] = mat[tmp][1] + obser_probs[i%VITERBI_PATH_LENGTH_SPEAKING][1];

        if (delta[0] > delta[1]) {
            delta[1] = delta[1] - delta[0];
            delta[0] = delta[0] - delta[0];
        }
        else {
            delta[0] = delta[0] - delta[1];
            delta[1] = delta[1] - delta[1];
        }
    }

    if (delta[0] > delta[1]) {
        prev_state = 0;
    }
    else {
        prev_state = 1;
    }

    for (i = VITERBI_PATH_LENGTH_SPEAKING - 1; i >= 0; i--){
        states[i] = xi[i][prev_state];
        prev_state = states[i];
    }
}


void SigHandler(int sig) {
    switch (sig) {
        case SIGALRM:
    //printf("Signal handler: Alarm received\n");
            break;
        case SIGINT:
    //puts("Signal handler: SIGINT (keyboard interrupt)");
            runflag=0;
            if(killflag) 
            {
                printf("Hard kill\n");
                exit(0);
            }
            else 
            {
                killflag = 1;
            }
            break;
        case SIGUSR1:
    //close nicely and exit     
            printf("Killed nicely.\n");
            exit(0);
            break;    
        default:
            printf("Signal handler received %d\n",sig);
    }
    return;
}
