/*****************************************************************************
 * File: MITeSignal.cpp
 * Author: Mark Blum (mark@media.mit.edu), MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Adapted from Sak2v2Signal.cpp
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************
 * This is the serial driver for MITes (MIT environmental sensors), accelero-
 * meters developped by the Media Lab group hous_n. The MITe-Receiver, which 
 * is connected to the serial port, is configured to listen on the channels 
 * requested. Acceleration values are posted as Enchantment Signals
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>
#include <math.h>
#include <strings.h>

#include <termios.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>

extern "C" {
#include <enchantment/whiteboard.h>
#include <enchantment/signals.h>
#include <enchantment/unl.h>
}

#define MITE_MAX_NUM_CHANNELS 6

#define MITE_DEBUG 0
#define MODEMDEVICE "/dev/ttyS0"
//#define MODEMDEVICE "/dev/ttyS1"  

#define BAUDRATE B57600  /*for Sak2: B115200*/
#define _POSIX_SOURCE 1 /* POSIX compliant source */
#define FALSE 0
#define TRUE 1

volatile int STOP=FALSE; 
int fd;
struct termios oldtio,newtio;

const int  kFailedReadsBeforeEcho = 2;
const int kLoopPause = 1;
int numValidChannels;

typedef struct {
  bool valid;
  Signal *sig;
  char *name;
  short *feature;
  struct timeval sampTime;
  int reqSampRate;
  int channelNumber;
  bool freshSample;
} ChannelRec;
    
void ProcessCommandLine(char *argv[], int argc, char **server, char **user, bool *printData);
void SetupSignalHandlers();
void SigHandler(int sig);
void SetupEnchantSignals(char *server, char *user);
int ReadChannels();  
void PrintChannels();
void SendSignals();
void UpdateSignalConnections(bool *bListeners);
int ConfigureMITeChannels();
int VerifyMITeChannels();
void CloseDown();
int MITe_init_serial();
int MITe_close_serial();
int SpinRead(char *buffer, int size);
int SpinWrite(char *buffer, int size);

volatile int runflag = 1;
volatile int killflag = 0;

struct timeval packTime;

ChannelRec channelRecs[MITE_MAX_NUM_CHANNELS];

int main(int argc, char *argv[]) 
{ 
  printf("start\n");
  char *server = NULL, *user = NULL;
  bool bListeners(false), bPrintData(false), bReadOnce(false);

  ProcessCommandLine(argv, argc, &server, &user, &bPrintData);
  
  SetupSignalHandlers();
  if(MITE_DEBUG) printf("SetupSignalHandlers() done\n");  
  
  SetupEnchantSignals(server, user);
  if(MITE_DEBUG) printf("SetupEnchantSignals() done\n");  
  
  while(runflag)
    {

      if(MITE_DEBUG) printf("Starting loop: Pause %d seconds.\n", kLoopPause);
      sleep(kLoopPause);

      if(!MITe_init_serial())
      {
        printf("error: Failed to init serial connection.\n");
        CloseDown();
        continue;
      }
      if(!ConfigureMITeChannels())
      {
          printf("error: Failed to configure MITe Receiver.\n");
	  CloseDown();
	  continue;
      }
      if(!VerifyMITeChannels())
      {
          printf("error: Failed to verify MITs Receiver configuration.\n");
	  CloseDown();
	  continue;
      }
      
      printf("MITeSignal: starting read cycle\n");
      int failedReads = 0;

      while(runflag) 
	{
	  if(ReadChannels()) 
	    {
	      bReadOnce = true;
	      failedReads = 0;

	      if(bPrintData)
		{
		  PrintChannels();
		}
	      
	      if(bListeners) SendSignals();
	    }
	  else
	    {
	      //We didn't have a successful read, but don't do anything
	      //unless we get a bunch in a row.
	      if(++failedReads < kFailedReadsBeforeEcho) { continue; }
	      //Haven't heard from the Sak2 in a while.  Echo the board to
	      //see if it is still connected and functional
	      else { fprintf(stderr, "error: MITe Receiver won't respond after several attempts!\n"); break; }
	    }
	  if(bReadOnce) UpdateSignalConnections(&bListeners);
	  killflag = 0;
	}
      CloseDown();
    }
  
  // free memory allocated for channelRec names
  for(int k=0; k < MITE_MAX_NUM_CHANNELS; k++)
  {
    if (channelRecs[k].valid) free(channelRecs[k].name);
  }
  fflush(stdout);
  printf("Exiting nicely.\n");     // Exiting will close signals.
  return 0;
}

void CloseDown()
{
  int ret;

  if(MITE_DEBUG) printf("CloseDown...");
  
  fflush(NULL);
  ret = MITe_close_serial();
  
  if(MITE_DEBUG) printf("  ...DONE\n");
} 
   
int ConfigureMITeChannels()
{
  char outbuffer[8];
  int i;
  
  if(MITE_DEBUG) printf("ConfigureMITeChannels...\n");
  
  //this to set the channels
  // send '9' <numer_of_active_channels> <channel 1> <channel 2>  ...
  outbuffer[0] = '9'; 
  outbuffer[1] = numValidChannels; 
  
  for(i = 0; i < MITE_MAX_NUM_CHANNELS; i++)
  {
    if (channelRecs[i].valid){
      outbuffer[2+i] = channelRecs[i].channelNumber;
    }
    else{
      outbuffer[2+i] = 0;
    }
  }
    
  if(MITE_DEBUG) printf("Sending '9' %d %d %d %d %d %d %d\n",outbuffer[1],outbuffer[2],outbuffer[3],outbuffer[4],outbuffer[5],outbuffer[6],outbuffer[7]);
  if (!SpinWrite(outbuffer,1)) {printf("error: SpinWrite() failed\n"); return 0;}
  usleep(150000);  //this seems necessary for successful communication (?)
  if (!SpinWrite(outbuffer+1,7)) {printf("error: SpinWrite() failed\n"); return 0;}
//   for (i=1; i<8; i++){
// 	if (SpinWrite(outbuffer+i,1))
//   }

  //this to set the receiver address (default is 255)
  // send '5' <receiver_address>
  usleep(200000);
  outbuffer[0] = '5'; outbuffer[1] = 255;
  
  if(MITE_DEBUG) printf("Sending '5' %d\n",outbuffer[1]);
  if (!SpinWrite(outbuffer,1)) {printf("error: SpinWrite() failed\n"); return 0;}
  usleep(150000);
  if (!SpinWrite(outbuffer+1,1)) {printf("error: SpinWrite() failed\n"); return 0;}
 
  return 1;
}

int VerifyMITeChannels()
{
  int numRead = 50; 
  char outbuffer[1], buffer[50];     // numRead MUST equal size of buffer !!!!!!!!
  bool ret1(false), ret2(false);
  int i;
  
  // this read cycle makes sure the serial in-buffer is empty
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}
  if(MITE_DEBUG) printf("VerifyMITeChannels...\n");
  
  //this to check channel settings
  // send '0' ---> Reveice 'N' 'C' 'H' 'A' <numer_of_active_channels> <channel 1> <channel 2>  ...
  outbuffer[0] =  '0'; 
  if (!SpinWrite(outbuffer,1)) {printf("error: SpinWrite() failed\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}  //it will take about 10 to 20 samples for the Receiver to responde, 50 is safe
  
  if (MITE_DEBUG) {
    for (i=0; i<numRead; i++)  printf("%c ",buffer[i]);
    printf("\n");
  }
  for (i=0; i<numRead; i++){ 
    if (((buffer[i] == 'N') && (buffer[i+1] == 'C')) && ((buffer[i+2] == 'H') && (buffer[i+3] == 'A'))) {
      if (buffer[i+4] != numValidChannels) {printf("error: set number of channels not correct. (%d)\n",buffer[i+4]); return 0;};
      for (int c=0; c<numValidChannels; c++){ 
        if (buffer[i+5+c] != channelRecs[c].channelNumber) {printf("error: number of channel %d not correct, was %d, expected %d\n",c,buffer[i+5+c],channelRecs[c].channelNumber); return 0;};
      }
      ret1 = true; 	//if we get this far, the channels were correct.   
      break;
    }
    if (i==numRead-1) printf("error: no NCHA received!\n");
  }
  
  //this is to check receiver address (default is 255)
  // send '6' ---> Reveice 'R' 'E' 'C' 'E' <reveicer_address>
  outbuffer[0] =  '6';  
  if (!SpinWrite(outbuffer,1)) {printf("error: SpinWrite() failed\n"); return 0;}
  if (!SpinRead(buffer,numRead)) {printf("error: SpinRead() failed.\n"); return 0;}  //it will take about 10 to 20 samples for the Receiver to respond, 50 is safe
  
  if (MITE_DEBUG) for (i=0; i<numRead; i++)  printf("%c ",buffer[i]);
  for (i=0; i<numRead; i++){  
    if ((buffer[i] == 'R') && (buffer[i+1] == 'E') && (buffer[i+2] == 'C') && (buffer[i+3] == 'E')) {
      if (((unsigned char)buffer[i+4]) != 255) {
	printf("error: Receiver address not correct (%d).\n",buffer[i+4]); 
	return 0;
      }
      ret2 = true; 	//if we get this far, the receiver address was correct.   
      break;
    } 
    if (i==numRead-1) printf("error: no RECE received!\n");
  }
  return (ret1 && ret2);
}

void PrintUsage()
{
  printf("Usage:MITesSignal"
	 "\n\t-c <number of channels> <channel #1> <channel #2> ...\t\t"
	 "\n\t-s <context server name>"
	 "\n\t[-u <user id>]"
	 "\n\t[-p]\t\t\t\t#Print data to stdout\n");
  printf("Default options:"
	 "\n\t-s <localhost>\n");
}

void ProcessCommandLine(char *argv[], int argc, char **server, char **user, bool *printData)
{
  bool serverSet(false), userSet(false), channelsAreSet(false);
  int i;
  char* buffer;
    
  *printData = false;

  for(i = 0; i < MITE_MAX_NUM_CHANNELS; i++)
  {
      channelRecs[i].valid = false;
      channelRecs[i].name = NULL;      
      channelRecs[i].feature = NULL;      
      channelRecs[i].channelNumber = 0;      
  }
  if (MITE_DEBUG) printf("Parsing input parameters\n");
    
  for(i = 1; i < argc; i++)
    {
      if(argv[i][0] == '-')
	switch(argv[i][1]) {
	case 'p':
	  *printData = true;
	  break;
	case 's':
	  i++;
	  *server = argv[i];
	  printf("Context server is %s.\n", argv[i]);
	  serverSet = true;
	  break;
	case 'u':
	  i++;
	  *user = argv[i];
	  printf("MITe signal user id %s.\n", argv[i]);
	  userSet = true;
	  break;
	case 'c':
          i++;	  
          numValidChannels = atoi(argv[i]);
        
          for(int k=0; k < numValidChannels; k++)
	    {
              i++;
              channelsAreSet = true;
              channelRecs[k].channelNumber = atoi(argv[i]);
              buffer = (char*) malloc(10);
              sprintf(buffer,"MITe_%d",channelRecs[k].channelNumber);
              channelRecs[k].name = buffer;
              channelRecs[k].valid = true;
              channelRecs[k].reqSampRate = -1;   //not used for anything
              printf("Reading MITe on channel %d --> Name is %s\n",channelRecs[k].channelNumber,buffer);
            }
	  break;
	case 'h':
	default:
	  PrintUsage();	      
	  exit(1);
	} 
    }
  
  if(!channelsAreSet) 
    {
      printf("There are no channels set for reading.\n");
      PrintUsage();
      exit(1);
    }

  if(!serverSet) 
    {
      char *eserver = getenv("MITE_SIGNAL_SERVER");  //probably not of use.
    
      if(eserver) *server = eserver;
      else *server = "localhost";
    }
}

void SetupSignalHandlers()
{
 
  struct itimerval *value =
    (struct itimerval *)(malloc(sizeof(struct itimerval)));;

  long loopTime = 20905l - 10000l; // something like 10ms overhead
  //  long loopTime = 31250l - 10000l; // something like 10ms overhead?

  //signal(SIGALRM,SigHandler);
  signal(SIGINT,SigHandler);
  signal(SIGUSR1,SigHandler);
  
  // Set timer value and reset value
  value->it_interval.tv_sec=0;
  value->it_interval.tv_usec=loopTime;
  value->it_value.tv_sec=0;
  value->it_value.tv_usec=loopTime;
  
  // Set real-time timer
  //setitimer(ITIMER_REAL,value,NULL);
  
  // Until we are interrupted...
}


void SigHandler(int sig) {
  switch (sig) {
  case SIGALRM:
    //printf("Signal handler: Alarm received\n");
    break;
  case SIGINT:
    //puts("Signal handler: SIGINT (keyboard interrupt)");
    runflag=0;
    if(killflag) 
      {
	printf("Hard kill, perhaps restart board.\n");
	exit(0);
      }
    else 
      {
	killflag = 1;
      }
    break;
  case SIGUSR1:
    //close nicely and exit
    MITe_close_serial();
    
    printf("Killed nicely.\n");
    exit(0);
    break;    
  default:
    printf("Signal handler received %d\n",sig);
  }
  return;
}

void SetupEnchantSignals(char *server, char *user)
{
  char path[128];
    
  for(int i = 0; i < MITE_MAX_NUM_CHANNELS; i++)
    {
      if(channelRecs[i].valid)
	{
	  if(channelRecs[i].name)
	    {
	      if(user) { sprintf(path, "signals/MITe/%s/%s", channelRecs[i].name, user); }
	      else { sprintf(path, "signals/MITe/%s", channelRecs[i].name); }
	    }
	  const char *chanUNL = NULL;
	  chanUNL = CreateUNL(server, path);
          if (MITE_DEBUG) printf("MITe connected on path: %s",path);
          
	  if(!(channelRecs[i].sig = NewSignalGenUNL(chanUNL)))
	  {
	    printf("Error creating signal generator for channel %d.\n", i);
	    exit(1);
	  }
	}
    }
}

void UpdateSignalConnections(bool *bListeners)
{
  *bListeners = false;

  // First, accept listeners on all the signals 
  for(int c = 0; c < MITE_MAX_NUM_CHANNELS; c++)
    {
      if(channelRecs[c].valid && channelRecs[c].feature) 
	{
	  SignalGenAccept(channelRecs[c].sig);
	  
	  if(SignalClientCount(channelRecs[c].sig) > 0)
	    {
	      *bListeners = true;
	    }    
	}
    }
}


int ReadChannels()
{
  char buff[4];
  int bReadPacket = 0, bReadChannels = 0, i;
  int value[3];
  
  for (i=0; i<MITE_MAX_NUM_CHANNELS; i++) {channelRecs[i].freshSample = false;}  //reset fresh state
   
  // packet structur:  'D' 'D' <channel number> <X-axis> <Y-axis> <Z-axis> <EMPTY> <next channel number> ...
  // or interpreted as:    'D' <channel number> <X-axis> <Y-axis> <Z-axis> <EMPTY> <next channel number> ...   ... 'D'
  
  while(!bReadPacket && runflag)
  {
        if(!SpinRead(buff,1)) {printf("error: SpinRead() failed.\n"); return 0;}
        if(buff[0] == 'D' && !bReadChannels) continue; //no channel identifier found --> we keep current state
        if(buff[0] == 'D' && bReadChannels) {
          bReadPacket = true;                          //we have found a closing 'D' after reading some channel data --> packet done
          continue;
        }
        
        for(i=0; i<MITE_MAX_NUM_CHANNELS; i++)
        {
          if(!channelRecs[i].valid) continue;   //no active channel
          if (buff[0] == channelRecs[i].channelNumber) {     //yes, we receive the data of a valid channel
              if(!SpinRead(buff,4)) {printf("error: SpinRead() failed.\n"); return 0;}
              if(channelRecs[i].feature == NULL) {           //this is only for the first call to initialize the feature (legacy from Sak2v2Signal)
		SignalSetType(channelRecs[i].sig, T_INT16);
		SignalSetDimension(channelRecs[i].sig, 1, 3, NULL, NULL);
		channelRecs[i].feature = (short*) malloc(SignalBufferSize(channelRecs[i].sig));
	      }
	      value[0] = (int)((unsigned char)buff[0]) + (int)((((unsigned char)buff[3]) & 0xc0) << 2);
              value[1] = (int)((unsigned char)buff[1]) + (int)((((unsigned char)buff[3]) & 0x30) << 4);
              value[2] = (int)((unsigned char)buff[2]) + (int)((((unsigned char)buff[3]) & 0x0c) << 6);

              for(int b = 0; b < 3; b++) {
                //channelRecs[i].feature->setElement(b, ((int)((signed char)buff[b])));      //perhaps interesting to shift the range of the data
                // it seems that the mean level (= zero g) is at about 170. So we deduct 42 to reach 128. This should solve the overflow problem.
                //temp = buff[b] - 42;   
                channelRecs[i].feature[b] = value[b]; 
	      }
              channelRecs[i].freshSample = true;
              bReadChannels = true;
              break;   //no need to continue checking other channels
           }
        }
  }  
  
  gettimeofday(&packTime, NULL);  

  for(i=0; i<MITE_MAX_NUM_CHANNELS; i++)       //timestamp all fresh samples
  {
    if (channelRecs[i].freshSample){
      channelRecs[i].sampTime = packTime; 
    }
  }
    
  return 1;
}


void SendSignals()
{
  for(int i = 0; i < MITE_MAX_NUM_CHANNELS; i++)
    {
      if(channelRecs[i].valid && 
	 (SignalClientCount(channelRecs[i].sig) > 0) &&
	 channelRecs[i].feature &&
	 channelRecs[i].freshSample)
	{
	  SignalSend(channelRecs[i].sig, &channelRecs[i].sampTime, channelRecs[i].feature);
	}
    }
}


void PrintChannels()
{
  for(int i = 0; i < MITE_MAX_NUM_CHANNELS; i++)
    {
      if(channelRecs[i].valid &&
	 channelRecs[i].feature &&
	 channelRecs[i].freshSample)
	{
	  
	  if(channelRecs[i].name) printf("%s: ", channelRecs[i].name);
	  else printf("dev %d: ", i);
	  
	  PrintSignalElements(stdout, channelRecs[i].sig, 
			      channelRecs[i].sampTime, channelRecs[i].feature);
	  
	}
    }
}


int MITe_init_serial(){                   //originally from sak2v2_serial.c
  
  fd = open(MODEMDEVICE, O_RDWR | O_NOCTTY ); 
  if (fd <0) {perror(MODEMDEVICE); exit(-1); }
  /* this is a comment */
  tcgetattr(fd,&oldtio); /* save current port settings */
  
  bzero(&newtio, sizeof(newtio));
  newtio.c_cflag = BAUDRATE | CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;
  
  /* set input mode (non-canonical, no echo,...) */
  newtio.c_lflag = 0;
   
  newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
  newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 chars received */
  //was 1
  
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd,TCSANOW,&newtio);
  
  return 1;
}

int MITe_close_serial()                 //originally from sak2v2_serial.c
{
  tcsetattr(fd,TCSANOW,&oldtio);
  return 1;
}


int SpinRead(char *buffer, int size)     //originally from sak2v2_serial.c
{
  int readBytes = 0;
  int newBytes;
  fd_set readSet;
  struct timeval tv;
  int retval;
  int i;

  FD_ZERO(&readSet);
  FD_SET(fd, &readSet);

  while(readBytes < size)
  {
    tv.tv_sec = 5;
    tv.tv_usec = 0;

    retval = select(fd+1, &readSet, NULL, NULL, &tv);

    if(retval)
    {
	  //fprintf(stderr, "<read:");
      newBytes = read(fd, buffer + readBytes, size - readBytes);
	  //fprintf(stderr, "byte>\n");
    }
    else
    {
      if(MITE_DEBUG) 
      {
        fprintf(stderr, "READ ERROR: got %d bytes of %d\n", readBytes, size);
	      
        for(i = 0; i < readBytes; i++)
        {
          fprintf(stderr, "%d ", buffer[i]);
        }	    
        fprintf(stderr, "\n");
      }
	  //set disconnect flag
	  //sleep?
      return 0;
    }
      
    readBytes += newBytes;
  }
  
  return 1;
}

int SpinWrite(char *buffer, int size)     //originally from sak2v2_serial.c
{
  int writtenBytes = 0;
  int newBytes;

  while(writtenBytes < size)
  {
    newBytes = write(fd, buffer + writtenBytes, size - writtenBytes);
      
    if(newBytes < 0)
    {
      if(MITE_DEBUG) fprintf(stderr, "WRITE ERROR\n");
	  //set disconnect flag
	  //sleep?
      return 0;
    }
      
    writtenBytes += newBytes;
  }
  
  return 1;

}









