/*****************************************************************************
*
* Author: Jonathan Gips, MIT Wearable Computing Group
* http://www.media.mit.edu/wearables/
* Copyright (C) 2005 MIT Media Lab
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************
* This class models a chain of the influence model. Includes XML parsing
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <strings.h>
#include <assert.h>

extern "C" {
#include <enchantment/matrix.h>
#include <enchantment/vector.h>
}

#include "Chain.hpp"


Chain::Chain() 
  : name(NULL), obsUNL(NULL), stateUNL(NULL),
    priors(NULL), B_disc(NULL), B_cont(NULL), 
    priorsXML(NULL), B_discXML(NULL), xml_state(SKIP)
{
  obsType = NONE;
}

Chain::~Chain() {
  clear();
}

void Chain::show() {

  printf("Chain:\n");
  printf("Name: %s\n", name);
  printf("Observation UNL: %s\n", obsUNL);
  printf("State UNL: %s\n", stateUNL);
  printf("Num States: %d\n", numStates);
  printf("State Priors:\n");
  PrintVector(stdout, priors);
  printf("Observation Dimensionality: %d\n", dimObs);
  if(obsType == DISCRETE) {
    printf("Observation Type: Discrete\n");
    printf("B:\n");
    PrintMatrix(stdout, B_disc);
  } else {
    printf("Observation Type: Continuous\n");
    for(int i = 0; i < numStates; i++) {
      printf("B %d:\n", i);
      getContinuousB(i)->show();
    }
  }
}

void Chain::clear(void)
{
  //do not clear dimension as it is only passed in constructor!
  if(name) free(name);
  if(obsUNL) free(obsUNL);//not DString
  if(stateUNL) free(stateUNL);
  if(priors) free(priors);
  if(B_cont) delete[] B_cont;
  if(B_disc) FreeMatrix(B_disc);
  if(B_discXML) delete B_discXML;
  if(priorsXML) delete priorsXML;

  name = NULL;
  obsUNL = NULL;
  stateUNL = NULL;
  priors = NULL;
  B_cont = NULL;
  B_disc = NULL;
  B_discXML = NULL;
  priorsXML = NULL;
  obsType = NONE;
}

bool Chain::setStateUNL(const char *s)
{
    if (stateUNL != NULL) free(stateUNL);
    stateUNL = (char*) malloc(sizeof(char)*(strlen(s)+1));
    strcpy(stateUNL,s);
    
    return true;
}

bool Chain::setObservationUNL(const char *s)
{
    if (obsUNL != NULL) free(obsUNL);
    obsUNL = (char*) malloc(sizeof(char)*(strlen(s)+1));
    strcpy(obsUNL,s);
    
    return true;
}


//--------------------------------------------------------------------

void Chain::startXMLElement(const char *name, const char **atts) 
{
  if(!strcasecmp(name, "chain")) {
    xml_state = CHAIN;
    clear();
    bLoadIdx = 0;
  }
  else if(!strcasecmp(name, "name")) {
    xml_state = NAME;
  } 
  else if(!strcasecmp(name, "observationunl")) {
    xml_state = OBSUNL;
  } 
  else if(!strcasecmp(name, "stateunl")) {
    xml_state = STATEUNL;
  } 
  else if(!strcasecmp(name, "numstates")) {
    xml_state = NUMSTATES;
  } 
  else if(!strcasecmp(name, "priors")) {
    xml_state = PRIORS;
    if(!(numStates > 0)) assert(0);
    priorsXML = new XMLVector(numStates);
    pushXMLObject(priorsXML, name, atts);
  }
  else if(!strcasecmp(name, "observationtype")) {
    xml_state = OBSTYPE;
  } 
  else if(!strcasecmp(name, "dimobservations")) {
    xml_state = DIMOBS;
  } 
  else if(!strcasecmp(name, "b_discrete")) {
    xml_state = B_DISC;    
    assert(numStates > 0 && dimObs > 0);
    assert(obsType == DISCRETE);

    B_discXML = new XMLMatrix(numStates,dimObs);
    pushXMLObject(B_discXML, name, atts);
  } 
  else if(!strcasecmp(name, "b_continuous")) {
    xml_state = B_CONT;
    assert(numStates > 0 && dimObs > 0);
    assert(obsType == CONTINUOUS);

    if(!B_cont)
      B_cont = new Gaussian[numStates](dimObs);
  }
  else if(!strcasecmp(name, "gaussian")) {
    pushXMLObject(&B_cont[bLoadIdx++], name, atts);
  }
  else {
    printf("Got unknown start tag \"%s\".\n", name);
    assert(0);
  }
}

void Chain::processXMLCharacterData(const char *s, int len)
{
  switch(xml_state) {
  case NAME:
    name = MyStrndup(s,len);
    break;
  case OBSUNL:
    obsUNL = MyStrndup(s,len);
    break;
  case STATEUNL:
    stateUNL = MyStrndup(s,len);
    break;
  case NUMSTATES:
    numStates = atoi(s);
    break;
  case DIMOBS:
    dimObs = atoi(s);
    break;
  case OBSTYPE:
    if(!strncasecmp(s,"discrete", strlen("discrete"))) {
      obsType = DISCRETE;
    } else if(!strncasecmp(s,"continuous", strlen("continuous"))) {
      obsType = CONTINUOUS;
    } else assert(0);
    break;
  default:
    break;
  }
  xml_state = SKIP; //only one parsing necessary per state
} 

void Chain::endXMLElement(const char *name)
{
  if(!strcasecmp(name, "chain")) {
    assert((obsType == CONTINUOUS && B_cont != NULL) || 
	   (obsType == DISCRETE && B_discXML != NULL ));
    assert(priorsXML != NULL);
    
    if(obsType == DISCRETE) {
      B_disc = CopyMatrix(B_discXML->m);
    }
    priors = CopyVector(priorsXML->v);
    popXMLObject();
  }
}
