/*****************************************************************************
*
* Author: Steve Dunn, MIT Wearable Computing Group
* http://www.media.mit.edu/wearables/
* Copyright (C) 2002 MIT Media Lab
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************
* Modified by Mark Blum (mark@media.mit.edu) in July 2005
* 
* The Classification class is used by GClassifier. It contains funcionality
* to manage labeled probabiliy vectors.
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern "C" {
#include <enchantment/node.h>
#include <enchantment/whiteboard.h>
}

#include "Classification.hpp"

Classification::Classification() {
    numElements = 0;
    contextLabel = NULL;
    postUNL = NULL;
    labels = NULL;
    probability = NULL;
}

Classification::Classification(char* contextLabel_in, char* postUNL_in) {
    numElements = 0;
    contextLabel = contextLabel_in;
    postUNL = postUNL_in;
    labels = NULL;
    probability = NULL;
}

Classification::~Classification() {
    empty();
//     if (contextLabel != NULL)
//         free(contextLabel);
}

//--------------------------------------------------

int Classification::getNumElements() {
    return numElements;
}

char* Classification::getLabel(int index) {
    if(index < 0 || index >= numElements)
        return 0;
    else
        return labels[index];
}

int Classification::getIndex(char *label) {
    int i;

    for(i=0; i<numElements; i++) {
        if(strcmp(labels[i], label) == 0)
            return i;
    }

    return -1;
}

double Classification::getProbability(int index) {
    if(index < 0 || index >= numElements)
        return MIN_PROBABILITY;
    else
        return probability[index];
}

double Classification::getProbability(char *label) {
    int index;
    index = getIndex(label);
    return getProbability(index);
}

void Classification::getProbabilities(double * buffer) {
    for(int i=0; i<numElements; i++){
        buffer[i] = probability[i];
    }
}


void Classification::setProbability(int index, double prob) {
    if(index < 0 || index >= numElements)
        return;
    else
        probability[index] = prob;
}

void Classification::setElement(int index, char* label, double prob) {
    if(index < 0 || index >= numElements)
        return;

    if(labels[index] != NULL)
        free(labels[index]);

    labels[index] = strdup(label);
    probability[index] = prob;
}

void Classification::appendElement(char* label, double prob) {
    double *tempP;
    char **tempC;
    int i;

        // Labels
    tempC = new char*[numElements+1];
    for(i=0; i<numElements; i++)
        tempC[i] = labels[i];

    if(numElements > 0)
        delete[] labels;
    labels = tempC;
    labels[numElements] = strdup(label);

        // Probs
    tempP = new double[numElements+1];
    for(i=0; i<numElements; i++)
        tempP[i] = probability[i];

    if(numElements > 0)
        delete[] probability;
    probability = tempP;
    probability[numElements] = prob;

    numElements++;
}

void Classification::addToLabel(char *label, double prob) {
    int i;
        
    i = getIndex(label);

    if(i == -1)
        appendElement(label, prob);
    else
        probability[i] += prob;
}

//--------------------------------------------------

// void Classification::publish(Context *context, char *unl) {
// 	// printf("Classification::Publish to %s\n", unl);
// 
// 	NodeType *updateNode;
// 	NodeType *labelNode;
// 	char *nodeName;
// 	int i;
// 
// 	if(context == NULL)
// 		return;
// 
// 	nodeName = strrchr(unl, '/');
// 	if(nodeName == NULL) {
// 	  printf("Error extracting node name from partial unl\n");
// 		return;
// 	}
// 	nodeName++;	
// 
// 	updateNode = CreateBranchNode(nodeName, T_BRANCH, 1);
// 
// 	for(i=0; i<numElements; i++) {
// 		// printf("Publishing node - element: %s - %.3f\n", labels[i], probability[i]);
// 		if(probability[i] >= 0.0) {
// 			labelNode = CreateNode(labels[i], T_FLOAT, 4, NULL, FALSE);
// 
// 			InsertNode(updateNode, labelNode);
// 			labelNode->data.float_val = (float) probability[i];
// 		}
// 	}
// 	context->overwriteNode(updateNode, unl);
// 
// 	// PrintNode(updateNode, NULL);
// 	DeleteNode(NULL, updateNode);
// }

void Classification::publish() {
    if (postUNL != NULL){
        publish(postUNL);
    }
    else
    {
        printf("Classification: Error, postUNL not specified!\n");
    }
}

void Classification::publish(char *fullUNL) {
// printf("Classification::Publish to %s\n", unl);

    NodeType *updateNode;
    NodeType *labelNode;
    PublishNodeInfo *pubInfo = NULL;
    Vector *probabilityVector = NewVector(numElements);
    char *nodeName;
    int i;

    nodeName = strrchr(fullUNL, '/');
    if(nodeName == NULL) {
        printf("Classification: Error extracting node name from partial unl\n");
        return;
    }
    nodeName++;	

    updateNode = CreateBranchNode(nodeName, T_BRANCH);

    for(i=0; i<numElements; i++) {
         // printf("Publishing node - element: %s - %.3f\n", labels[i], probability[i]);
        if(probability[i] >= 0.0) {
            labelNode = CreateFloatNode(labels[i], T_DOUBLE, (double) probability[i]);
            InsertNode(updateNode, labelNode);
            
        }
    }
    
    labelNode = CreateStringNode("decision", T_STRING, this->mostLikely());
    InsertNode(updateNode, labelNode);
    
    for (i = 0; i<numElements; i++)
        VectorSet(probabilityVector, i, (double)(probability[i]));
    labelNode = CreateVectorNode("probs", T_VECTOR, probabilityVector);
    InsertNode(updateNode, labelNode);
    //PrintNode(updateNode, NULL);
    //PublishNode(NULL, NULL, &pubInfo);
    if (!PublishNode(fullUNL,updateNode, pubInfo))
        printf("Classification.cpp: Error publishing node\n");
            
    FreeNode(updateNode);
}

void Classification::publishVector(char *postUNL) {
    //if would be more efficient to call a initPublications function that sets up the node
    NodeType *updateNode;
    PublishNodeInfo *pubInfo = NULL;
    Vector *probabilityVector = NewVector(numElements);
    char *nodeName = strrchr(postUNL, '/');
    if(nodeName == NULL) {
        printf("Classification: Error extracting node name from partial unl %s\n",postUNL);
        exit(1);
    }
    nodeName++;
    updateNode = CreateVectorNode(nodeName, T_VECTOR, probabilityVector);

    for (int i = 0; i<numElements; i++)
        VectorSet(probabilityVector, i, (double)(probability[i]));
    if (!PublishNode(postUNL, updateNode, pubInfo))
        printf("Classification.cpp: Error publishing node\n");
    if (updateNode != NULL) FreeNode(updateNode); // this also seems to free probabilityVector!
}

void Classification::show() {
    int i;

    printf("Classification::show()\n");
    printf("  numElements: %d\n", numElements);

    for(i=0; i<numElements; i++) {
        printf("  %s : %.4f\n", labels[i], probability[i]);
    }
}

//--------------------------------------------------

void Classification::normalize() {
    int i;
    double sum;

    sum = 0;
    for(i=0; i<numElements; i++)
        sum += probability[i];

    if(sum > 0) {
        for(i=0; i<numElements; i++)
            probability[i] /= sum;
    }
    else {
                // printf("Empty Classification\n");
        for(i=0; i<numElements; i++)
            probability[i] = MIN_PROBABILITY;
    }
}

void Classification::zero() {
    int i;

    for(i=0; i<numElements; i++)
        probability[i] = 0.0;
}

void Classification::empty() {
    int i;
    if(numElements > 0) {
        for(i=0; i<numElements; i++) {
            free(labels[i]);
        }
        delete[] labels;
        delete[] probability;
    }
    labels = NULL;
    probability = NULL;
}

//--------------------------------------------------

// void Classification::load(NodeType *node) {
//     IterType *label_set;
//     NodeType *labelNode = NULL;
// 
//     label_set = NewBranchIter(node);
// 
//     zero();
//     while( (labelNode=NextNode(label_set)) != NULL ) {
//         if (IsType(labelNode->type,T_FLOAT)) {
// 			// printf("%s %.4f\n",labelNode->name, labelNode->data.float_val);
//             addToLabel(labelNode->name, labelNode->data.float_val);
//         }
// 
//         DeleteNode(NULL, labelNode);
//     };
//     free(label_set);
// }

//--------------------------------------------------

char* Classification::mostLikely() {
    int i;
    double best = 0.0;
    char *ptr = "None";

    for(i=0; i<numElements; i++) {
        if(probability[i] > best) {
            best = probability[i];
            ptr = labels[i];
        }
    }

    return ptr;
}

int Classification::mostLikelyIndex() {
    int i;
    double best = 0.0;
    int index = -1;

    for(i=0; i<numElements; i++) {
        if(probability[i] > best) {
            best = probability[i];
            index = i;
        }
    }

    return index;
}

int Classification::meetsThreshold(char *label, double threshold) {
    int id = getIndex(label);

    if(id == -1)
        return 0;

    return (probability[id] >= threshold);
}

