/*****************************************************************************
 *
 * Author: Mark Blum, Wen Dong, Jonathan Gips, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <pthread.h>

extern "C" {
#include <enchantment/unl.h>
#include <enchantment/signals.h>
#include <enchantment/whiteboard.h>
#include <enchantment/vector.h>
#include <enchantment/darray.h>
}

#include "Classification.hpp"
#include "InfluenceModel.hpp"

#define USE_THREADS

//Static so we can flush it in SigHandler()
FILE** outFile_fids;
int numChains = 1;
bool fileOutput;
volatile int runflag = 1;
volatile int killflag = 0;


int timeCompare(struct timeval *t1, struct timeval *t2);

//-----------------------------------------------------------
DArray chainSigs = NULL;

//Holds data pertaining to one input signal
//that has a thread assigned to it
typedef struct ChainSignal 
{
  int chainIdx;
  char *inBuf;
  struct timeval time;
  NodeType *sigNode;
  char *sigUNL;
  Signal *signal;
  bool bRunning;
  bool bDataAvailable;
  pthread_cond_t dataSem;
  pthread_mutex_t dataSemMutex;
  pthread_t thread;
  bool bDestroy;
} ChainSignal;


void StartChainSignalThread(NodeType *sigNode, int chainIdx);
void *doRunChainSignal(void *param);
ChainSignal *GetNextChainSignalObservation();
void AdvanceChainSignal(ChainSignal *cs);



void StartChainSignalThread(NodeType *sigNode, int chainIdx)
{
  int idxToUse(-1);
  
  //set any old signals pertaining to new chain index for
  //destruction
  for(int i = 0; i < ArrayLength(chainSigs); i++) {
    ChainSignal *cs = *(ChainSignal**)ArrayNth(chainSigs, i);

    if(cs->chainIdx == chainIdx) {
      cs->bDestroy = true;
    }
    if(idxToUse == -1 && cs->bRunning == false) {
      idxToUse = i;
    }
  }
  
  //initialize a new record
  ChainSignal *newCS = (ChainSignal*)malloc(sizeof(ChainSignal));
  
  newCS->chainIdx = chainIdx;
  newCS->inBuf = NULL;
  newCS->sigNode = sigNode;
  newCS->signal = NULL;
  newCS->bDataAvailable = false;
  newCS->bRunning = true;
  newCS->bDestroy = false;
  pthread_cond_init(&newCS->dataSem, NULL);
  pthread_mutex_init(&newCS->dataSemMutex, NULL);
  
  //add new record into array
  if(idxToUse != -1) {
    ArrayReplaceAt(chainSigs, &newCS, idxToUse);
  } else {
    ArrayAppend(chainSigs, &newCS);
    idxToUse = ArrayLength(chainSigs) - 1;
  } 

  ChainSignal *arrCS = *(ChainSignal**)ArrayNth(chainSigs, idxToUse);

  //create thread and set the threadID in the actual array memory
  pthread_create(&arrCS->thread, NULL, doRunChainSignal, arrCS);
  
}

void ChainSignalDestroy(void *param)
{
  ChainSignal *cs = *(ChainSignal**)param;

  if(cs->sigNode) FreeNode(cs->sigNode);
  cs->sigNode = NULL;
  
  if(cs->inBuf) free(cs->inBuf);
  cs->inBuf = NULL;
  
  if(cs->signal) SignalClose(cs->signal);
  cs->signal = NULL;

  pthread_cond_destroy(&cs->dataSem);
  pthread_mutex_destroy(&cs->dataSemMutex);

  free(cs); //free the allocated ChainSignal
}


void *doRunChainSignal(void *param)
{
  ChainSignal *cs = (ChainSignal*)param;
  
  //Set thread to detached so that it will clean itself up
  //and we don't have to call join
  pthread_detach(pthread_self());
  
  cs->signal = NewSignalNode(cs->sigNode);
			     
  while (!cs->bDestroy)
    {
      if (!SignalConnected(cs->signal))
        {
	  if (!SignalConnectWait(cs->signal))
	    {
	      printf("IClassifier:: Failed to connect to signal for chain %d\n", cs->chainIdx);
	      cs->bRunning = false;
	      pthread_exit(NULL);
	    }
          if (SignalConnected(cs->signal))
	    {
	      cs->inBuf = (char*)malloc(SignalBufferSize(cs->signal));
	    }
        }
      else
        {
	  if (SignalRecv(cs->signal, &cs->time, cs->inBuf))
            {
	      if(!cs->bDestroy)
		{
		  //Set the flag for new data and then wait for semaphore
		  //to be signaled before continuing.
		  cs->bDataAvailable = true;
		  pthread_cond_wait(&cs->dataSem,&cs->dataSemMutex);
		}
	    }
        }
    }   

  cs->bRunning = false;
  pthread_exit(NULL);
}


ChainSignal *GetNextChainSignalObservation()
{
  ChainSignal *oldestCS(NULL);  
  bool hasChainData[numChains];
  int numWithData(0);

  //we require all chains to have new data before returning data
  for(int i = 0; i < numChains; i++) hasChainData[i] = false;

  for(int i = 0; i < ArrayLength(chainSigs); i++) {
    ChainSignal *cs = *(ChainSignal**)ArrayNth(chainSigs, i);
    
    if(cs->bRunning && cs->bDataAvailable) {
      
      if(!hasChainData[cs->chainIdx]) numWithData++;
      hasChainData[cs->chainIdx] = true;
      
      if(oldestCS == NULL) {
	oldestCS = cs;
      } else if(timeCompare(&oldestCS->time, &cs->time) > 0) {
	oldestCS = cs;
      }
    }
  }
  
  //don't return unless all chains have data
  if(numWithData < numChains) return NULL;

  return oldestCS;
}

void AdvanceChainSignal(ChainSignal *cs)
{
  cs->bDataAvailable = false;
  pthread_cond_signal(&cs->dataSem);
}

//---------------------------------------------------------------------------------
void PrintUsage()
{
    printf("Usage: IClassifier -m [model.xml] [OPTION]...\n");
    printf("                                                  \n");
    printf("       -m [FILE]     XML file representing the Influence Model \n");
    printf("       -v            verbose, write output to stdout \n");
    printf("\n");
    printf("       For multiple chains, the input and output UNLs must be contained\n");
    printf("       in the XML file. Optionally a string may be prepended as follows: \n");
    printf("       -i [baseUNL]      prepend to [baseUNL] to all input UNLs \n");
    printf("       -o [baseUNL]      prepend to [baseUNL] to all output UNLs \n");
    printf("       -f                treat all UNLs as files \n");
    printf("       -g                treat output UNLs as files \n");
    printf("       -t [size]         windowSize \n");
    printf("\n");
    // not sure if we should deal with this case in this program, or if we should crate a new one called HMMClassifier
    printf("       If there is only one chain, the input and output may be completely\n");
    printf("       specified as follows:\n");
    printf("       -i [UNL]      read input signal from [UNL] \n");
    printf("       -o [UNL]      write output signal to [UNL] \n");
    printf("       -p [UNL]      post output [UNL] on whiteboard\n");
    printf("       -r [FILE]     read input from [FILE] \n");
    printf("       -w [FILE]     write output to [FILE] \n");
    printf("\n");
    printf("       Example: IClassifier -m model.xml -i ./input o ./output -f\n");
    printf("       will run a the model with multiple chains entirely file-based\n");
    printf("\n");
}

void SigHandler(int sig) {
    switch (sig) {
        case SIGINT:
            runflag=0;
            if(killflag)
            {
                fflush(stdout);
                if (fileOutput){
                    for (int i=0; i<numChains; i++){
                        fflush(outFile_fids[i]);
                    }
                }
                fprintf(stderr,"Hard kill, flushing output streams and exiting.\n");
                exit(0);
            }
            else
            {
                killflag = 1;
            }
            
            break;
            default:
                printf("Signal handler received %d\n",sig);
    }
    return;
}

int timeCompare(struct timeval *t1, struct timeval *t2)
{
    if (t1->tv_sec > t2->tv_sec){
        return 1;
    }
    else if (t1->tv_sec < t2->tv_sec) {
        return -1;
    }
    else if (t1->tv_usec > t2->tv_usec) {
        return 1;
    }
    else if (t1->tv_usec < t2->tv_usec) {
        return -1;
    }
    else
        return 0; // timestamps are equal;
}

int usec_difference(struct timeval *t1, struct timeval *t2)
{
    return (t1->tv_sec - t2->tv_sec) * 1000000 + (t1->tv_usec - t2->tv_usec);
}


int main(int argc, char *argv[]) {
    //FILE* inFile_fids;   // these 3 are declared further down
    //double** inBuffers;
    //double** outBuffers;
    //FILE* outFile_fids;  // these 2 are declared globally
    //bool fileOutput;
    char *inUNL = NULL; // (ex. /signals/feat)
    char *outUNL = NULL;   // (ex. /classifiers/posture)
    char *postUNL = NULL;   // (ex. /classifiers/posture)
    char *inFile = NULL;
    char *outFile = NULL;
    char *modelFile = NULL;
    char textBuffer[256];
    int verbose = 0;
    int windowSize = 1;
   
    //int numFeatures, numClasses;
    bool fileInput = false;
    bool newObsAvailable;
        
    //Signal *inSignal;
    //Signal *outSignal;
    struct timeval outTime;
    Vector *feature = NULL;
    Classification *classifications;
    int newObsIndex;
    int i,k;
    
    signal(SIGINT,SigHandler);  //die nicely
    
    ///// Parse input arguments ///////////////////////////////////////////
    
    i = 1;
    while ( i < argc ) {
        if (argv[i][0] == '-' ) {
            switch ( argv[i][1] ) {  // options without arguments
                case 'v': {
                    verbose = true;
                    break;
                }
                case 'f': {
                    fileInput = true;
                    fileOutput = true;
                    break;
                }
                case 'g': {
                    fileOutput = true;
                    break;
                }
                default: {
                    i++;
                    if ( i == argc ) // check if argument follows
                    {
                        PrintUsage();
                        exit(1);
                    }
                    switch ( argv[i-1][1] ) { // options with arguments
                        case 'i': {
                            inUNL = argv[i];
                            break;
                        }
                        case 'o': {
                            outUNL = argv[i];
                            break;
                        }
                        case 'p': {
                            postUNL = argv[i];
                            break;
                        }
                        case 'r': {
                            inFile = argv[i];
                            break;
                        }
                        case 'w': {
                            outFile = argv[i];
                            break;
                        }
                        case 'm': {
                            modelFile = argv[i];
                            break;
                        }
                        case 't': {
                            windowSize = atoi(argv[i]);
                            break;
                        }
                        default: {
                            printf("Unknown option -%c\n\n",argv[i-1][1]);
                            PrintUsage(); 
                            exit(1);
                        }
                    } // end options with arguments
                }
            } //end switch ( argv[i][1]
            i++;
        }
        else  // no '-' detected
        {
            PrintUsage();
            exit(1);
        }
    }  ///// end parsing input arguments
    
    
    
    if (modelFile==NULL){
        printf("IClassifier: Error, no model file specified!\n\n");
        PrintUsage();
        exit(1);
    }
    
    printf("IClassifier: Creating and loading influence model\n");
    InfluenceModel influenceModel(windowSize);
    if (influenceModel.loadXML(modelFile) != 0){
        printf("IClassifier: Error loading influence Model\n");
        exit(1);
    }
    if (verbose) influenceModel.show();
    
    printf("IClassifier: Allocating memory for I/O\n");
    numChains = influenceModel.getNumChains();
    classifications = new Classification[numChains];
    FILE *inFile_fids[numChains];
    bool endOfFile[numChains];
    bool newObsSet[numChains];
    bool dataPending[numChains];
    struct timeval time[numChains];
    if (fileOutput) outFile_fids = new (FILE*)[numChains];
    double* inBuffers[numChains];
    double* outBuffers[numChains];
    double* newObservations[numChains];
    
    //Do not touch this directly, use functions at top of file
    chainSigs = ArrayNew(sizeof(ChainSignal*), numChains, ChainSignalDestroy);
    
    /*if (numChains == 1) {
        if (inUNL==NULL && inFile==NULL){
            printf("IClassifier: Error, no input specified!\n\n");
            PrintUsage();
            exit(1);
        }
        if (inUNL!=NULL) {
            inSignal = NewSignalUNL(inUNL);
            if(inSignal == NULL) {
                printf("IClassifier: Error allocating signal %s\n",inUNL);
                exit(1);
            }
        }
        else
        {
            inFile_fid = fopen(inFile,"r");
            if (inFile_fid == NULL) {
                printf("IClassifier: Error, could not open file %s\n", inFile);
                exit(1);
            }
        }
        numFeatures = influenceModel.getDim();
        if (inSignal) {
            if (!SignalConnectWait(inSignal)){
                printf("IClassifier: Error connecting to %s\n",inUNL);
                exit(1);
            }
            if (SignalGetLength(inSignal, 1) != numFeatures)
            {
                printf("IClassifier: Error, numFeatures (%d) does not match length of input signal (%d)\n",
                numFeatures, SignalGetLength(inSignal,1));
                exit(1);
            }
        }
        inBuffer = (float *) malloc(numFeatures * sizeof(float));
        feature = NewVector(numFeatures);
     
        numClasses = influenceModel.getCount();
        outBuffer = (double *) malloc(numClasses * sizeof(double));
     
        printf("%d classes, %d feature vector\n",numClasses,numFeatures);
     
        if (outUNL != NULL) {
            outSignal = NewSignalGenUNL(outUNL);
            if(outSignal == NULL) {
                printf("IClassifier: Error allocating outSignal %s\n",outUNL);
                exit(1);
            }
            SignalSetType(outSignal, T_DOUBLE);
            SignalSetDimension(outSignal, 1, numClasses, NULL, NULL);
        }
        if (outFile != NULL) {
            outFile_fid = fopen(outFile,"w");
            if (outFile_fid == NULL) {
                printf("IClassifier: Error, could not create file file %s\n", outFile);
                exit(1);
            }
        }
     
     
    }
    else  // multiple chains */
    {
        for (i=0; i<numChains; i++){
            sprintf(textBuffer,"%s%s",((inUNL == NULL)?"":inUNL),influenceModel.getChain(i)->getObservationUNL());
            if (fileInput) {
                influenceModel.setObservationUNL(i, textBuffer);
                inFile_fids[i] = fopen(textBuffer,"r");
                if (inFile_fids[i] == NULL) {
                    printf("IClassifier: Error, could not open input file '%s' for chain %d\n",textBuffer,i);
                    exit(1);
                }
            }
            else // input from whiteboard
            {
                if (ExtractUNLServer(influenceModel.getChain(i)->getObservationUNL()) == NULL) {
                    strcpy(textBuffer,CreateUNL("localhost", influenceModel.getChain(i)->getObservationUNL()));
                }
                influenceModel.setObservationUNL(i, textBuffer);
                printf("IClassifier: Subscribing to %s\n",textBuffer);
                SubscribeUNLInfo info;
                SubscribeUNL(NULL, &info);
                if (!SubscribeUNL(textBuffer, &info))
                {
                    fprintf(stderr, "IClassifier: Error subscribing to %s\n", textBuffer);
                    exit(1);
                }
            }
            sprintf(textBuffer,"%s%s",((outUNL == NULL)?"":outUNL),influenceModel.getChain(i)->getStateUNL());
            if (fileOutput) {
                influenceModel.setStateUNL(i, textBuffer);
                //extract path and create directory
                char temp[strlen(textBuffer)+10];
                sprintf(temp,"mkdir -p ");
                int n = (int)(strrchr(textBuffer,'/')-textBuffer);
                strncpy (temp+9,textBuffer,n); 
                temp[n+9]='\0';
                system(temp); 
                outFile_fids[i] = fopen(textBuffer,"w");
                if (outFile_fids[i] == NULL) {
                    printf("IClassifier: Error, could not open output file '%s' for chain %d\n",textBuffer,i);
                    exit(1);
                }
            }
            else // input to whiteboard
            {
                if (ExtractUNLServer(influenceModel.getChain(i)->getStateUNL()) == NULL) {
                    strcpy(textBuffer,CreateUNL("localhost", influenceModel.getChain(i)->getStateUNL()));
                }
            }
            inBuffers[i] = (double *) malloc(influenceModel.getChain(i)->getDimObservations() * sizeof(double));
            outBuffers[i] = (double *) malloc(influenceModel.getChain(i)->getNumStates() * sizeof(double));
            newObservations[i] = (double *) malloc(influenceModel.getChain(i)->getDimObservations() * sizeof(double));
            // Fill with NaN because we cant expect to have all observations for the first classification, 
            // but do want to (try and) classify with what ever observation we have got.
            for (k=0; k < influenceModel.getChain(i)->getDimObservations(); k++) newObservations[i][k] = influenceModel.NaN; 
            endOfFile[i] = false;
            dataPending[i] = false;
            newObsSet[i] = false;
        }
        
    }
    
    
    printf("IClassifier: Reading Signals -----------\n");
    int count = 1;
    while(runflag) {
        /*if (inSignal != NULL) {
            if (SignalRecv(inSignal, &time, inBuffer) != 1) {
                continue;  //SignalRecv did not recieve new data, so do not compute classification
            }
        }
        else // (inFile_fid != NULL)
        {
            if (fscanf(inFile_fid,"%ld.%ld",&time.tv_sec,&time.tv_usec) == EOF) {
                printf("GClassifier: Reached end of input file\n");
                break;  //this is the normal way to terminate since we end after a newLine
            }
            else {
                for (i = 0; i<numFeatures; i++) {
                    if (fscanf(inFile_fid,"%f",&inBuffer[i]) == EOF) {
                        printf("GClassifier: Warning, unfinished feature vector at end of input file\n");
                        endOfFile = 1;
                        break;  //no runtime error, but the last classification will not be correct
                    }
                }
            }
            if (endOfFile) break;
            //fgets(inBuffer,numFeatures * sizeof(float),inFile_fid);
        }
        for (i = 0; i<numFeatures; i++)
            VectorSet(feature, i,(double)(inBuffer[i]));
        //printf("received new feature Vector: ");
        //PrintVector(stdout,feature); */
        
	newObsAvailable = false;
        ChainSignal *nextCS = NULL;
	
	if (fileInput) {
            for (i=0; i<numChains; i++){
                int dimObs = influenceModel.getChain(i)->getDimObservations();
                if (!endOfFile[i] && !dataPending[i]){
                    if (fscanf(inFile_fids[i],"%ld.%ld",&((time[i]).tv_sec),&((time[i]).tv_usec)) == EOF) {
                        printf("IClassifier: Reached end of input file for chain %d\n",i);
                        endOfFile[i] = true;
                        continue;  //this is the normal way to terminate since we end after a newLine
                    }
                    else {
                        for (k = 0; k < dimObs; k++) {
                            if (fscanf(inFile_fids[i],"%lf",&(inBuffers[i][k])) == EOF) {
                                printf("IClassifier: Warning, unfinished feature vector at end of input file for chain %d\n",i);
                                endOfFile[i] = true;
                                break;  //no runtime error, but the last classification will not be correct
                            }
                        }
                    }
                    dataPending[i] = true;
                }
                newObsAvailable = newObsAvailable || dataPending[i]; // is there any data left at all?
            }
            
            if (!newObsAvailable) break; // exit while(runflag) since all files have been read
            outTime.tv_sec = 2147483647;  //should be LONG_MAX
            // find the observation among all new ones, which is the oldest
            for (i=0; i<numChains; i++){
                if (!endOfFile[i]) if (timeCompare(&(time[i]),&outTime) != 1){  // if (time[i] <= outTime)
                    outTime = time[i];
                    newObsIndex = i;
                }
            }
            memcpy(newObservations[newObsIndex],inBuffers[newObsIndex],sizeof(double)*influenceModel.getChain(newObsIndex)->getDimObservations());
            dataPending[newObsIndex] = false;
            newObsSet[newObsIndex] = true;
        }
        else // input from Whiteboard
        {
            //printf("waiting for update\n");
            GetUpdateInfo upInfo;
            GetUpdate(&upInfo, NULL);
            if (upInfo.node == NULL) {
                //printf("upInfo.node == NULL\n");
            }
            else {
                if (upInfo.node->type == T_VECTOR){
                    //printf("received vector from %s:\n",upInfo.UNL);
                    if (verbose) PrintVector(stdout,upInfo.node->data.vector_val);
                    for (i=0; i<numChains; i++){
                        if (strcmp(upInfo.UNL,influenceModel.getChain(i)->getObservationUNL()) == 0) {
                            memcpy(newObservations[i],upInfo.node->data.vector_val->values,
                            sizeof(double)*influenceModel.getChain(i)->getDimObservations());
                            //newObsSet[i] = true;
                            newObsAvailable = true;
                        }
		    }
		    
		    FreeNode(upInfo.node);
                }
		else if(upInfo.node->type == T_SIGNAL) {
		  for (i=0; i<numChains; i++){
		    if (strcmp(upInfo.UNL,influenceModel.getChain(i)->getObservationUNL()) == 0) {
		      StartChainSignalThread(upInfo.node, i);
		      break;
		    }
		  }
		}
                else
                {
                    printf("IClassifier: Warning, node is not a vector\n");
		    FreeNode(upInfo.node);
                }  
            }
	    
	    //only look at signal data if we didn't get whiteboard data
	    if(!newObsAvailable) {
	      nextCS = GetNextChainSignalObservation();
	      
	      if(nextCS != NULL) {
		TypeID sigType = SignalGetType(nextCS->signal);
		char *endPtr = nextCS->inBuf + SignalBufferSize(nextCS->signal);
		
		char *ptr = nextCS->inBuf;
		int i(0);
		
		printf("next: chain %d, time %d:%d\n", nextCS->chainIdx, 
		       nextCS->time.tv_sec, nextCS->time.tv_usec);


		while(ptr <= endPtr) {
		  switch(sigType) {
		  case T_INT32:
		    newObservations[nextCS->chainIdx][i] = *(int*)ptr;
		    ptr += sizeof(int);
		    break;
		  case T_INT16:
		    newObservations[nextCS->chainIdx][i] = *(short*)ptr;
		    ptr += sizeof(short);
		    break;
		  case T_INT8:
		    newObservations[nextCS->chainIdx][i] = *(unsigned char*)ptr;
		    ptr += sizeof(unsigned char *);
		    break;
		  case T_FLOAT:
		    newObservations[nextCS->chainIdx][i] = *(float*)ptr;
		    ptr += sizeof(float);
		    break;
		  case T_DOUBLE:
		    newObservations[nextCS->chainIdx][i] = *(double*)ptr;
		    ptr += sizeof(double);
		    break;
		  default:
		    printf("IClassifier: Unhandled signal type for chain %d.\n", nextCS->chainIdx);
		    exit(-1);
		  }
		  i++;
		}
		
		outTime = nextCS->time;
		
		//let the thread continue
		AdvanceChainSignal(nextCS);

		newObsAvailable = true;
	      }
	    }
	}
	
	
        /*bool ready = true;
        for (i=0; i<numChains; i++){
	ready = ready && newObsSet[i];
        }   
        if (!ready) continue;*/
        
	if (!newObsAvailable) {
	  usleep(10000);
	  continue;
        }

        if (verbose){
            printf("newObservations[%d]:\n",count++);
            for (i=0; i<numChains; i++){
                for (k = 0; k < influenceModel.getChain(i)->getDimObservations(); k++) {
                    printf("%-30.30g ",newObservations[i][k]);
                }
                printf("\n");
            }
        }

        influenceModel.classify(classifications, newObservations);
        
        if (verbose){
            for (i=0; i<numChains; i++){
                classifications[i].show();
            }
        }
        for (i=0; i<numChains; i++){
            classifications[i].getProbabilities(outBuffers[i]);
        }            
        if (fileOutput){
            for (i=0; i<numChains; i++){
                if (!fileInput && !nextCS) gettimeofday(&outTime, NULL);
                    fprintf(outFile_fids[i],"%06ld.%06ld",outTime.tv_sec,outTime.tv_usec);
                for (k = 0; k < influenceModel.getChain(i)->getNumStates(); k++) {
                    fprintf(outFile_fids[i]," %-30.30g",outBuffers[i][k]);
                }
                fprintf(outFile_fids[i],"\n");
            }
        }
        /*
        classification->getProbabilities(&outBuffer[0]);
        //organize resulting probabilities in float *outBuffer
        if (postUNL != NULL) classification->publish(postUNL);
        
        if (outFile_fid != NULL) {
            fprintf(outFile_fid,"%06ld.%06ld", time.tv_sec, time.tv_usec);
            for (i=0; i<numClasses; i++) {
                fprintf(outFile_fid," %g",outBuffer[i]);
            }
            fprintf(outFile_fid," %s\n",classification->mostLikely());
        }
        
        if (verbose){
            for (i=0; i<numClasses; i++) {
                printf("%g ",outBuffer[i]);
            }
            printf(" %s\n",classification->mostLikely());
        }
        
        // Important: SignalSend is placed at the end because it overwrites outBuffer!
        if (outSignal != NULL) {
            SignalGenAccept(outSignal);
            SignalSend(outSignal, &time, outBuffer);
        }
        */
        
    }
    
    if (fileOutput){
        for (i=0; i<numChains; i++){
            fflush(outFile_fids[i]);
        }
    }
    if (feature != NULL) FreeVector(feature);
    delete[] classifications;
    if (fileOutput) delete[] outFile_fids;
    for (i=0; i<numChains; i++){
        free(inBuffers[i]);
        free(outBuffers[i]);
    }
    printf("Exiting nicely\n");
    return 0;
}
