/*****************************************************************************
*
* Author: Jonathan Gips, MIT Wearable Computing Group
* http://www.media.mit.edu/wearables/
* Copyright (C) 2005 MIT Media Lab
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*****************************************************************************
* The class models the influence model, used by IClassifier. It inludes XML 
* parsing.
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>

extern "C" {
#include <enchantment/vector.h>
#include <enchantment/matrix.h>
}

#include "Classification.hpp"
#include "InfluenceModel.hpp"

//double InfluenceModel::NAN = 0xFFF80000;
int InfluenceModel::DEFAULT_HD_T = 30;
int InfluenceModel::HD_VIEW_MULT = 4;

InfluenceModel::InfluenceModel(int T) :
  A(NULL), D(NULL), versionXML(NULL), numChains(0), chains(NULL), 
  AiXML(NULL), DXML(NULL), c1XML(-1), c2XML(-1), HXML(NULL), 
  chainLoadIdx(0), hdSampView(NULL), hdInf(NULL), NaN(0)
{
  long hexNaN[2] = {0xFFF80000, 0x00000000 };
  memcpy(&this->NaN, hexNaN, sizeof(double));

  if(T >= 0) 
    hdT = T;
  else 
    hdT = DEFAULT_HD_T;
}

InfluenceModel::~InfluenceModel()
{
  clear();
}

void InfluenceModel::show() {

  printf("InfluenceModel:\n");
  printf("Version: %s\n", versionXML);
  printf("NumChains: %d\n", numChains);
  for(int i = 0; i < numChains; i++) {
    chains[i].show();
  }
  for(int i = 0; i <numChains; i++) {
    for(int j = 0; j < numChains; j++) {
      printf("A: (%d, %d)\n", i,j);
      PrintMatrix(stdout, getA(i,j));
    }
  } 
  printf("D:\n");
  PrintMatrix(stdout, D);
}

void InfluenceModel::clear(void)
{
  //do not clear dimension as it is only passed in constructor!
  if(versionXML != NULL) free(versionXML);
  if(chains != NULL) delete[] chains;
  if(AiXML != NULL) delete AiXML;
  if(DXML != NULL) delete DXML;
  if(HXML != NULL) delete HXML;
  if(A != NULL) {
    for(int i = 0; i < numChains; i++)
      if(A[i] != NULL) {
	FreeMatrix(A[i]);
      }
  }
  
  if(D != NULL) FreeMatrix(D);
  if(hdInf != NULL) hd_influence_free(hdInf);
  if(hdSampView != NULL) hd_sample_view_free(hdSampView);

  AiXML = NULL;
  DXML = NULL;
  HXML = NULL;
  A = NULL;
  D = NULL;
  hdInf = NULL;
  hdSampView = NULL;

  versionXML = NULL;
  chains = NULL;
  c1XML = -1;
  c2XML = -1;

  freeModel();
}

Matrix *InfluenceModel::getA(int c1, int c2)
{
  if(A==NULL || c1>numChains || c2>numChains)
    return NULL;
  else return A[c1*numChains + c2]; 
}

bool InfluenceModel::setD(Matrix *d)
{
  if(d == NULL) return false;
  
  if(D != NULL) {
    FreeMatrix(D);
  } 
  
  D = CopyMatrix(d);

  return true;
}

bool InfluenceModel::setA(Matrix *a, int c1, int c2) 
{
  c1--;
  c2--; //starts at 0
  
  if(a == NULL || c1>=numChains || c2>=numChains) {
    return false;
  } 
  
  if(A == NULL) {
    A = (Matrix **)malloc(sizeof(Matrix*)*numChains*numChains);
    memset(A,0,sizeof(Matrix*)*numChains*numChains);
  }
  
  A[c1*numChains + c2] = CopyMatrix(a);
 
  return true;
}

bool InfluenceModel::setStateUNL(int c, const char *s)
{
    if(c < 0 || c >= numChains) return false;
    return chains[c].setStateUNL(s);
}

bool InfluenceModel::setObservationUNL(int c, const char *s)
{
    if(c < 0 || c >= numChains) return false;
    return chains[c].setObservationUNL(s);
}


const Chain* InfluenceModel::getChain(int c) 
{
  if(c < 0 || c >= numChains) return NULL;
  
  return &(chains[c]);
}

//--------------------------------------------------------------------

void InfluenceModel::startXMLElement(const char *name, const char **atts) 
{
  if(!strcasecmp(name, "influencemodel")) {
    xml_state = INFLUENCEMODEL;
    clear();
    chainLoadIdx = 0;
  }
  else if(!strcasecmp(name, "version")) {
    xml_state = VERSION;
  } 
  else if(!strcasecmp(name, "numchains")) {
    xml_state = NUMCHAINS;
  } 
  else if(!strcasecmp(name, "chain")) {
    xml_state = CHAIN;
    pushXMLObject(&chains[chainLoadIdx++], name, atts);
  } 
  else if(!strcasecmp(name, "d")) {
    xml_state = D_MATRIX;
    if(chainLoadIdx != numChains) {
      assert(0);
    }
    DXML = new XMLMatrix(numChains, numChains);
    pushXMLObject(DXML, name, atts);
  }
  else if(!strcasecmp(name, "h")) {
    xml_state = H_MATRIX;
    if(chainLoadIdx != numChains) {
      assert(0);
    }
    HXML = new XMLMatrix();
    pushXMLObject(HXML, name, atts);
  }
  else if(!strcasecmp(name, "a")) {
    xml_state = A_MATRIX;
  }
  else if(!strcasecmp(name, "a_sub")) {
    xml_state = A_SUB;
    if(chainLoadIdx != numChains) {
      assert(0);
    }
  }
  else if(!strcasecmp(name, "c1")) {
    xml_state = C1;
  }
  else if(!strcasecmp(name, "c2")) {
    xml_state = C2;
  }
  else if(!strcasecmp(name, "transitionmatrix")) {
    xml_state = TRANS_MAT;
    if(AiXML != NULL) delete AiXML;
    AiXML = new XMLMatrix();
    pushXMLObject(AiXML, name, atts);
  }
  else {
    printf("Got unknown start tag \"%s\".\n", name);
    assert(0);
  }
}

void InfluenceModel::processXMLCharacterData(const char *s, int len)
{
  switch(xml_state) {
  case VERSION:
    versionXML = MyStrndup(s,len);
    break;
  case NUMCHAINS:
    numChains = atoi(s);
    chains = new Chain[numChains]();
    break;
  case C1:
    c1XML = atoi(s);
    break;
  case C2:
    c2XML = atoi(s);
    break;
  default:
    break;
  }
  xml_state = SKIP; //only one parsing necessary per state
} 

void InfluenceModel::endXMLElement(const char *name)
{
  if(!strcasecmp(name, "influencemodel")) {
    setD(DXML->m);
    delete DXML;
    DXML = NULL;
    popXMLObject();
  }
  else if(!strcasecmp(name, "a_sub")) {
    setA(AiXML->m, c1XML, c2XML);
    delete AiXML;
    AiXML = NULL;
  }
}

int InfluenceModel::loadXML(char *file)
{
  if(XMLObject::loadXML(file) == 1) return 1;

  //need to invert return value...fun gnu return value (0==success)
  return !initModel();
}

int InfluenceModel::initModel()
{
  double num_states[numChains];
  double num_outputs[numChains];
  double o_type[numChains];

  freeModel();

  //initialize the structure of the influence model
  for(int i = 0; i < numChains; i++) {
    num_states[i] = chains[i].getNumStates();
    num_outputs[i] = chains[i].getDimObservations();
    o_type[i] = (chains[i].getObservationType() == Chain::DISCRETE) ? 'd' : 'c';
  }
  
  hdInf = hd_influence_init(numChains, num_states, num_outputs, o_type);
  
  int numPriors(0);
  for(int i = 0; i < numChains; i++) {
    numPriors += chains[i].getNumStates();
  }
  
  double priors[numPriors];
  int numCopied(0);
  
  //initialize each chain
  for(int i = 0; i < numChains; i++) {
    //copy priors
    memcpy(priors + numCopied, chains[i].getPriors()->values, 
	   sizeof(double)*chains[i].getNumStates());
    numCopied += chains[i].getNumStates();
    
    //initialize the observations
    if(chains[i].getObservationType() == Chain::DISCRETE) {
      Matrix *b = chains[i].getDiscreteB();
      double trans[b->rows*b->cols];
      
      //Matrix struct is column major, need to transpose
      assert(b->values);
      XMLMatrix::GetTranspose(b, trans);
      hd_influence_set_B(hdInf, i, trans); 
    } else {
      double bvals[chains[i].getNumStates()][chains[i].getDimObservations()];
      
      for(int s = 0; s < chains[i].getNumStates(); s++) {
	Gaussian *b = chains[i].getContinuousB(s);
	
	//Copy mean
	memcpy(bvals[s], b->u->values, sizeof(double)*chains[i].getDimObservations());

	//Matrix struct is column major, need to transpose
	double trans[b->cov->rows*b->cov->cols];
	assert(b->cov->values);
	XMLMatrix::GetTranspose(b->cov,trans);
	hd_influence_set_cov(hdInf, i, s, trans); //orientation of rows?
      }
      hd_influence_set_B(hdInf, i, bvals[0]);
    }

    //initialize A from this chain to all chains
    for(int j = 0; j < numChains; j++) {
      Matrix *a = getA(i,j);
      double trans[a->rows*a->cols];
      
      XMLMatrix::GetTranspose(a,trans);
      hd_influence_set_A(hdInf, i, j, trans);
    }
  }
  
  //set priors
  hd_influence_set_prior(hdInf, priors);

  //initialize alpha values
  double dtrans[D->rows*D->cols];
  XMLMatrix::GetTranspose(D,dtrans);
  hd_influence_set_D(hdInf, dtrans);

  //need to call after As and D are set
  hd_influence_set_H(hdInf);

  //allocate the sliding window 
  hdSampView = hd_sample_view_alloc(hdInf, hdT*HD_VIEW_MULT, hdT);
  
  //get a sample and zero it out
  hd_sample *hdSamp = hd_sample_view_shift(hdSampView, 1);
  resetSample(hdSamp);

  hd_influence_reset_ess(hdInf); 
  
  if(hdInf != NULL) hd_influence_print(stdout, hdInf);

  return (hdInf != NULL && hdSampView != NULL);
}

void InfluenceModel::freeModel()
{
  if(hdInf != NULL) hd_influence_free(hdInf);
  if(hdSampView != NULL) hd_sample_view_free(hdSampView);
  
  hdInf = NULL;
  hdSampView = NULL;
}

void InfluenceModel::resetSample(hd_sample *hdSamp)
{
  for(int c = 0; c < numChains; c++) {
    for(int t = 0; t < hdT; t++) {
      //one hidden state
      hd_sample_set_s(hdSamp, c, t, NaN);
     
      if(chains[c].getObservationType() == Chain::DISCRETE) {
	//one discrete observation
	double chainSamp = NaN;
	hd_sample_set_o(hdSamp, c, t, &chainSamp);
      } else {
	//vector of continuous values
	double chainSamp[chains[c].getDimObservations()];
	for(int z = 0; z < chains[c].getDimObservations(); z++)
	  chainSamp[z] = NaN;
	hd_sample_set_o(hdSamp, c, t, chainSamp);
      }
    }
  }
}

void InfluenceModel::classify(Classification *classification, double **obs)
{
  hd_sample *hdSamp = hd_sample_view_shift(hdSampView, 1);
  int PRINT = 0;

  for(int i = 0; i < numChains; i++) {
    //clear out the state estimations
    for(int t = 0; t < hdT; t++) {
      hd_sample_set_s(hdSamp, i, t, NaN);
    }
    
    //set the new observations
    hd_sample_set_o(hdSamp, i, hdT - 1, obs[i]);
  }
  
  if(PRINT) {
    printf("Sample before enter evidence...\n");
    hd_sample_print(stdout, hdSamp);
  }
  
  //enter the new evidence
  hd_influence_enter_evidence(hdInf, hdSamp);
  
  //find most gamma values for most recent hidden states of 
  //each chain
  hd_influence_find_mpe(hdInf, hdSamp);
  
  if(PRINT) {
    printf("Sample after enter evidence...\n");
    hd_sample_print(stdout, hdSamp);
    hd_sample_print_statistics(stdout, hdSamp);
    printf("------------------------------------\n");
  }
  
  for(int c = 0; c < numChains; c++) {
    double gamma[chains[c].getNumStates()];
    
    hd_sample_get_gamma(hdSamp, c, hdT - 1, gamma);
    
    for(int s = 0; s < chains[c].getNumStates(); s++) {
      if(classification[c].getLabel(s) == NULL) {
	//Set the label the first time
	char buff[256];
	sprintf(buff, "s%d", s);
	classification[c].appendElement(buff, gamma[s]);
      } else {
	classification[c].setProbability(s, gamma[s]);
      }
    }
  }
} 
