/**********************************************************************
** Copyright (C) 2002 Michael 'Mickey' Lauer.  All rights reserved.
**
** This file is part of Opie Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
**********************************************************************
* Modifications by Mark Blum (mark@media.mit.edu) July 2005 MIT Media Lab
*
* This is the driver for the CF-card camera. It was modified to suit
* the SignalRecorder application
**********************************************************************/

#include "CameraIO.h"

//#include <sys/types.h>
//#include <sys/stat.h>
//#include <sys/ioctl.h>
//#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <iostream.h>

#include <qimage.h>
#include <qdatetime.h>

#define SHARPZDC "/dev/sharp_zdc"

#define CAM_DEBUG 0


CameraIO* CameraIO::_instance = 0;

CameraIO* CameraIO::instance()
{
    if ( !CameraIO::_instance )
    {
      if (CAM_DEBUG) cout << "CameraIO: Creating CameraIO::_instance" << endl;
        CameraIO::_instance = new CameraIO();
    }
    return CameraIO::_instance;
}


CameraIO::CameraIO()
    : _pressed( false ), _height( 120 ), _width( 160 ), _zoom( 1 ),
           _flip( -1 ), _rot( 0 ), _readlen( 0 )

{
    _driver = open( SHARPZDC, O_RDWR );
    if ( _driver == -1 ){
      cout << "CameraIO: Error, Can't open camera driver" << endl;
      exit(1);
    }
    else
        init();
}


void CameraIO::init()
{
    if ( CameraIO::_instance )
      cout << "CameraIO: Don't create more than one CameraIO instances." << endl;
    else
    {
        _timer = new QTime();
        setReadMode( STATUS );
    }
}


CameraIO::~CameraIO()
{
    if ( _driver != -1 )
    {
        setReadMode( 0 );
            ::close( _driver );
    }
}


bool CameraIO::isOpen() const
{
    return _driver != -1;
}


bool CameraIO::isShutterPressed()
{
    if ( _status[0] == 'S' )
    {
        if ( !_pressed )    // wasn't pressed before, but is now!
        {
            _pressed = true;
            _timer->start();
            return true;
        }

        if ( _timer->elapsed() > 2000 ) // the press is pretty old now
        {
            clearShutterLatch();
            _status[0] = 's';
            _pressed = false;
        }
    }

    return false;
}


bool CameraIO::isFinderReversed() const
{
    return _status[1] == 'M';
}


bool CameraIO::isCapturing() const
{
    return _status[2] == 'C';
}


bool CameraIO::isAvailable() const
{
    return _status[3] == 'A';
}


bool CameraIO::setCaptureFrame( int width, int height, int zoom, bool rot )
{
  if (CAM_DEBUG) cout << "CameraIO: setCaptureFrame( " << width << ", " << height << ", " << zoom << ", " << rot << " )" << endl;
    char b[100];
    sprintf( b, "%c=%d,%d,%d,%d", rot ? 'R':'S', width, height, zoom, width*2 );
    if ( camWrite( b ) )
    {
        _width = width;
        _height = height;
        _zoom = zoom;
        _rot = rot;
        _readlen = 2 * _width * _height; // camera is fixed @ 16 bits per pixel
        return true;
    }
    cout << "CameraIO: couldn't write to driver!" << endl;
    return false;
}


bool CameraIO::setZoom( int zoom )
{
    return setCaptureFrame( _width, _height, zoom*256, _rot );
}


void CameraIO::setReadMode( int mode )
{
    char b[10];
    sprintf( b, "M=%d", mode );
    camWrite( b, mode <= 9 ? 3 : 4 );
    if ( mode & STATUS ) // STATUS bit is set
    {
        camRead( _status, 4 );
        if ( isShutterPressed() )
        {
      //cout << "emit shutterClicked()" << endl; //emit shutterClicked();
        }
    }
}


void CameraIO::setFlip( int flip )
{
    _flip = flip;
}


void CameraIO::clearShutterLatch()
{
    camWrite( "B", 1 );
}


bool CameraIO::camRead( char* b, int len )
{
#ifndef NO_TIMING
    QTime t;
    t.start();
#endif
    int rlen = read( _driver, b, len );
#ifndef NO_TIMING
    int time = t.elapsed();
#else
    int time = -1;
#endif
    if ( rlen ) {
    if (CAM_DEBUG) { cout << "read " << rlen << " ('" << b[0] << b[1] << b[2] << b[3] << "') [" << time << " ms] from driver." << endl; }
    }
    else
        cout << "CameraIO: error, read nothing from driver." << endl;
    return rlen == len;
}


bool CameraIO::camWrite( char* buf, int len )
{
    if ( !len )
        len = strlen( buf );

    if (CAM_DEBUG) cout << "writing '" << buf << "' to driver." << endl;
    int ret = write( _driver, buf, len );
    return (ret == len);
}


bool CameraIO::snapshot( QImage* image )
{
    setReadMode( STATUS );

    if (CAM_DEBUG) cout << "w=" << _width << " h= " << _height << " readlen= " << _readlen << "rotation = " << _rot << endl;

    int readmode;
    if ( _flip == -1 ) // AUTO
    {
        if ( _rot ) // Portrait
        {
            readmode = IMAGE | isFinderReversed() ? XFLIP | YFLIP : 0;
        }
        else // Landscape
        {
            readmode = IMAGE | XFLIP | YFLIP;
        }
    }
    else // OVERRIDE
    {
        readmode = IMAGE | _flip;
    }

    setReadMode( readmode );

    char buf[_readlen];
    char* bp = buf;
    unsigned char* p;

    //usleep(500000);   // this is the exposure time !!!!!!!!!
    sleep(1);
    camRead( bp, _readlen );

    image->create( _width, _height, 16 );
    for ( int i = 0; i < _height; ++i )
    {
        p = image->scanLine( i );
        for ( int j = 0; j < _width; j++ )
        {
            *p = *bp;
            p++;
            bp++;
            *p = *bp;
            p++;
            bp++;
        }
    }

    return true;
}


void CameraIO::captureFrame( int w, int h, int zoom, QImage* image )
{
    setCaptureFrame( w, h, zoom*256, w>h );
    snapshot( image );
}
