/*************************************************
 * File: SignalRecorder.cpp
 * Author: Mark Blum (mark@media.mit.edu)
 * Birth Date: 05/12/05
 *
 * Project: ASSIST
 * 
 * 
 * Copyright (C) 2005 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ************************************************
 * SignalRecorder GUI-application that takes care of recording data and
 * starting and stopping classifiers. Upon request it will launch
 * required processes with system calls and will kill them on termination.
 * It includes the following intellegent components:
 * - Sensor file size, battery and CPU logging
 * - Managing sensor calibration using MITeCalibrate
 * - Manages picture taking
 * - polls the shutter button if requested
 * - runs the WiFi sniffing procedure
 * - manages online labeling
 * - computes the interest prediction algorithm and logs moments of interst
 ***********************************************/

#include <qlcdnumber.h>
#include <qslider.h>
#include <qcheckbox.h>
#include <qbuttongroup.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qlineedit.h>
#include <qdialog.h>
#include <qspinbox.h>
#include <qtextview.h>
#include <qtabwidget.h>
#include "SignalRecorder.h"
#include <qmessagebox.h>
#include <string.h>
#include <qstring.h>
#include <stdlib.h>
#include <iostream.h>
#include <sys/time.h>

#include "sharp_char.h"

extern "C" {
#include <enchantment/whiteboard.h>
}

#define SR_DEBUG 1
#define BUZZER_DEVICE "/dev/sharp_buz"
#define LED_DEVICE "/dev/sharp_led"

GroupMedia::GroupMedia(QWidget* parent,const char*name, WFlags f): GroupMediaBase(parent,name,f), 
  audioFileSize(0), MITe1FileSize(0), MITe2FileSize(0), MITe3FileSize(0), MITe4FileSize(0)
{
  // Done using designer; 
  //   setFocusPolicy(this->StrongFocus);
  //  setKeyCompression(TRUE);
	
/*	nnn = new QString("SessionName");
    q=new Question(nnn);
  

    q->setGeometry(30,70,180,100);
    connect(q->line,SIGNAL(textChanged(const QString &)),this,SLOT(AcceptName(const QString &)));
    q->exec();
*/	
    int c,k;
    
    connect(startButton,      SIGNAL(clicked()), this, SLOT(StartRecording()));
    connect(stopButton,       SIGNAL(clicked()), this, SLOT(StopRecording()));
    connect(quitButton,       SIGNAL(clicked()), this, SLOT(Quit()));
    connect(serverButton,     SIGNAL(clicked()), this, SLOT(LaunchServer()));
    connect(fileSizeButton,   SIGNAL(clicked()), this, SLOT(CheckAllSignals()));
    connect(fotoButton,       SIGNAL(clicked()), this, SLOT(CaptureImageSlot()));
    connect(posture_ButtonGroup, SIGNAL(clicked(int)), this, SLOT(SelectPosture(int)));
    connect(activity_ButtonGroup,SIGNAL(clicked(int)), this, SLOT(SelectActivity(int)));
    connect(classify_startButton,SIGNAL(clicked()), this, SLOT(StartClassifier()));
    connect(classify_stopButton, SIGNAL(clicked()), this, SLOT(StopClassifier()));
    connect(calibrateButton,     SIGNAL(clicked()), this, SLOT(Calibrate()));
    connect(tempButton,          SIGNAL(clicked()), this, SLOT(PlayBuzzer()));
    connect(threshold_SpinBox,             SIGNAL(valueChanged(int)), this, SLOT(setInterestThreshold(int)));
    stopButton->setEnabled(false);
    calibrateButton->setEnabled(false);
    classify_stopButton->setEnabled(false);
    classify_startButton->setEnabled(false);
    serverButton->setEnabled(false);
    posture_ButtonGroup->setEnabled(false);
    activity_ButtonGroup->setEnabled(false);
    MITe3_checkBox->setEnabled(false);  //temporary !
    MITe4_checkBox->setEnabled(false);  //temporary !
    posture_lineEdit->setText("N/A");
    event_lineEdit->setText("N/A");
    speech_lineEdit->setText("N/A");
    location_lineEdit->setText("N/A");
    posture_c45_lineEdit->setText("N/A");
    event_c45_lineEdit->setText("N/A");
    speech_c45_lineEdit->setText("N/A");
    locationWiFi_lineEdit->setText("N/A");
    
    checkSensorCounter = new QTimer( this );
    connect( checkSensorCounter, SIGNAL(timeout()), this, SLOT(CheckAllSignals()) );
    captureImageCounter = new QTimer( this );
    connect( captureImageCounter, SIGNAL(timeout()), this, SLOT(CaptureImageSlot()) );
    buttonPollCounter = new QTimer( this );
    connect( buttonPollCounter, SIGNAL(timeout()), this, SLOT(PollButton()) );
    showClassificationCounter = new QTimer( this );
    connect( showClassificationCounter, SIGNAL(timeout()), this, SLOT(ShowClassification()) );
    wifiCounter = new QTimer( this );
    connect( wifiCounter, SIGNAL(timeout()), this, SLOT(WifiSniffer()) );
    calibrationCounter = new QTimer( this );
    connect( calibrationCounter, SIGNAL(timeout()), this, SLOT(Calibrate()) );
    
    for (c=0;c<NUM_CHAINS;c++) {
        currentClassification[c] = 0;
        currentSuperState[c] = 0;
        for (k=0;k<MAX_LABELS_PER_CHAIN; k++) {
            interestPoints[c][k] = 0;
            labels[c][k] = NULL;
        }
        for (k=0; k<RING_BUFFER_SIZE; k++) {
            classificationRingBuffer[c][k] = 0;
        }
    }
    
    SetStandbyLED(1);
    system("mkdir -p ./pids");
    system("mkdir -p ./data");
        
}


GroupMedia::~GroupMedia()
{
    SetStandbyLED(0);
    delete checkSensorCounter;
    delete captureImageCounter;
    delete buttonPollCounter;
    delete showClassificationCounter;
    delete wifiCounter;
    printf("SignalRecorder: Deleted counters\n");
    for (int c=0;c<NUM_CHAINS;c++) for (int k=0;k<MAX_LABELS_PER_CHAIN; k++) {
        if (labels[c][k] != NULL) free(labels[c][k]);
    }
    //system("./sigrec_stop.sh &");
    printf("SignalRecorder: freed labels\n");
    KillAllProcesses();
    printf("SignalRecorder: Destructor done\n");
}

void GroupMedia::KillAllProcesses()
{
    char buffer[64];
    sprintf(buffer, "kill `cat ./pids/af.pid 2> /dev/null` 2> /dev/null");           system(buffer);
    sprintf(buffer, "kill `cat ./pids/AudioFeatures.pid 2> /dev/null` 2> /dev/null");system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl.pid 2> /dev/null` 2> /dev/null");        system(buffer);
    sprintf(buffer, "kill `cat ./pids/C45cl.pid 2> /dev/null` 2> /dev/null");        system(buffer);
    sprintf(buffer, "kill `cat ./pids/sv_accel1.pid 2> /dev/null` 2> /dev/null");    system(buffer);
    sprintf(buffer, "kill `cat ./pids/sv_accel2.pid 2> /dev/null` 2> /dev/null");    system(buffer);
    sprintf(buffer, "kill `cat ./pids/sv_accel3.pid 2> /dev/null` 2> /dev/null");    system(buffer);
    sprintf(buffer, "kill `cat ./pids/sv_accel4.pid 2> /dev/null` 2> /dev/null");    system(buffer);    
    sprintf(buffer, "kill `cat ./pids/sv_audio.pid 2> /dev/null` 2> /dev/null");     system(buffer);    
    sprintf(buffer, "kill `cat ./pids/mitesig.pid 2> /dev/null` 2> /dev/null");      system(buffer);    
    sprintf(buffer, "kill `cat ./pids/lasig.pid 2> /dev/null` 2> /dev/null");        system(buffer);
    sprintf(buffer, "kill `cat ./pids/af.pid 2> /dev/null` 2> /dev/null");           system(buffer);
    sprintf(buffer, "kill `cat ./pids/AudioFeatures.pid 2> /dev/null` 2> /dev/null");system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl_1.pid 2> /dev/null` 2> /dev/null");      system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl_2.pid 2> /dev/null` 2> /dev/null");      system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl_3.pid 2> /dev/null` 2> /dev/null");      system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl_4.pid 2> /dev/null` 2> /dev/null");      system(buffer);
    sprintf(buffer, "kill `cat ./pids/C45cl_1.pid 2> /dev/null` 2> /dev/null");      system(buffer);
    sprintf(buffer, "kill `cat ./pids/C45cl_2.pid 2> /dev/null` 2> /dev/null");      system(buffer);
    sprintf(buffer, "kill `cat ./pids/C45cl_3.pid 2> /dev/null` 2> /dev/null");      system(buffer);
    sprintf(buffer, "kill `cat ./pids/sv_accelfeat.pid 2> /dev/null` 2> /dev/null"); system(buffer);
    sprintf(buffer, "kill `cat ./pids/sv_audiofeat.pid 2> /dev/null` 2> /dev/null"); system(buffer);
    system("killall kismet_server 2> /dev/null");
    system("usbctl off 1 2> /dev/null");
}

void GroupMedia::Quit()
{
    printf("Shutting down...");

    cout << "killing processes...";
    KillAllProcesses();
    cout << "done..." ;
    close();
    cout << "Audiorecorder closed." << endl;
}

void GroupMedia::CheckName(){
    QString temp(nameBox->text());
    temp=temp.simplifyWhiteSpace();
    for(unsigned int i=0;i<temp.length();i++){
        QChar c=temp[i];
        if(c.isLetterOrNumber())
            continue;
        else if(c.isSpace())
            temp[i]=static_cast<QChar>('_');
        else if(c=='_')
            continue;
        else
            temp.remove(i--,1);
    }
    temp.truncate(20);
    nameBox->setText(temp);
}

void GroupMedia::StartRecording()
{	
    char buffer[256];
    char MITeChannels [16];
    int numMITeChannels = 0;
  
    if (!TestServer()) return;
    
    startButton->setEnabled(false);
    MITe1_checkBox->setEnabled(false);
    MITe2_checkBox->setEnabled(false);
    MITe3_checkBox->setEnabled(false);
    MITe4_checkBox->setEnabled(false);
    audio_checkBox->setEnabled(false);
    wifi_checkBox->setEnabled(false);
    camera_checkBox->setEnabled(false);
    nameBox->setEnabled(false);
    posture_ButtonGroup->setEnabled(true);
    activity_ButtonGroup->setEnabled(true);
    stopButton->setEnabled(true);
      
    strcpy(MITeChannels,"");      
    if(MITe1_checkBox->isChecked()) {
        sprintf(MITeChannels,"%s %s",MITeChannels,MITe1_lineEdit->text().latin1()); 
        numMITeChannels++;
    }
    if(MITe2_checkBox->isChecked()) {
        sprintf(MITeChannels,"%s %s",MITeChannels,MITe2_lineEdit->text().latin1()); 
        numMITeChannels++;
    }
    if(MITe3_checkBox->isChecked()) {
        sprintf(MITeChannels,"%s %s",MITeChannels,MITe3_lineEdit->text().latin1());    //not yet used!
        numMITeChannels++;
    }
    if(MITe4_checkBox->isChecked()) {
        sprintf(MITeChannels,"%s %s",MITeChannels,MITe4_lineEdit->text().latin1());    //not yet used!
        numMITeChannels++;
    }
    if (numMITeChannels > 0) {
        sprintf(buffer,"MITeSignal -c %d%s > ./data/MITeSignal.log & echo $! > ./pids/mitesig.pid",numMITeChannels,MITeChannels);
        system(buffer);
        printf("SignalRecorder: Recording MITe acceleration signals on channels%s\n",MITeChannels);
        if (!skipCalibration_checkBox->isChecked()){
            calibrating = 0;
            //Calibrate();
	    calibrationCounter->start( 5*1000 );
        }
        calibrateButton->setEnabled(true);
    }
    
    if(audio_checkBox->isChecked()) {
        printf("SignalRecorder: Recording audio signal\n");
        system("LinuxAudioSignal  /signals/audio 8000 1 2 1024 /dev/dsp & echo $! > ./pids/lasig.pid");   //huray, the new Zaurus 6000 can record at 8000Hz!
    //system("/mnt/card/borglab/enchant-signal/LinuxAudioDownSampler localhost:/rawaudio localhost:/signals/audio 3 0 & echo $! > /var/run/lads.pid");
    }
        
    if(nameBox->text().isEmpty()){
        strcpy(outputFile,"sr");  // default name "sr" as in SignalRecorder
    }
    else {
        CheckName();
        strcpy(outputFile,nameBox->text().latin1());
    }
  
  //sleep(1);
  
    strcat(outputFile,"_`hostname`_`date \"+%H_%M_%S\"`");   // e.g. Mark_z2_11_33_60
    sprintf(buffer, "echo %s > ./data/outputfile.txt",outputFile);
    system(buffer);
  
    FILE* fid = fopen("./data/outputfile.txt","r");
    if (fid != NULL) {
        fgets (outputFile, 64 , fid);   //now outputFile should is the real filename string
        strcpy(outputFile,strtok(outputFile,"\n"));
        fclose(fid);
    }
    else{
        printf("SignalRecoder: Error opening file ./data/outputfile.txt\n");
    }
  
    text->setText("Running...  Press Stop to end.");
      
    if(MITe1_checkBox->isChecked()) {
        sprintf(buffer, "SignalViewer -s localhost -p /signals/MITe/MITe_%s -t > ./data/%s.accel1 & echo $! > ./pids/sv_accel1.pid",MITe1_lineEdit->text().latin1(),outputFile);
        system(buffer);
        sprintf(buffer,"%s.accel1",outputFile);
        MITe1FileName_lineEdit->setText(buffer);
    }
    if(MITe2_checkBox->isChecked()) {
        sprintf(buffer, "SignalViewer -s localhost -p /signals/MITe/MITe_%s -t > ./data/%s.accel2 & echo $! > ./pids/sv_accel2.pid",MITe2_lineEdit->text().latin1(),outputFile);
        system(buffer);
        sprintf(buffer,"%s.accel2",outputFile);
        MITe2FileName_lineEdit->setText(buffer);
    }
    if(MITe3_checkBox->isChecked()) {
        sprintf(buffer, "SignalViewer -s localhost -p /signals/MITe/MITe_%s -t > ./data/%s.accel3 & echo $! > ./pids/sv_accel3.pid",MITe3_lineEdit->text().latin1(),outputFile);
        system(buffer);
        sprintf(buffer,"%s.accel3",outputFile);
        MITe3FileName_lineEdit->setText(buffer);
    }
    if(MITe4_checkBox->isChecked()) {
        sprintf(buffer, "SignalViewer -s localhost -p /signals/MITe/MITe_%s -t > ./data/%s.accel4 & echo $! > ./pids/sv_accel4.pid",MITe4_lineEdit->text().latin1(),outputFile);
        system(buffer);
        sprintf(buffer,"%s.accel4",outputFile);
        MITe4FileName_lineEdit->setText(buffer);
    }
    if(audio_checkBox->isChecked()) {
        sprintf(buffer, "SignalViewer -s localhost -p /signals/audio -t > ./data/%s.audio & echo $! > ./pids/sv_audio.pid",outputFile);
        system(buffer);
        sprintf(buffer,"%s.audio",outputFile);
        audioFileName_lineEdit->setText(buffer);
    }
    if(camera_checkBox->isChecked()) {
        printf("SignalRecorder: Recording camera, foto every %s seconds\n",camera_lineEdit->text().latin1());
    }
    if(wifi_checkBox->isChecked()) {
        printf("SignalRecorder: Recording WiFi\n");
        sprintf(buffer,"%s.wifi",outputFile);
        wifiFileName_lineEdit->setText(buffer);
        wifiState=0;
        wifiCounter->start( 20*1000 ); 
    }
    
    sprintf(buffer,"./data/%s.labels",outputFile);
    labelFileID = fopen(buffer,"w");
    if (labelFileID == NULL) {
        printf("SignalRecoder: Error opening file %s\n",buffer);
    }
  
    checkSensorCounter->start( 100*1000 );  //100 seconds
   
    if(camera_checkBox->isChecked()) {
        captureImageCounter->start( atoi(camera_lineEdit->text().latin1())*1000 );  
    }
    if (shutterPolling_checkBox->isChecked()) {
        buttonPollCounter->start( 5*1000 );  // wait 5 secs before we start polling for things to settle 
        pollingButton = 0; capturingImage = 0; cameraDriver = -1;
    }
        
    classify_startButton->setEnabled(true);
    tabWidget->setCurrentPage(1);
    
    if (SR_DEBUG) printf("SignalRecorder: StartRecording() done\n\n");
}

void GroupMedia::StopRecording()
{
    //char buffer[128];    
    
    stopButton->setEnabled(false);
    calibrateButton->setEnabled(false);
    classify_stopButton->setEnabled(false);
    classify_startButton->setEnabled(false);
    posture_ButtonGroup->setEnabled(false);
    activity_ButtonGroup->setEnabled(false);
    startButton->setEnabled(true);
    MITe1_checkBox->setEnabled(true);
    MITe2_checkBox->setEnabled(true);
    MITe3_checkBox->setEnabled(true);
    MITe4_checkBox->setEnabled(true);
    audio_checkBox->setEnabled(true);
    wifi_checkBox->setEnabled(true);
    camera_checkBox->setEnabled(true);
    nameBox->setEnabled(true);
    text->setText("Enter Name, select sensors, click start!");
  
    captureImageCounter->stop();
    checkSensorCounter->stop();
    buttonPollCounter->stop();
    fclose(labelFileID);
    
    //system("./sigrec_stop.sh &");
    KillAllProcesses();
        
    printf("SignalRecorder: StopRecording() done\n");
}

void GroupMedia::InitInterestMeasurement() 
{
    char buffer[128];
    int c,k;
    gettimeofday(&label_timestamp, NULL);
    lastMomentOfInterest = label_timestamp.tv_sec;
    ringBufferIndex = 0;
    
    if (SR_DEBUG) printf("SignalRecorder: Reading interest map from file...");
    FILE *fid = fopen("./interestMap.txt","r");
    if (fid == NULL){
        printf("SignalRecorder: WARNING, error opening ./interestMap.txt not performing interest measure.\n");
    }
    c = -1;
    k = 0;
    while (fgets(buffer, 128, fid) != NULL) {
        char* pch = strtok(buffer," \t");
        pch = strtok(NULL,"\n");
        if (strcmp(buffer,"Location") == 0) {
            c = 0; k = 0;
            continue;
        }
        if (strcmp(buffer,"Speech") == 0) {
            c = 1; k = 0;
            continue;
        }
        if (strcmp(buffer,"Posture") == 0) {
            c = 2; k = 0;
            continue;
        }
        if (strcmp(buffer,"Event") == 0) {
            c = 3; k = 0;
            continue;
        }
        if (c > -1) {
            if (k > MAX_LABELS_PER_CHAIN) {
                printf("SignalRecorder: WARNING, number of labels in chain %c exceeds hard-coded maximum (%d)!\n",c,MAX_LABELS_PER_CHAIN);
                k++;
                continue;
            }
            interestPoints[c][k] = atoi(buffer);
            labels[c][k] = (char *) malloc(strlen(pch)+2);
            strcpy(labels[c][k],pch);
            k++;
        }
    }
    if (SR_DEBUG) printf("done\n");
    for (c=0;c<4;c++) for (k=0;k<MAX_LABELS_PER_CHAIN; k++) {
        printf("c%d, %d: %d (%s)\n",c,k,interestPoints[c][k],(labels[c][k]==NULL) ? "NULL" : labels[c][k]);
    }
    interestThreshold = 5;
    
    
}

void GroupMedia::StartClassifier() {
    char buffer[256];
    SubscribeUNLInfo info;
    if (shutterPolling_checkBox->isChecked()) {
        buttonPollCounter->start( 10*1000 );  // wait 10 secs before we start polling for things to settle 
        pollingButton = 0; capturingImage = 0;
    }
    
    InitInterestMeasurement();
    
    printf("SignalRecorder: Starting AccelFeature\n");
    sprintf(buffer, "AccelFeature -i /signals/MITe/MITe_%s /signals/MITe/MITe_%s -o /feat/accel -m acc12_mean_var -c ./data/Config_MITes.txt -g 1 > /dev/null & echo $! > ./pids/af.pid",
            MITe1_lineEdit->text().latin1(),MITe2_lineEdit->text().latin1());
    system(buffer);
    printf("SignalRecorder: Starting AudioFeatures -s 0.0000024 \n");
    sprintf(buffer, "AudioFeatures -s 0.0000024 -p /signals/audio /feat/audio > /dev/null & echo $! > ./pids/AudioFeatures.pid");
    system(buffer);
    //sleep(1);
    printf("SignalRecorder: Starting GClassifier for Posture\n");
    sprintf(buffer, "GClassifier -m ../inference/models/posture_GMM.xml -i localhost:/feat/accel -p localhost:/class/posture -w ./data/%s.posture.class -d > /dev/null & echo $! > ./pids/GMMcl_1.pid",outputFile);
    printf("                %s\n",buffer);
    system(buffer);
    printf("SignalRecorder: Starting GClassifier for Event\n");
    sprintf(buffer, "GClassifier -m ../inference/models/event_GMM.xml -i localhost:/feat/accel -p localhost:/class/event -w ./data/%s.event.class -d > /dev/null & echo $! > ./pids/GMMcl_2.pid",outputFile);
    printf("                %s\n",buffer);
    system(buffer);
    printf("SignalRecorder: Starting GClassifier for Speech\n");
    sprintf(buffer, "GClassifier -m ../inference/models/speech_GMM.xml -i localhost:/feat/audio/states/bigFeatures -p localhost:/class/speech -w ./data/%s.speech.class -d > /dev/null & echo $! > ./pids/GMMcl_3.pid",outputFile);
    printf("                %s\n",buffer);
    system(buffer);
    printf("SignalRecorder: Starting GClassifier for Location\n");
    sprintf(buffer, "GClassifier -m ../inference/models/location_GMM.xml -i localhost:/feat/audio/states/bigFeatures -p localhost:/class/location -w ./data/%s.location.class -d > /dev/null & echo $! > ./pids/GMMcl_4.pid",outputFile);
    printf("                %s\n",buffer);
    system(buffer);
    
    printf("SignalRecorder: Starting C45Classifier for Posture\n");
    sprintf(buffer, "C45Classifier -m ../inference/models/posture_c45 -i localhost:/feat/accel -p localhost:/class/postureC45 -d -w ./data/%s.posture_c45.class > /dev/null & echo $! > ./pids/C45cl_1.pid",outputFile);
    printf("                %s\n",buffer);
    system(buffer);    
    printf("SignalRecorder: Starting C45Classifier for Event\n");
    sprintf(buffer, "C45Classifier -m ../inference/models/event_c45 -i localhost:/feat/accel -p localhost:/class/eventC45 -d -w ./data/%s.event_c45.class > /dev/null & echo $! > ./pids/C45cl_2.pid",outputFile);
    printf("                %s\n",buffer);
    system(buffer);    
    printf("SignalRecorder: Starting C45Classifier for Speech\n");
    sprintf(buffer, "C45Classifier -m ../inference/models/speech_c45 -i localhost:/feat/audio/states/bigFeatures -p localhost:/class/speechC45 -d -w ./data/%s.speech_c45.class > /dev/null & echo $! > ./pids/C45cl_3.pid",outputFile);
    printf("                %s\n",buffer);
    system(buffer);   
    
    sprintf(buffer, "SignalViewer -s localhost -p /feat/accel -t > ./data/%s.accelfeat & echo $! > ./pids/sv_accelfeat.pid",outputFile);
    system(buffer);
    sprintf(buffer, "SignalViewer -s localhost -p /feat/audio/states/bigFeatures -t > ./data/%s.audiofeat & echo $! > ./pids/sv_audiofeat.pid",outputFile);
    system(buffer);
    
    strcpy(buffer,"localhost:/class/posture/decision");
    printf("SignalRecorder: Subscribing to %s\n",buffer);
    SubscribeUNL(NULL, &info);
    if (!SubscribeUNL(buffer, &info)){
        printf("SignalRecorder: Error subscribing to classification node %s\n",buffer);
    }
    strcpy(buffer,"localhost:/class/event/decision");
    printf("SignalRecorder: Subscribing to %s\n",buffer);
    SubscribeUNL(NULL, &info);
    if (!SubscribeUNL(buffer, &info)){
        printf("SignalRecorder: Error subscribing to classification node %s\n",buffer);
    }
    strcpy(buffer,"localhost:/class/speech/decision");
    printf("SignalRecorder: Subscribing to %s\n",buffer);
    SubscribeUNL(NULL, &info);
    if (!SubscribeUNL(buffer, &info)){
        printf("SignalRecorder: Error subscribing to classification node %s\n",buffer);
    }
    strcpy(buffer,"localhost:/class/location/decision");
    printf("SignalRecorder: Subscribing to %s\n",buffer);
    SubscribeUNL(NULL, &info);
    if (!SubscribeUNL(buffer, &info)){
        printf("SignalRecorder: Error subscribing to classification node %s\n",buffer);
    }
    strcpy(buffer,"localhost:/class/postureC45/decision");
    printf("SignalRecorder: Subscribing to %s\n",buffer);
    SubscribeUNL(NULL, &info);
    if (!SubscribeUNL(buffer, &info)){
        printf("SignalRecorder: Error subscribing to classification node %s\n",buffer);
    }
    strcpy(buffer,"localhost:/class/eventC45/decision");
    printf("SignalRecorder: Subscribing to %s\n",buffer);
    SubscribeUNL(NULL, &info);
    if (!SubscribeUNL(buffer, &info)){
        printf("SignalRecorder: Error subscribing to classification node %s\n",buffer);
    }
    strcpy(buffer,"localhost:/class/speechC45/decision");
    printf("SignalRecorder: Subscribing to %s\n",buffer);
    SubscribeUNL(NULL, &info);
    if (!SubscribeUNL(buffer, &info)){
        printf("SignalRecorder: Error subscribing to classification node %s\n",buffer);
    }
    
    showClassificationCounter->start( 1000 );  
    classify_startButton->setEnabled(false);
    classify_stopButton->setEnabled(true);
}

void GroupMedia::StopClassifier() {
    showClassificationCounter->stop();
    char buffer[128];
    sprintf(buffer, "kill `cat ./pids/af.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/AudioFeatures.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl_1.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl_2.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl_3.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/GMMcl_4.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/C45cl_1.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/C45cl_2.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/C45cl_3.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/sv_accelfeat.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    sprintf(buffer, "kill `cat ./pids/sv_audiofeat.pid 2> /dev/null` 2> /dev/null");
    system(buffer);
    posture_lineEdit->setText("N/A");
    event_lineEdit->setText("N/A");
    speech_lineEdit->setText("N/A");
    location_lineEdit->setText("N/A");
    posture_c45_lineEdit->setText("N/A");
    event_c45_lineEdit->setText("N/A");
    speech_c45_lineEdit->setText("N/A");
    classify_startButton->setEnabled(true);
    classify_stopButton->setEnabled(false);
    system("ench unpublish /class");
    system("ench unpublish /feat");
}

int GroupMedia::TestServer() {
    GetNodeInfo info;
    if (!GetNode("localhost:/time", &info)) {
        QMessageBox::information(this,"Error","Unable to connect to server.",QMessageBox::Ok);
        fprintf(stderr, "SignalRecorder: Couldn't connect to server. Please restart server now!\n");
        serverButton->setEnabled(true);
        return 0;
    }
    if (SR_DEBUG) printf("SignalRecorder: TestServer() done\n");
    return 1;
}

int GroupMedia::DiscEnchant()
{
    return(1);
}


void GroupMedia::SelectPosture(int id)
{
    gettimeofday(&label_timestamp, NULL);  
    fprintf(labelFileID, "%06ld.%06ld %d\n",label_timestamp.tv_sec,label_timestamp.tv_usec,id);
    if (SR_DEBUG) printf("SignalRecorder: SelectActivity() selected %d\n",id);
}

void GroupMedia::SelectActivity(int id)
{
    gettimeofday(&label_timestamp, NULL);  
    fprintf(labelFileID, "%06ld.%06ld %d\n",label_timestamp.tv_sec,label_timestamp.tv_usec,id);
    if (SR_DEBUG) printf("SignalRecorder: SelectActivity() selected %d\n",id);
}


void GroupMedia::CheckAllSignals()
{
    char result[256];
    char buffer[128];
    int fileSize, totalFileSize = 0;
    int problem = 0;
  
    system("echo SignalRecorder: Checking status of signals\n");
    printf("SignalRecorder: Checking status of signals\n");
    //sprintf(result,"SignalRecorder has checked status of:\n\n");
    sprintf(result,"-");
  

    QPalette pal_green;
    QColorGroup cg;
    cg.setColor( QColorGroup::Foreground, black );
    cg.setColor( QColorGroup::Button, QColor( 192, 192, 192) );
    cg.setColor( QColorGroup::Light, white );
    cg.setColor( QColorGroup::Midlight, QColor( 223, 223, 223) );
    cg.setColor( QColorGroup::Dark, QColor( 96, 96, 96) );
    cg.setColor( QColorGroup::Mid, QColor( 128, 128, 128) );
    cg.setColor( QColorGroup::Text, black );
    cg.setColor( QColorGroup::BrightText, white );
    cg.setColor( QColorGroup::ButtonText, black );
    cg.setColor( QColorGroup::Base, QColor( 85, 255, 127) );
    cg.setColor( QColorGroup::Background, white );
    cg.setColor( QColorGroup::Shadow, black );
    cg.setColor( QColorGroup::Highlight, black );
    cg.setColor( QColorGroup::HighlightedText, white );
    pal_green.setActive( cg ); pal_green.setInactive( cg );  pal_green.setDisabled( cg );
  
    QPalette pal_red;
    cg.setColor( QColorGroup::Base, QColor( 255, 0, 0) );
    pal_red.setActive( cg ); pal_red.setInactive( cg );  pal_red.setDisabled( cg );
 
    if(audio_checkBox->isChecked()) {
        sprintf(buffer,"%s.audio",outputFile);
        fileSize = GetFileSize(buffer);
        totalFileSize += fileSize;
        sprintf(buffer,"%d byte",fileSize);
        audioFileSize_lineEdit->setText(buffer);
        if (fileSize > audioFileSize){
            strcat(result,"audio ok ");
            audioFileSize_lineEdit->setPalette(pal_green); 
        }
        else {
            strcat(result,"audio DOWN ");
            problem = 1;
            audioFileSize_lineEdit->setPalette(pal_red); 
        }
        audioFileSize = fileSize;
    }
    if(MITe1_checkBox->isChecked()) {
        sprintf(buffer,"%s.accel1",outputFile);
        fileSize = GetFileSize(buffer);
        totalFileSize += fileSize;
        sprintf(buffer,"%d byte",fileSize);
        MITe1FileSize_lineEdit->setText(buffer);
        if (fileSize > MITe1FileSize){
            sprintf(result,"%s MITe%s ok ",result,MITe1_lineEdit->text().latin1());
            MITe1FileSize_lineEdit->setPalette(pal_green); 
        }
        else {
            sprintf(result,"%s MITe%s DOWN ",result,MITe1_lineEdit->text().latin1());
            problem = 1;
            MITe1FileSize_lineEdit->setPalette(pal_red); 
        }
        MITe1FileSize = fileSize;
    }
    if(MITe2_checkBox->isChecked()) {
        sprintf(buffer,"%s.accel2",outputFile);
        fileSize = GetFileSize(buffer);
        totalFileSize += fileSize;
        sprintf(buffer,"%d byte",fileSize);
        MITe2FileSize_lineEdit->setText(buffer);
        if (fileSize > MITe2FileSize){
            sprintf(result,"%s MITe%s ok ",result,MITe2_lineEdit->text().latin1());
            MITe2FileSize_lineEdit->setPalette(pal_green); 
        }
        else {
            sprintf(result,"%s MITe%s DOWN ",result,MITe2_lineEdit->text().latin1());
            problem = 1;
            MITe2FileSize_lineEdit->setPalette(pal_red); 
        }
        MITe2FileSize = fileSize;
    }
    if(MITe3_checkBox->isChecked()) {
        sprintf(buffer,"%s.accel3",outputFile);
        fileSize = GetFileSize(buffer);
        totalFileSize += fileSize;
        sprintf(buffer,"%d byte",fileSize);
        MITe3FileSize_lineEdit->setText(buffer);
        if (fileSize > MITe3FileSize){
            sprintf(result,"%s MITe%s ok ",result,MITe3_lineEdit->text().latin1());
            MITe3FileSize_lineEdit->setPalette(pal_green); 
        }
        else {
            sprintf(result,"%s MITe%s DOWN ",result,MITe3_lineEdit->text().latin1());
            problem = 1;
            MITe3FileSize_lineEdit->setPalette(pal_red); 
        }
        MITe3FileSize = fileSize;
    }
    if(MITe4_checkBox->isChecked()) {
        sprintf(buffer,"%s.accel4",outputFile);
        fileSize = GetFileSize(buffer);
        totalFileSize += fileSize;
        sprintf(buffer,"%d byte",fileSize);
        MITe4FileSize_lineEdit->setText(buffer);
        if (fileSize > MITe4FileSize){
            sprintf(result,"%s MITe%s ok ",result,MITe4_lineEdit->text().latin1());
            MITe4FileSize_lineEdit->setPalette(pal_green); 
        }
        else {
            sprintf(result,"%s MITe%s DOWN ",result,MITe4_lineEdit->text().latin1());
            problem = 1;
            MITe4FileSize_lineEdit->setPalette(pal_red); 
        }
        MITe4FileSize = fileSize;
    }
    if(wifi_checkBox->isChecked()) {
        sprintf(buffer,"%s.wifi",outputFile);
        fileSize = GetFileSize(buffer);
        totalFileSize += fileSize;
        sprintf(buffer,"%d byte",fileSize);
        wifiFileSize_lineEdit->setText(buffer);
        if (fileSize > wifiFileSize){
            sprintf(result,"%s wifi ok ",result);
            wifiFileSize_lineEdit->setPalette(pal_green); 
        }
        else {
            sprintf(result,"%s wifi DOWN ",result);
            //problem = 1;
            wifiFileSize_lineEdit->setPalette(pal_red); 
        }
        wifiFileSize = fileSize;
    }
          
    sprintf(buffer,"%d byte",totalFileSize);
    totalFileSize_lineEdit->setText(buffer);
    
    if(camera_checkBox->isChecked()) {
        sprintf(buffer,"?");
        system("ls ./data/*.jpg | wc | awk '{print $1}' > ./data/size.temp"); 
        FILE* fid = fopen("./data/size.temp","r");
        if (fid != NULL) {
            fgets (buffer, 32, fid);      
            strcpy(buffer,strtok(buffer,"\n"));
            fclose(fid);
            printf("SignalRecorder:      Number of pictures taken: %s\n",buffer);
        }
        else{
            printf("SignalRecoder: Error opening file size.temp\n");
        }
        
        CameraFileSize_lineEdit->setText(buffer);
        sprintf(result,"%sCamera --> %s pics\n",result,buffer);
    }
        
    status_lineEdit->setText(result);
    
    CheckBattery();                
    CheckCPU();
           
    if (problem) {
        //PlayBuzzer();
        //QMessageBox::information(this,"CheckSignals()",result,QMessageBox::Ok);
        SetStandbyLED(0);
    }
    else {
        SetStandbyLED(1);
    }
}


int GroupMedia::GetFileSize(char* fileName)
{
    char buffer[256];
  
    sprintf(buffer,"ls ./data -al | grep %s | awk '{print $5}' > ./data/size.temp",fileName);
    system(buffer);
    FILE* fid = fopen("./data/size.temp","r");
    if (fid != NULL) {
        fgets (buffer, 32, fid);      
        strcpy(buffer,strtok(buffer,"\n"));
        fclose(fid);
        printf("SignalRecorder:      Size of output file (%s) is %s\n",fileName,buffer);
        return atoi(buffer);
    }
    else{
        printf("SignalRecoder: Error opening file size.temp\n");
        return -1;
    }
}


void GroupMedia::CaptureImageSlot(){
    CaptureImage(NULL);
}
void GroupMedia::CaptureImage(char* fileName)
{
    char buffer[128];
    pollingButton = 0;
    capturingImage = 1;
    //printf("CaptureImage(): cameraDriver on entry is %d\n",cameraDriver);
    if (shutterPolling_checkBox->isChecked()) {
        // hold off polling for 8 seconds, not to interfere with the picture taking
        //buttonPollCounter->changeInterval(8*1000);   
        buttonPollCounter->start(8*1000);   
    }
    if (cameraDriver != -1) {
        ::close( cameraDriver );
        //printf("CaptureImage(): closed driver, (cameraDriver now %d)\n",cameraDriver);
        cameraDriver = -1;
    }
    if (fileName == NULL) {
        gettimeofday(&label_timestamp, NULL);
        sprintf(buffer,"./Capture -x 480 -y 464 -q 70 ./data/%06ld.jpg &",label_timestamp.tv_sec);
    }
    else{
        sprintf(buffer,"./Capture -x 480 -y 464 -q 70 ./data/%s.jpg &",fileName);
    }
    if (SR_DEBUG) { printf(buffer);  printf("\n"); }
    system(buffer);
    
    if (!audio_checkBox->isChecked()) {       // this means the audio device driver is free
        char *device;
  
        device = "/dev/dsp";
    
        if (! linux_init_write_sound(device, 22050, 1, 0)) 
        {
            fprintf(stderr,"SignalRecorder: Unable to configure linux audio input\n");
            return;
        }
      
        if (! linux_play_wave("/home/QtPalmtop/sounds/shtr.wav"))
        {
            fprintf(stderr,"SignalRecorder: Unable to play wav file\n");
            return;
        }
    
        linux_audio_close();
    }
    capturingImage = 0;
}

// void GroupMedia::timerEvent( QTimerEvent *e )
// {
// }

void GroupMedia::PollButton()
{
    char b[4];
    int rlen,ret;
    
    // open() for the camera driver takes 0.6 seconds. So we only close and reopen when the polling is
    // interrupted by an attempt to capture an image. This needs to be done here because the program
    // taking the foto, Capture, is a seperate process and we dont know when it ends.
    if (capturingImage) {
        printf("SignalRecorder: PollButton() no good, capturing Image (cameraDriver == %d)\n",cameraDriver);
        return;
    }
    
    if (!pollingButton){
        if (cameraDriver != -1) {
            printf("SignalRecorder: PollButton() warning, cameraDriver == %d\n",cameraDriver);
            //::close(cameraDriver);
            //cameraDriver = -1;
            //buttonPollCounter->start(300);   //give time to close driver
            //return;
        }
        else
        {
            cameraDriver = open("/dev/sharp_zdc", O_RDWR );
            if ( cameraDriver == -1 ){
                printf("SignalRecorder: Error: Could not open camera driver!\n");
                return;
            }
            //printf("SignalRecorder: PollButton() opened cameraDriver == %d\n",cameraDriver);
        }
        pollingButton = 1;
        buttonPollCounter->changeInterval(300);   // we start polling every 0.3 seconds
    }
    
    if ( cameraDriver == -1 ){
        printf("SignalRecorder, PollButton(): Error: camera driver was not open!\n");
        return;
    }
    ret = write( cameraDriver, "M=1" , 3 );
    if (ret != 3){
        printf("SignalRecorder: Error: could not write to camera!\n");
        return;
    }
    if ( cameraDriver == -1 ){
        printf("SignalRecorder, PollButton(): Error: camera driver was not open!\n");
        return;
    }
    rlen = read( cameraDriver, b, 4 );
    if ( rlen ) {
        //if (SR_DEBUG) printf("SignalRecorder: Read %d bytes: %c %c %c %c\n",rlen,b[0], b[1],b[2],b[3]); 
    }
    else {
        printf("SignalRecorder: Error, Read nothing from camera driver!\n");
        return;
    }
    
    if (b[0]=='S'){
        if (SR_DEBUG) printf("SignalRecorder: Shutter pressed --> taking image\n");
        CaptureImage(NULL);
    }
    return;
}

void GroupMedia::LaunchServer()
{
    system("/mnt/card/borglab/enchant-base/EnchantmentServer -d");
    serverButton->setEnabled(false);
}


void GroupMedia::ShowClassification()
{
    GetUpdateInfo upInfo;
    char interestStr[] = "                ";
    int c,i, count = 0;
    int staticInterest = 0;
    gettimeofday(&label_timestamp, NULL);
    while(1) {
        GetUpdate(&upInfo, NULL);
        count++;
        // there is an issue with the way children of published branch nodes arrive using GetUpdate() :
        // Server: Start
        // Client 1: Subscribe /A/B/C
        // Client 2: Publish B/C (branch B with child node C) to A/B
        // Client1: GetUpdate() -> {node = NULL, unl = /A/B/C}
        // Client1: GetUpdate() ->{node = C, unl = /A/B/C}
        // ... the sequence repeats ...
        // Client 2: Publish B/C (branch B with child node C) to A/B
        // Client1: GetUpdate() -> {node = NULL, unl = /A/B/C}
        // Client1: GetUpdate() ->{node = C, unl = /A/B/C}
        // ...
        if (upInfo.node != NULL) {
            //printf("received update from %s:\n",upInfo.UNL);
            if (strcmp(upInfo.UNL,"localhost:/class/location/decision") == 0) {
                //if (SR_DEBUG) printf("                Current location Classification is %s\n",upInfo.node->data.string_val);
                location_lineEdit->setText(upInfo.node->data.string_val);
                for (i=0; i<MAX_LABELS_PER_CHAIN; i++) {
                    if (strcmp(upInfo.node->data.string_val,labels[0][i]) == 0) {
                        currentClassification[0] = i;                        
                    }
                }
            }
            else if (strcmp(upInfo.UNL,"localhost:/class/speech/decision") == 0) {
                //if (SR_DEBUG) printf("                Current speech Classification is %s\n",upInfo.node->data.string_val);
                speech_lineEdit->setText(upInfo.node->data.string_val);
                for (i=0; i<MAX_LABELS_PER_CHAIN; i++) {
                    if (strcmp(upInfo.node->data.string_val,labels[1][i]) == 0) {
                        currentClassification[1] = i;                        
                    }
                }
                
            }
            else if (strcmp(upInfo.UNL,"localhost:/class/posture/decision") == 0) {
                //if (SR_DEBUG) printf("                Current posture Classification is %s\n",upInfo.node->data.string_val);
                posture_lineEdit->setText(upInfo.node->data.string_val);
                for (i=0; i<MAX_LABELS_PER_CHAIN; i++) {
                    if (strcmp(upInfo.node->data.string_val,labels[2][i]) == 0) {
                        currentClassification[2] = i;                        
                    }
                }
            }
            else if (strcmp(upInfo.UNL,"localhost:/class/event/decision") == 0) {
                //if (SR_DEBUG) printf("                Current event Classification is %s\n",upInfo.node->data.string_val);
                event_lineEdit->setText(upInfo.node->data.string_val);
                for (i=0; i<MAX_LABELS_PER_CHAIN; i++) {
                    if (strcmp(upInfo.node->data.string_val,labels[3][i]) == 0) {
                        currentClassification[3] = i;                        
                    }
                }
            }
            
            else if (strcmp(upInfo.UNL,"localhost:/class/postureC45/decision") == 0) {
                //if (SR_DEBUG) printf("                Current postureC45 Classification is %s\n",upInfo.node->data.string_val);
                posture_c45_lineEdit->setText(upInfo.node->data.string_val);
            }
            else if (strcmp(upInfo.UNL,"localhost:/class/eventC45/decision") == 0) {
                //if (SR_DEBUG) printf("                Current postureC45 Classification is %s\n",upInfo.node->data.string_val);
                event_c45_lineEdit->setText(upInfo.node->data.string_val);
            }
            else if (strcmp(upInfo.UNL,"localhost:/class/speechC45/decision") == 0) {
                //if (SR_DEBUG) printf("                Current postureC45 Classification is %s\n",upInfo.node->data.string_val);
                speech_c45_lineEdit->setText(upInfo.node->data.string_val);
            }
            else 
            {
                printf("Warning, received unknown update: %s\n",upInfo.UNL);
            }
            FreeNode(upInfo.node);
        }
        else {// (upInfo.node == NULL)
            //printf("count = %d\n",count);
            if ((upInfo.UNL == NULL) || (count >= 10)) break; 
        }
    }
    
    ///// interest calculation ////////////////////////////////////////
    for (c=0; c<NUM_CHAINS; c++){
        staticInterest += interestPoints[c][currentClassification[c]];
    }
    
    //printf("--transitionInterest = ");
    ringBufferIndex = (ringBufferIndex + 1) % RING_BUFFER_SIZE;
    int stateCounter[MAX_LABELS_PER_CHAIN];
    char newSuperState = 0;
    for (c=0; c<NUM_CHAINS; c++){
        classificationRingBuffer[c][ringBufferIndex] = currentClassification[c];
        //stateCounter(i) = length(find(classificationRingBuffer == i))
        for (i=0; i<MAX_LABELS_PER_CHAIN; i++) stateCounter[i] = 0;
        for (i=0; i<RING_BUFFER_SIZE; i++){
            stateCounter[classificationRingBuffer[c][i]]++;
        }
        //[foo newSuperState] = max(stateCounter)
        int maxCount = 0;
        for (i=0; i<MAX_LABELS_PER_CHAIN; i++){
            if (stateCounter[i] > maxCount){
                newSuperState = i;
                maxCount = stateCounter[i];
            }
        }
        if (currentSuperState[c] != newSuperState)  transitionInterest += 0.5;
        currentSuperState[c] = newSuperState;
    }
    //printf("%2.1f\n",transitionInterest);

    float timeoutInterest = ((float)(label_timestamp.tv_sec - lastMomentOfInterest)) / 120;  //equivalent to 1 interestPoint for 2min without picture
        
    currentInterest = staticInterest + timeoutInterest + transitionInterest;
    //printf("idx %d  static = %d, timeout = %2.1f, transition = %2.1f %d %d %d %d\n",ringBufferIndex,staticInterest,timeoutInterest,transitionInterest,
    //currentSuperState[0],currentSuperState[1],currentSuperState[2],currentSuperState[3]);
    
    sprintf(interestStr,"%2.1f",currentInterest);
    interest_lineEdit->setText(interestStr);
    if (currentInterest > interestThreshold) MomentOfInterest();
    return;
}

void GroupMedia::MomentOfInterest()
{
    FILE* fid;
    char buffer[128];
    gettimeofday(&label_timestamp, NULL);
    if ((label_timestamp.tv_sec - lastMomentOfInterest) < 5){
        //don't take pictures continuously..
        return;
    }
        
    if (SR_DEBUG) {
            printf("MoI: ");
            printf("%06ld.%06ld %d %d %d %d %d %d %d %d %2.2f\n",label_timestamp.tv_sec,label_timestamp.tv_usec,
            currentSuperState[0],currentSuperState[1],currentSuperState[2],currentSuperState[3],
            currentClassification[0],currentClassification[1],currentClassification[2],currentClassification[3],
            currentInterest);
    }
    sprintf(buffer,"./data/%s.interest",outputFile);
    fid = fopen(buffer,"a");
    if (fid == NULL){
        printf("SignalRecorder: Error could not create %s\n",buffer);
        //return;
    }
    else
    {
        fprintf(fid,"%06ld.%06ld %d %d %d %d %d %d %d %d %2.2f\n",label_timestamp.tv_sec,label_timestamp.tv_usec,
            currentSuperState[0],currentSuperState[1],currentSuperState[2],currentSuperState[3],
            currentClassification[0],currentClassification[1],currentClassification[2],currentClassification[3],
            currentInterest);
        fclose(fid);
    }
    transitionInterest = 0;
    lastMomentOfInterest = label_timestamp.tv_sec;
    
    if(camera_checkBox->isChecked()) {
        sprintf(buffer,"%06ld.i",label_timestamp.tv_sec);
        CaptureImage(buffer);
    }
}


void GroupMedia::CheckBattery()
{
    char buffer[128];
    float foo;
    int x1, x2, x3, x4, percentage;
    char percentageStr[5];
    
    FILE* fid = fopen("/proc/apm","r");
    if (fid == NULL){
        printf("SignalRecorder: Error opening /proc/apm  \n");
        return;
    }
    fscanf(fid,"%f %f %x %x %x %x %s",&foo,&foo,&x1,&x2,&x3,&x4,percentageStr);
    //printf("foo %f   foo %f   x1 %d   x2 %d   x3 %d   x4 %d   percentage %s\n",foo,foo,x1,x2,x3,x4,percentageStr);
    fclose(fid);
    
        //buffer = NULL;
    
    if (x2 == 1){
        sprintf(buffer,"power online");
    }
    else{
        sprintf(buffer, "power offline");
    }
    if (x3 == 3)
        sprintf(buffer, "%s, charging",buffer);
    
    percentage = atoi(strtok(percentageStr,"%"));
    sprintf(buffer,"%s %d %%\n",buffer,percentage);
    
    battery_lineEdit->setText(buffer);
    if (SR_DEBUG) {
        printf("SignalRecorder:      Battery status:");
        printf("%s",buffer);
    }
}

void GroupMedia::CheckCPU()
{
    char buffer[8];
    float f1, f2, f3;
    
    FILE* fid = fopen("/proc/loadavg","r");
    if (fid == NULL){
        printf("SignalRecorder: Error opening /proc/loadavg  \n");
        return;
    }
    fscanf(fid,"%f %f %f",&f1,&f2,&f3);
    fclose(fid);
    
    if (SR_DEBUG) printf("SignalRecorder:      CPU usage: %0.1f (1 minute) %0.1f (5 minute) %0.1f (10minute average)\n",f1, f2, f3);
    sprintf(buffer,"%0.1f%%\n",f1);
    cpu_lineEdit->setText(buffer);
}

void GroupMedia::SetStandbyLED(int status)
{
    int fd;
    sharp_led_status ledStatus;
    
    if (SR_DEBUG) printf("SignalRecorder: Setting StandbyLED to %d\n",status);
    
    //printf("opening leds\n");
    if ((fd = ::open (LED_DEVICE, O_RDWR|O_NONBLOCK)) == -1) {
        printf ("SignalRecorder: Error, Problems opening led device\n");
        return;
    }
    //printf("set led to 1\n");
    ledStatus.which =  SHARP_LED_MAIL_EXISTS;
    ledStatus.status = status;
    if (ioctl(fd, SHARP_LED_SETSTATUS, &ledStatus) == -1) {
        printf ("SignalRecorder: Error, Problems setting standby led\n");
        return;
    }
    ::close(fd);           
}


void GroupMedia::PlayBuzzer()
{
    int fd;
    sharp_buzzer_status zbs, zbs_save;
    sharp_led_status ledStatus,ledStatus_save;
    
    if (SR_DEBUG) printf("SignalRecorder: Playing buzzer\n");
    
    //printf("opening buzzer\n");
    if ((fd = ::open (BUZZER_DEVICE, O_RDWR|O_NONBLOCK)) == -1) {
        printf ("SignalRecorder: Error, Problems opening buzzer device\n");
        return;
    }

    //printf("save old settings\n");
    zbs_save.which =  SHARP_BUZ_SCHEDULE_ALARM;
    if (ioctl(fd, SHARP_BUZZER_GETVOLUME, &zbs_save) == -1) {
        printf ("SignalRecorder: Error, Problems saving current buzzer settings\n");
        return;
    }

    //printf("set new volume\n");
    zbs.which =  SHARP_BUZ_SCHEDULE_ALARM;
    zbs.volume = SHARP_BUZ_VOLUME_HIGH;
    zbs.mute = 0;
    if (ioctl(fd, SHARP_BUZZER_SETVOLUME, &zbs) == -1) {
        printf ("SignalRecorder: Error, Problems setting new buzzer volume\n");
        return;
    }

    //printf("make sound\n");      // note: SHARP_BUZ_SCHEDULE_ALARM == 11 , this is the only sound that plays on the SL-6000
    if (ioctl(fd, SHARP_BUZZER_MAKESOUND,SHARP_BUZ_SCHEDULE_ALARM) == -1) {
        printf ("SignalRecorder: Error, Problems playing buzzer sound\n");
            //return;
    }
  
    //printf("reset to saved\n");
    if (ioctl(fd, SHARP_BUZZER_SETVOLUME, &zbs_save) == -1) {
        printf ("SignalRecorder: Error, Problems resetting buzzer settings\n");
        return;
    }

    ::close(fd);           
    
    if (SR_DEBUG) printf("SignalRecorder: Blinking LEDs\n");
    
    //printf("opening leds\n");
    if ((fd = ::open (LED_DEVICE, O_RDWR|O_NONBLOCK)) == -1) {
        printf ("SignalRecorder: Error, Problems opening led device\n");
        return;
    }

    //printf("save old settings\n");
    ledStatus_save.which =  SHARP_LED_MAIL_EXISTS;
    if (ioctl(fd, SHARP_LED_GETSTATUS, &ledStatus_save) == -1) {
        printf ("SignalRecorder: Error, Problems saving current led settings\n");
        return;
    }

    for (int i=0; i<10; i++) {
    //printf("toggling LED\n");
        ledStatus.which =  SHARP_LED_MAIL_EXISTS;
        ledStatus.status = 1;
        if (ioctl(fd, SHARP_LED_SETSTATUS, &ledStatus) == -1) {
            printf ("SignalRecorder: Error, Problems setting led\n");
            return;
        }
        usleep(40000);
        ledStatus.which =  SHARP_LED_MAIL_EXISTS;
        ledStatus.status = 0;
        if (ioctl(fd, SHARP_LED_SETSTATUS, &ledStatus) == -1) {
            printf ("SignalRecorder: Error, Problems setting led\n");
            return;
        }
        usleep(40000);
    }
    //printf("reset to saved\n");
    if (ioctl(fd, SHARP_LED_SETSTATUS, &ledStatus_save) == -1) {
        printf ("SignalRecorder: Error, Problems resetting led settings\n");
        return;
    }
    ::close(fd);           
}


void GroupMedia::WifiSniffer()
{
    // This function is called on each timeout of wifiCounter. The timing is necessary because
    // the different system calls in this function require a certain time to securly execute.
    // wifiState is a GroupMedia member which keeps track of the current state of WifiSniffer.
    
    switch (wifiState) {
        case 0:
            if (SR_DEBUG) printf("SignalRecorder: starting WifiSniffer()\n");
            system("rm /mnt/card/borglab/wifidata/Kismet.network");
            system("usbctl on 1 2> /dev/null");
            wifiState++;
            wifiCounter->start( 2000 );
            break;
        case 1:
            system("wlanctl-ng wlan0 lnxreq_ifstate ifstate=enable > /dev/null");
            wifiState++;
            wifiCounter->start( 2000 );
            break;
        case 2:
            system("ifconfig wlan0 1.1.1.1 up");
            wifiState++;
            wifiCounter->start( 2000 );
            break;
        case 3:
            system("kismet_server 2> /dev/null &");
            locationWiFi_lineEdit->setText("searching...");
            wifiState++;
            wifiCounter->start( 15000 );
            break;
        case 4:
            system("killall kismet_server");
            wifiState++;
            wifiCounter->start( 2000 );
            break;
        case 5:
            system("usbctl off 1 2> /dev/null");
            system("grep \"Network\" ../wifidata/Kismet.network > ./data/wifi.temp");
            
            FILE* fid_wifitemp;
            FILE* fid_wifilog;
            FILE* fid_lookupTable;
            char textViewBuffer[512];
            char buffer[128];
            char netName[64];
            char SSID[32];
            char SSID_file[32];
            char locationStr[32];
            char * pch;
            struct timeval time;
            int n;
            
            fid_wifitemp = fopen("./data/wifi.temp","r");
            if (fid_wifitemp == NULL){
                printf("SignalRecorder: Error opening ./data/wifi.temp\n");
                wifiState = 0;
                wifiCounter->start( 100000 );  // start again in 100 seconds
                return;
            }
            
            sprintf(buffer,"./data/%s.wifi",outputFile);
            fid_wifilog = fopen(buffer,"a");
            if (fid_wifilog == NULL){
                printf("SignalRecorder: Error opening %s\n",buffer);
                wifiState = 0;
                wifiCounter->start( 100000 );  // start again in 100 seconds
                return;
            }
            fid_lookupTable = fopen("../inference/models/wifiLookupTable.txt","r");
            if (fid_lookupTable == NULL){
                printf("SignalRecorder: Error opening ../inference/models/wifiLookupTable.txt\n");
                wifiState = 0;
                wifiCounter->start( 100000 );  // start again in 100 seconds
                return;
            }
            
            gettimeofday(&time, NULL);
            n = 0;
            strcpy(textViewBuffer,""); strcpy(locationStr,"no match");
            while (fgets(buffer, 128, fid_wifitemp) != NULL) {
                // extract netName and SSID
                pch = strtok(buffer,"\"");
                pch = strtok(NULL,"\"");
                strcpy(netName,pch);
                pch = strtok(NULL,"\"");
                pch = strtok(NULL,"\"");
                strcpy(SSID,pch);
                fprintf(fid_wifilog, "%06ld.%06ld %s %s\n",time.tv_sec,time.tv_usec,netName,SSID);
                sprintf(textViewBuffer,"%s %s   %s\n",textViewBuffer,netName,SSID);
                
                //go through lookup table
                while (fgets(buffer, 128, fid_lookupTable) != NULL) {
                    sscanf(buffer,"%s %s",SSID_file,locationStr);
                    if (strcmp(SSID,SSID_file) == 0) { 
                        if (SR_DEBUG) printf("SignalRecorder: Found wifi location match: %s\n",locationStr);
                        locationWiFi_lineEdit->setText(locationStr);
                        
                        PublishNodeInfo *pubInfo = NULL;
                        NodeType *updateNode = CreateStringNode("locationWiFi", T_STRING, locationStr);
                        if (!PublishNode("/class/locationWiFi", updateNode, pubInfo))
                            printf("SignalRecorder.cpp: Error publishing node\n");
                        FreeNode(updateNode);
                        
                        break; //we found a matching entry, disregard other entries
                    }
                }
                n++;
            }
            
            
            if (n == 0) {
                fprintf(fid_wifilog, "%06ld.%06ld none\n",time.tv_sec,time.tv_usec);
                locationWiFi_lineEdit->setText("none");
            }
            if (SR_DEBUG) printf("SignalRecorder: WifiSniffer found %d networks\n",n);
            wifi_TextView->setText(textViewBuffer);
            
            fclose(fid_lookupTable);
            fclose(fid_wifitemp);
            fclose(fid_wifilog);
            wifiState = 0;
            wifiCounter->start( 100000 );  // start again in 100 seconds
            break;
            
        default:
            printf("SignalRecorder: Error, entered impossible wifiState '%d'\n",wifiState);
        return;
    }
    return;
}


void GroupMedia::Calibrate()
{
    // Calibrate will 1. launch MITeCalibrate, set a timer, and will then 2. read the calibration file, only to display the results
    // The function is called by StartRecording, or during the recording process by clicking the Calibrate button (usefull if initial calibration was bad).
    
    FILE* fid;
    char* calibrationFilePrefix = "./data/Config";
    char buffer[256];
    char tempBuffer[128];
    
    if (!calibrating){  // 1. start MITeCalibrate
        if (MITe3_checkBox->isChecked() || MITe4_checkBox->isChecked()){
            printf("SignalRecorder: Calibration does not support more than 2 MITes");
            return;
        }
        QMessageBox::information(this,"MITe Calibration","You have 30sec to perform MITe calibration.\n"
                "Rotate MITes arround all axis, hold still at least 2sec in all 6 positions!\n\nPress Ok to start",QMessageBox::Ok);
        printf("SignalRecorder: Starting MITe calibration");
                        
        if (MITe1_checkBox->isChecked() && MITe2_checkBox->isChecked()){
            sprintf(buffer,"MITeCalibrate -d /signals/MITe/MITe_%s /signals/MITe/MITe_%s -o %s -r %d -t 30 &",
                    MITe1_lineEdit->text().latin1(),MITe2_lineEdit->text().latin1(),calibrationFilePrefix,90);
        }
        else // only MITe1 can be checked
        {
            sprintf(buffer,"MITeCalibrate -d /signals/MITe/MITe_%s -o %s -r %d -t 30 &",MITe1_lineEdit->text().latin1(),calibrationFilePrefix,180);
        }
        system(buffer);
        calibrating = 1;
        calibrationCounter->start( 38*1000 );
    }
    else // 2. calibrating, read and interpret results
    {
        calibrationCounter->stop();
        printf("reading resulsts\n");
        sprintf(tempBuffer,"%s_MITes.txt",calibrationFilePrefix);
        fid = fopen(tempBuffer,"r");
        if (fid == NULL){
            printf("SignalRecorder: Error opening %s\n",tempBuffer);
            return;
        }
        strcpy(buffer,"Calibration Result:\n\n");
        while(fgets(tempBuffer,128,fid) != NULL){
            sprintf(buffer,"%s%s",buffer,tempBuffer);
        }
        fclose(fid);
        QMessageBox::information(this,"MITe Calibration",buffer,QMessageBox::Ok);
        printf("SignalRecorder: %s",buffer);
        calibrating = 0;
    }
    return;        
}

void GroupMedia::setInterestThreshold(int k) {
    interestThreshold = k;
    if (SR_DEBUG) printf("SignalRecorder: Interest threshold was set to %d\n",k);
}
