/*****************************************************************************
 *
 * Author: Daniel Barkalow, MIT Wearable Computing Group
 * http://www.media.mit.edu/wearables/
 * Copyright (C) 2003 MIT Media Lab, 
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <enchantment/whiteboard.h>
#include <enchantment/signals.h>
#include <enchantment/time-utils.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <string.h>
#include <errno.h>

#define USE_THREADS

#ifdef USE_THREADS
#include <pthread.h>
#include <semaphore.h>
#endif

#define CERT_FILE "cert.pem"
#define KEY_FILE "key.pem"

#ifdef USE_THREADS
pthread_mutex_t fileLock;
#define lock(name) pthread_mutex_lock(&name)
#define unlock(name) pthread_mutex_unlock(&name)
#else
#define lock(name)
#define unlock(name)
#endif

int ignoreEOF = 0;

void usage(char *arg0)
{
  fprintf(stderr, "Usage: %s [OPTION]...\n", arg0);
  fprintf(stderr, 
	  "Perform various buffering and replay operations on signals.\n\n");
  fprintf(stderr, "  -i [UNL]	read from [UNL]\n");
  fprintf(stderr, "  -o [UNL]	write to [UNL]\n");
  fprintf(stderr, "  -b [UNL]	write to UNLs based on [UNL]\n");
  fprintf(stderr, "  -r [FILE]	read from [FILE]\n");
  fprintf(stderr, "  -w [FILE]	write to [FILE]\n");
  fprintf(stderr, "  -a [FILE]	append to [FILE]\n");
  fprintf(stderr, "  -c [AUTH]	hold data if there is no connected reader;\n");
  fprintf(stderr, "		when reading, send receipts after the data\n");
  fprintf(stderr, "		is committed.\n");
  fprintf(stderr, "  -t [FILE]	use [FILE] for temporary storage if needed\n");
  fprintf(stderr, "  -s [SVR]	make [SVR] the default enchantment server\n");
  fprintf(stderr, "  -f       	fast; ignore original speed\n");
  fprintf(stderr, "  -p		proxy through the server\n");
  fprintf(stderr, "  -n		use new file format\n");
  fprintf(stderr, "  -l [UNL]	include labels from [UNL]\n");
  fprintf(stderr, "  -e		wait for more file data on EOF\n");
  fprintf(stderr, "\n");
  fprintf(stderr, 
	  "Exactly one source (-i, -r) must be given. If no sink (-o, -w)\n"
	  "is specified, standard out is used. In caching mode, enables\n"
	  "the reliable delivery of a signal when the receiver might get\n"
	  "disconnected temporarily, provided this client does not get\n"
	  "disconnected from the signal.\n");
  fprintf(stderr, "\n");
}

typedef struct SystemInfo SystemInfo;

typedef struct
{
  Signal *signal;
  char *sigUNL;
  char *labelUNL;
} InputInfo;

typedef struct
{
  Signal *signal;
  char *name;
} StreamInfo;

char *auth = NULL;

typedef struct
{
  void *data;
  void (*got_signal)(struct SystemInfo *, unsigned stream, 
		     struct timeval timestamp, void *buffer);
  void (*got_label)(struct SystemInfo *, unsigned stream, char *label);
} OutputInfo;

// There can be any number of streams.

// Each stream has its own input driver.

// There is only one output driver, which takes all of the streams of
// data.

struct SystemInfo
{
  unsigned timeset : 1; // have we chosen the time offset?

  unsigned stream_count; // number of things going through
  StreamInfo *streams;

  InputInfo **inputs;

  signed long offset;

  OutputInfo *output;
};

FILE *infile = NULL;

int fast = 0;

// Output modules

typedef struct
{
  Signal *signal;
  int channel_fd;
} outsig;

static void outsig_signal(SystemInfo *info, unsigned stream, 
			  struct timeval timestamp, void *buffer)
{
  outsig *data = info->output->data;
  SignalGenAccept(data->signal);
  if (data->channel_fd >= 0)
    {
      unsigned char channel = stream;
      write(data->channel_fd, &channel, 1);
    }
  SignalSend(data->signal, &timestamp, buffer);
}

OutputInfo *output_signal(Signal *signal, int channel_fd)
{
  OutputInfo *ret = malloc(sizeof(OutputInfo));
  outsig *data = malloc(sizeof(outsig));
  data->signal = signal;
  data->channel_fd = channel_fd;
  ret->data = data;
  ret->got_signal = outsig_signal;
  ret->got_label = NULL;
  return ret;
}

static void text_signal(SystemInfo *info, unsigned stream,
			struct timeval timestamp, void *buffer)
{
  FILE *outfile = info->output->data;
  Signal *inSig = info->streams[stream].signal;
  char *name = info->streams[stream].name;
  fprintf(outfile, "vector ");
  if (name)
    fprintf(outfile, "%s ", name);
  if (!inSig)
    printf("No inSig?\n");
  PrintSignalElements(outfile, inSig, timestamp, buffer);
  if (auth)
    SignalRecvd(inSig, timestamp);
  fflush(outfile);
}

static void text_label(SystemInfo *info, unsigned stream, char *label)
{
  FILE *outfile = info->output->data;
  fprintf(outfile, "label %s\n", label);
}

OutputInfo *output_text(FILE *file)
{
  OutputInfo *ret = malloc(sizeof(OutputInfo));
  ret->data = file;
  ret->got_signal = text_signal;
  ret->got_label = text_label;
  return ret;
}

// Functions to call the virtual methods

void got_signal_data(SystemInfo *info, unsigned stream,
		     struct timeval timestamp, void *buffer)
{
  if (!fast)
    {
      struct timeval now;
      if (!info->timeset)
	{
	  gettimeofday(&now, NULL);
	  info->offset = timeval_sub(&now, &timestamp);
	  info->timeset = 1;
	}
      else
	{
	  struct timeval until = timestamp;
	  timeval_add(&until, info->offset);
	  wait_until(&until, NULL);
	}
    }
  lock(fileLock);
  (*info->output->got_signal)(info, stream, timestamp, buffer);
  unlock(fileLock);
}

void got_label(SystemInfo *info, unsigned stream, char *label)
{
  if (info->output->got_label)
    {
      lock(fileLock);
      (*info->output->got_label)(info, stream, label);
      unlock(fileLock);
    }
}

// XXXX for loop_from_text, loop_from_signal
Signal *signal_spec;

void loop_from_text(SystemInfo *system)
{
  struct timeval timestamp;
  void *buffer = malloc(SignalBufferSize(signal_spec));
  do
    {
      if (fscanf(infile, " vector ") < 0)
	{
	  if (ignoreEOF)
	    continue;
	  exit(0);
	}
      while (fgetc(infile) != ' ')
	;
      if (!ScanSignalElements(infile, signal_spec, &timestamp, buffer))
	{
	  printf("Error reading file\n");
	  continue;
	  //exit(1);
	}
      got_signal_data(system, 0, timestamp, buffer);
    }
  while (1);
}

int proxy = 0;

int read_channel_fd = -1;

void loop_from_signal(SystemInfo *system, unsigned stream)
{
  struct timeval timestamp;
  void *buffer = NULL;
  Signal *inSig;
  InputInfo *info = system->inputs[stream];

  if (info->signal)
    {
      inSig = info->signal;
      if (SignalConnected(inSig))
	{
	  buffer = malloc(SignalBufferSize(inSig));
	  if (signal_spec)
	    SignalCopyProperties(inSig, signal_spec);
	}
    }
  else
    {
      inSig = NewSignalUNL(info->sigUNL);
      info->signal = inSig;
    }
  system->streams[stream].signal = inSig;

  if (info->labelUNL)
    SubscribeUNL(info->labelUNL, NULL);

  do
    {
      // Loop on getting data
      do
	{
	  GetUpdateInfo updateInfo;
	  // Loop on connecting
	  while (!SignalConnected(inSig))
	    {
	      if (info->sigUNL)
		printf("Waiting for %s\n", info->sigUNL);
	      if (!(proxy ? SignalConnectByProxy(inSig) :
		    SignalConnectWait(inSig)))
		{
		  return;
		}
	      printf("Done waiting for %s\n", info->sigUNL);
	      if (!buffer)
		{
		  printf("Properties set\n");
		  buffer = malloc(SignalBufferSize(inSig));
		  if (signal_spec)
		    SignalCopyProperties(inSig, signal_spec);
		}
	    }
	  if (info->labelUNL)
	    {
	      GetUpdate(&updateInfo, info->labelUNL);
	      if (updateInfo.publisher && updateInfo.node && 
		  HasFormat(updateInfo.node->type, T_STRING))
		{
		  // XXXX stream...
		  got_label(system, 1, updateInfo.node->data.string_val);
		}
	    }
	  if (read_channel_fd >= 0)
	    {
	      char channel;
	      read(read_channel_fd, &channel, 1);
	    }
	}
      while (!SignalRecv(inSig, &timestamp, buffer));
      
      got_signal_data(system, stream, timestamp, buffer);
    }
  while (1);
}

struct sig_thread_info
{
  SystemInfo *system;
  unsigned stream;
};

void *sig_thread_proc(void *data)
{
  struct sig_thread_info *info = data;
  loop_from_signal(info->system, info->stream);
  return NULL;
}

int main(int argc, char **argv)
{
  int opt;
  char *inUNL = NULL;
  DArray inUNLs = NULL;
  char *outUNL = NULL;
  char *outUNLBase = NULL;
  char *inFileName = NULL;
  char *outFileName = NULL;
  char *labelUNL = NULL;
  int append = 0;
  int help = 0;
  int newproto = 0;
  void *buffer = NULL;
  struct timeval timestamp;
  struct timeval base;

  SystemInfo *system;

  do
    {
      opt = getopt(argc, argv, "I:i:o:r:w:a:t:c:s:fpl:b:ne");
      if (opt == 's')
	{
	  SetDefaultServer(optarg);
	}
      else if (opt == 'i')
	{
	  if (inFileName)
	    {
	      fprintf(stderr, "Only one input source allowed\n\n");
	      help = 1;
	    }
	  if (inUNLs || inUNL)
	    {
	      if (!inUNLs)
		{
		  inUNLs = ArrayNew(sizeof(char *), 2, free);
		  ArrayAppend(inUNLs, &inUNL);
		  inUNL = NULL;
		}
	      ArrayAppend(inUNLs, &optarg);
	    }
	  else
	    {
	      inUNL = optarg;
	    }
	  // If you're playing from a signal, assume that it comes in at the
	  // desired rate.
	  fast = 1;
	}
      else if (opt == 'p')
	{
	  proxy = 1;
	}
      else if (opt == 'b')
	{
	  if (outUNL || outUNLBase)
	    fprintf(stderr, "Only one output UNL currently supported\n");
	  outUNLBase = optarg;
	}
      else if (opt == 'o')
	{
	  if (outUNL || outUNLBase)
	    fprintf(stderr, "Only one output UNL currently supported\n");
	  outUNL = optarg;
	}
      else if (opt == 'r')
	{
	  if (inFileName || inUNL)
	    {
	      fprintf(stderr, "Only one input source allowed\n\n");
	      help = 1;
	    }
	  inFileName = optarg;
	}
      else if (opt == 'w')
	{
	  if (outFileName)
	    fprintf(stderr, "Only one output file currently supported\n");
	  outFileName = optarg;
	}
      else if (opt == 'a')
	{
	  if (outFileName)
	    fprintf(stderr, "Only one output file currently supported\n");
	  outFileName = optarg;
	  append = 1;
	}
      else if (opt == 'f')
	{
	  fast = 1;
	}
      else if (opt == 'e')
	{
	  ignoreEOF = 1;
	}
      else if (opt == 'c')
	{
	  auth = optarg;
	}
      else if (opt == 'l')
	{
	  labelUNL = optarg;
	}
      else if (opt == 'I')
	{
          char *certname = malloc(strlen(optarg) + strlen(CERT_FILE) + 1);
          char *keyname = malloc(strlen(optarg) + strlen(KEY_FILE) + 1);
          strcpy(certname, optarg);
          strcpy(certname + strlen(optarg), CERT_FILE);
          strcpy(keyname, optarg);
          strcpy(keyname + strlen(optarg), KEY_FILE);
          if (!UseIdentity(certname, keyname))
            printf("Identity in %s and %s could not be used.\n",
                   certname, keyname);
	}
      else if (opt == 'n')
	{
	  newproto = 1;
	}
      else if (opt > 0)
	{
	  printf("Unknown option\n");
	  help = 1;
	}
    }
  while (opt > 0);

  if (!inUNL && !inFileName && !inUNLs)
    {
      fprintf(stderr, "Input source required\n");
      help = 1;
    }

  if (outUNL && outFileName)
    {
      fprintf(stderr, "Only one output supported\n");
      help = 1;
    }

  if (help)
    {
      usage(argv[0]);
      exit(0);
    }

  system = malloc(sizeof(SystemInfo));

  if (outUNL)
    {
      Signal *outSig = NewSignalGenUNL(outUNL);
      if (auth)
	SignalGenGetReceipts(outSig, auth);
      signal_spec = outSig;
      system->output = output_signal(outSig, -1);
    }
  else
    {
      if (newproto)
	{
	  int channel_fd;
	  Signal *outSig;
	  char channel = 0;
	  printf("Opening %s\n", outFileName);
	  if (!outFileName || !strcmp(outFileName, "-"))
	    channel_fd = 1;
	  else
	    channel_fd = open(outFileName, O_WRONLY | O_CREAT |
			      (append ? O_APPEND : 0), 00666);
	  if (channel_fd < 0)
	    {
	      fprintf(stderr, "Unable to write to %s: %s\n", outFileName,
		      strerror(errno));
	      exit(-1);
	    }
	  write(channel_fd, &channel, 1);
	  outSig = NewSignalGenFile(channel_fd);
	  signal_spec = outSig;
	  system->output = output_signal(outSig, channel_fd);
	}
      else
	{
	  FILE *outfile;
	  if (!outFileName || !strcmp(outFileName, "-"))
	    outfile = stdout;
	  else
	    {
	      outfile = fopen(outFileName, append ? "a" : "w");
	      if (!outfile)
		{
		  fprintf(stderr, "Unable to write to %s: %s\n", outFileName,
			  strerror(errno));
		  exit(-1);
		}
	    }
	  system->output = output_text(outfile);
	}
    }

  if (inUNL)
    {
      InputInfo info;
      info.signal = NULL;
      info.sigUNL = inUNL;
      info.labelUNL = labelUNL;
      system->stream_count = 1;
      system->inputs = malloc(sizeof(InputInfo *));
      system->inputs[0] = &info;
      system->streams = malloc(sizeof(StreamInfo));
      system->streams[0].name = inUNL;
      loop_from_signal(system, 0);
    }
  else if (inUNLs)
    {
      int i;
      InputInfo info;
      system->stream_count = ArrayLength(inUNLs);
      system->inputs = malloc(sizeof(InputInfo *) * system->stream_count);
      system->streams = malloc(sizeof(StreamInfo) * system->stream_count);
      for (i = 0; i < ArrayLength(inUNLs) - 1; i++)
	{
	  InputInfo *subinfo = malloc(sizeof(InputInfo));
	  pthread_t thread;
	  struct sig_thread_info *thread_info =
	    malloc(sizeof(struct sig_thread_info));
	  subinfo->signal = NULL;
	  subinfo->sigUNL = *(char **)ArrayNth(inUNLs, i);
	  subinfo->labelUNL = NULL;
	  system->inputs[i] = subinfo;
	  system->streams[i].name = *(char **)ArrayNth(inUNLs, i);
	  thread_info->system = system;
	  thread_info->stream = i;
	  pthread_create(&thread, NULL, sig_thread_proc, thread_info);
	}
      
      info.signal = NULL;
      info.sigUNL = *(char **)ArrayNth(inUNLs, i);
      system->streams[i].name = *(char **)ArrayNth(inUNLs, i);
      info.labelUNL = labelUNL;
      system->inputs[i] = &info;
      loop_from_signal(system, i);
    }
  else if (newproto)
    {
      InputInfo info;
      char channel = 0;

      read_channel_fd = open(inFileName, O_RDONLY);
      if (read_channel_fd < 0)
	{
	  fprintf(stderr, "Unable to read from %s: %s\n", inFileName,
		  strerror(errno));
	  exit(-1);
	}
 
      printf("Reading from file\n");
      if (read(read_channel_fd, &channel, 1) != 1)
	{
	  printf("Empty file?\n");
	}
      printf("Channel: %d\n", channel);
      info.signal = NewSignalFile(read_channel_fd);
      info.sigUNL = NULL;
      info.labelUNL = NULL;
      system->stream_count = 1;
      system->inputs = malloc(sizeof(InputInfo *));
      system->streams = malloc(sizeof(StreamInfo));
      system->inputs[0] = &info;
      system->streams[0].name = NULL;
      if (!SignalConnected(info.signal))
	printf("Empty file?\n");
      loop_from_signal(system, 0);
    }
  else
    {
      if (!strcmp(inFileName, "-"))
	infile = stdin;
      else
	infile = fopen(inFileName, "r");
      if (!infile)
	{
	  fprintf(stderr, "Unable to read from %s: %s\n", inFileName, 
		  strerror(errno));
	  exit(-1);
	}
      buffer = malloc(7);
      
      fread(buffer, 1, 7, infile);
      if (!strncmp(buffer, "vector ", 7))
	{
	  while (fgetc(infile) != ' ')
	    ;
	  ScanSignalElements(infile, signal_spec, &base, NULL);
	  if (SignalGetUnitSamples(signal_spec))
	    {
	      long diff;
	      while (fscanf(infile, " vector ") < 0)
		{
		  if (ignoreEOF)
		    continue;
		  exit(0);
		}
	      while (fgetc(infile) != ' ')
		;
	      ScanSignalElements(infile, signal_spec, &timestamp, NULL);

	      diff = timeval_sub(&timestamp, &base);

	      if (diff)
		SignalSetSampleRate(signal_spec, 
				    1000000.0 * 
				    SignalGetUnitSamples(signal_spec) / 
				    timeval_sub(&timestamp, &base),
				    SignalGetUnitSamples(signal_spec));
	      printf("Sample rate: %g\n", SignalGetSampleRate(signal_spec));
	    }
	}
      free(buffer);
      buffer = NULL;
      fseek(infile, 0, SEEK_SET);
      loop_from_text(system);
    }
  return 0;
}
