/*************************************************
 * File: identity.c
 * Author: Daniel Barkalow
 * Birth Date: 11/24/03
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#include <enchantment/identity.h>

#include <stdlib.h>

#ifndef WANT_SSL
Identity *NewIdentity(char *cert, char *key)
{
  return NULL;
}

int UseIdentity(char *cert, char *key)
{
  return 0;
}

Identity *DefaultIdentity()
{
  return NULL;
}

#else // WANT_SSL

#include <enchantment/types.h>

#include <string.h>

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/ssl.h>
#include <openssl/x509v3.h>

#define ENCHANT_PASS_ENV "ENCHANT_PASS"

struct Identity
{
  SSL_CTX *ctx;
};

static Identity *def_ident = NULL;
static Identity *fallback_ident = NULL;

#define BUFSIZE 1024

static void print_openssl_errors()
{
  int flags;
  int line;
  const char *data;
  const char *file;
  unsigned long code;
  char errbuf[BUFSIZE];
  
  do
    {
      code = ERR_get_error_line_data(&file, &line, &data, &flags);
      ERR_error_string_n(code, errbuf, (BUFSIZE - 1));
      
      printf("%s\n", errbuf);
      printf("in %s line %d.\n\n", file, line);
      if (data && (flags && ERR_TXT_STRING))
	{
	  printf("error data: %s\n", data);
	}
    }
  while (code);
}

#define MESSAGE "Passphrase required for %s: "

static int get_passphrase(char *buf, int size, int flag, void *userdata)
{
  char *pass = getenv(ENCHANT_PASS_ENV);
  if (pass)
    {
      strncpy(buf, pass, size);
      return strlen(pass);
    }
  if (1)
    {
      char *prompt = malloc(strlen(MESSAGE) - 2 + strlen(userdata) + 1);
      sprintf(prompt, MESSAGE, (char *) userdata);
      if (!EVP_read_pw_string(buf, size, prompt, flag))
	return strlen(buf);
    }
  return 0;
}

static int verify_none(int preverify_ok, X509_STORE_CTX *ctx)
{
  // Accept everyone for now.
  return 1;
}

static int init_ssl()
{
  if (!SSL_library_init())
    {
      print_openssl_errors();	
      fprintf(stderr, "OpenSSL initialization failed.");
      return -1;
    }
  SSL_load_error_strings();
  return 0;
}

#define SHOW_INIT_ERRORS 0

Identity *NewIdentity(char *cert, char *key)
{
  Identity *ret;
  SSL_CTX *ctx;
  if (init_ssl() != 0)
    return NULL;

  ctx = SSL_CTX_new(SSLv23_method());
  
  SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, &verify_none);

  SSL_CTX_set_default_passwd_cb(ctx, &get_passphrase);
  SSL_CTX_set_default_passwd_cb_userdata(ctx, key);
  if (cert && SSL_CTX_use_certificate_chain_file(ctx, cert) != 1)
    {
      if (SHOW_INIT_ERRORS)
	print_openssl_errors();
      SSL_CTX_free(ctx);
      return NULL;
    }
  if (key && SSL_CTX_use_PrivateKey_file(ctx, key, SSL_FILETYPE_PEM) != 1)
    {
      if (SHOW_INIT_ERRORS)
	print_openssl_errors();
      SSL_CTX_free(ctx);
      return NULL;
    }
  SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY | SSL_MODE_ENABLE_PARTIAL_WRITE);
  ret = (Identity *)malloc(sizeof(Identity));
  ret->ctx = ctx;
  return ret;
}

int UseIdentity(char *certfile, char *keyfile)
{
  def_ident = NewIdentity(certfile, keyfile);
  return !!def_ident;
}

Identity *DefaultIdentity()
{
  return def_ident;
}

Identity *FallbackIdentity()
{
  if (def_ident)
    return def_ident;
  if (!fallback_ident)
    {
      fallback_ident = NewIdentity(NULL, NULL);
    }
  return fallback_ident;
}

#endif
