/*************************************************
 * File: time-utils.c
 * Birth Date: Aug 23, 2004
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2003 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 *************************************************/

#include <enchantment/time-utils.h>
#include <stdlib.h>
#include <time.h>

void timeval_add(struct timeval *tm, long usecs)
{
  tm->tv_usec += usecs;
  if (usecs < 0)
    {
      while (tm->tv_usec < 0)
        {
          tm->tv_sec--;
          tm->tv_usec += 1000000;
        }
    }
  else
    {
      while (tm->tv_usec > 1000000)
        {
          tm->tv_sec++;
          tm->tv_usec -= 1000000;
        }
    }
}

void timeval_add_many(struct timeval *tm, long long usecs)
{
  tm->tv_sec += (usecs / 1000000);
  tm->tv_usec += (usecs % 1000000);
  if (usecs < 0)
    {
      while (tm->tv_usec < 0)
        {
          tm->tv_sec--;
          tm->tv_usec += 1000000;
        }
    }
  else
    {
      while (tm->tv_usec > 1000000)
        {
          tm->tv_sec++;
          tm->tv_usec -= 1000000;
        }
    }
}

long timeval_sub(struct timeval *t1, struct timeval *t2)
{
  return (t1->tv_sec - t2->tv_sec) * 1000000 +
    (t1->tv_usec - t2->tv_usec);
}

void timeval_round(struct timeval *tm, unsigned long period)
{
  int64_t t = (((int64_t) tm->tv_sec) * 1000000) + tm->tv_usec;
  t -= t % period;
  tm->tv_sec = t / 1000000;
  tm->tv_usec = t % 1000000;
}

static long minsleep = 1000;
static long spintime = 1000;

void wait_until(struct timeval *tm, struct timeval *actual)
{
  struct timeval timestamp;
  struct timespec timer;
  long diff;
  if (!actual)
    actual = &timestamp;
  gettimeofday(actual, NULL);
  diff = timeval_sub(tm, actual);
  if (diff > 0)
    {
      while (diff > minsleep)
	{
	  timer.tv_sec = (diff - spintime) / 1000000;
	  timer.tv_nsec = ((diff - spintime) % 1000000) * 1000;
	  nanosleep(&timer, NULL);
	  gettimeofday(actual, NULL);
	  diff = timeval_sub(tm, actual);
	}
      while (diff > 0)
	{
	  gettimeofday(actual, NULL);
	  diff = timeval_sub(tm, actual);
	}
    }
  if (actual)
    *actual = timestamp;
}

struct timestamp_sequence
{
  unsigned wait : 1;
  unsigned started : 1;
  unsigned correct : 1;
  float rate;
  unsigned length;

  long step;

  struct timeval current;
};

timestamp_sequence *new_timestamp_sequence(float rate, unsigned length,
					   unsigned wait)
{
  timestamp_sequence *ret = malloc(sizeof(timestamp_sequence));
  ret->wait = !!wait;
  ret->correct = !wait;
  ret->rate = rate;
  ret->length = length;
  ret->step = 1000000 / ret->rate * ret->length;
  ret->started = 0;
  return ret;
}

struct timeval *timestamp_sequence_next(timestamp_sequence *seq)
{
  if (!seq->started)
    {
      gettimeofday(&seq->current, NULL);
      seq->started = 1;
    }
  else
    {
      long step = seq->step;
      struct timeval now;
      long offset;
      gettimeofday(&now, NULL);
      offset = timeval_sub(&seq->current, &now) + step;
      step -= offset / 10000;
      timeval_add(&seq->current, step);
    }
  return &seq->current;
}
