/*************************************************
 * File: clients.c
 * Birth Date: Dec 03, 2003
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2003, 2004 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *************************************************/

#include <enchantment/types.h>

#include "features.h"
#include "clients.h"

#include <enchantment/node.h>

#include "tree-manip.h"

#include "introduction.h"
#include "protocol.h"
#include "ipc.h"
#include <enchantment/unl.h>

#include "logging.h"

#include <stdlib.h>
#include <string.h>

#include <sys/time.h>

#include "request.h"
#include "whiteboard-protocol.h"

#include "main.h"

#include <assert.h>

typedef struct {
  ClientIDType *client;
  NodeType *replace;
  NodeType *with;
} SendNodeData;

static NodeType *SendNodeProc(SendNodeData *data, NodeType *node)
{
  if (InTree(node))
    {
      if (!CheckNodePermission(node, data->client, 'r'))
	return NULL;
      if (IsSpecial(node))
	{
	  if (node->type == T_TIMEVAL)
	    {
	      gettimeofday(node->data.timeval_val, NULL);
	    }
	}
    }
  if (node == data->replace)
    return data->with;
  return node;
}

static int SendNode(ClientIDType *client, Socket *socket, NodeType *node,
		    NodeType *replace, NodeType *with)
{
  SendNodeData data = {
    .client = client,
    .replace = replace,
    .with = with
  };
  plogf("Send to %s:\n", SocketName(socket));
  logNode(node);
  WriteNode(socket, node, (IPCNodeProc) SendNodeProc, &data);
  return 1;
}

static NodeType *RecvNodeProc(void *data, NodeType *node)
{
  return node;
}

static int ReceiveNode(Socket *socket, NodeType **node)
{
  ReadNode(socket, node, NULL, RecvNodeProc);
  return 1;
}

void SendUpdate(SubscriptionType *sub, NodeType *node, const char *UNL,
		const char *publisherName, NodeType *replace, NodeType *with)
{
  Socket *socket = sub->subscriber->socket;
  if (SHOW_UPDATES)
    {
      plogf("Update to %s: %s published to %s:\n", sub->subscriber->name,
	    publisherName, UNL);
      logNode(node);
    }
  
  SocketSendRecord(socket, "type", 4);
  WriteString(socket, S_UPDATE, strlen(S_UPDATE));
  if (UNL)
    {
      SocketSendRecord(socket, "unl", 3);
      WriteString(socket, UNL, strlen(UNL));
    }
  SocketSendRecord(socket, "id", 2);
  WriteString(socket, InstanceNumber(sub->id), 
	      strlen(InstanceNumber(sub->id)));
  if (publisherName)
    {
      SocketSendRecord(socket, "publisher", 9);
      WriteString(socket, publisherName, strlen(publisherName));
    }
  if (node)
    {
      SocketSendRecord(socket, "node", 4);
      SendNode(sub->subscriber, socket, node, replace, with);
    }
  SocketSendRecord(socket, NULL, 0);
  SocketFlush(socket);
}

void SendResult(ClientIDType *client, NodeType *dataNode, 
		const char *callType, const char *UNL, int result, 
		const char **others)
{  
  Socket *socket = client->socket;
  char resultStr[6];
  int i;
  
  if (!socket || !callType)
    {
      printf("Error in args to SendResultNodeSocket\n");
      return;
    }

  sprintf(resultStr, "0x%03x", result);

  SocketSendRecord(socket, "type", 4);
  WriteString(socket, S_RETURN, strlen(S_RETURN));
  SocketSendRecord(socket, "command", 7);
  WriteString(socket, callType, strlen(callType));
  SocketSendRecord(socket, "result", 6);
  WriteString(socket, resultStr, 5);
  if (UNL)
    {
      SocketSendRecord(socket, "unl", 3);
      WriteString(socket, UNL, strlen(UNL));
    }
  if (dataNode)
    {
      SocketSendRecord(socket, "node", 4);
      SendNode(client, socket, dataNode, NULL, NULL);
    }
  if (others)
    {
      for (i = 0; others[i * 2]; i++)
	{
	  SocketSendRecord(socket, others[i * 2], strlen(others[i * 2]));
	  WriteString(socket, others[i * 2 + 1], strlen(others[i * 2 + 1]));
	}
    }
  SocketSendRecord(socket, NULL, 0);
  SocketFlush(socket);
}

extern DArray g_clients;

NodeType *GetClientNode(ClientIDType *clientID)
{
  NodePathReturnType npReturn;

  npReturn.UNL = AppendUNLPath("/clients", clientID->name);
  npReturn.cross_symlinks = 0;
  npReturn.follow_symlinks = 0;
  npReturn.getSubscriptions = 0;
  GetNodePathInfo(serverID, &npReturn, 0);
  return npReturn.existNode;
}

int HandleNewClient(Socket *socket)
{
  NodeType *childNode = NULL;
  NodeType *newClient;
  NodeType *info = NULL;
  static int serialID = 0;
  char clientName[MAX_NAME_LEN];
  const char *clientPath;
  ClientIDType *client;
  unsigned char *ip;
  int use_ssl = 0;
  int i;
  int done;

  do
    {
      IntroMessage message;
      int ret;
      done = 1;
      do
	{
	  ret = IntroRecv(socket, whiteboardClientProperties, &message);
	  if (ret)
	    {
	      IntroFree(&message);
	      done = 0;
	    }
	}
      while (ret);
      if (!done)
	{
	  done = 1;
	  IntroDone(socket);
	}
    }
  while (!done);
    
  sprintf(clientName, "Client%d", serialID);
  serialID++;

  printf("Accept %s\n", clientName);

  newClient = CreateBranchNode(clientName, T_BRANCH);
  
  clientPath = AppendUNLPath("/clients", newClient->name);
  
  client = (ClientIDType *) malloc(sizeof(ClientIDType));

  client->subscribedNodes = NULL;
  client->locks = NULL;
  client->socket = socket;
  client->name = newClient->name;
  client->nextID = 1;

  ArrayAppend(g_clients, &client);

  if (use_ssl) 
    {
      SocketSecure(socket, NULL, 1);
    }

  // Determine the authentication of this client
  client->auth = ArrayNew(sizeof(char *), 0, NULL);
  if (1)
    {
      char *auth = malloc(strlen(newClient->name) + strlen(NAME_AUTH) + 1);
      strcpy(auth, NAME_AUTH);
      strcpy(auth + strlen(NAME_AUTH), newClient->name);
      ArrayAppend(client->auth, &auth);
    }
  ip = (char *) &SocketRemote(socket)->ip;
  if (ip[0] == 127 &&
      ip[1] == 0 &&
      ip[2] == 0 &&
      ip[3] == 1)
    {
      char *auth = strdup(NETWORK_AUTH "localhost");
      ArrayAppend(client->auth, &auth);
    }
  if (SocketIsSecure(socket))
    {
      char *fingerprint = SocketRemoteFingerprint(socket);
      if (fingerprint)
	{
	  char *auth = malloc(strlen(fingerprint) + 
			      strlen(CERTIFICATE_AUTH) + 1);
	  strcpy(auth, CERTIFICATE_AUTH);
	  strcpy(auth + strlen(CERTIFICATE_AUTH), fingerprint);
	  ArrayAppend(client->auth, &auth);
	  free(fingerprint);
	}
    }
  
  if (ArrayLength(client->auth))
    {
      info = CreateBranchNode("Auth", T_BRANCH);
      for (i = 0; i < ArrayLength(client->auth); i++)
	{
	  char *auth = *((char **)ArrayNth(client->auth, i));
	  childNode = CreateStringNode(InstanceNumber(i), T_STRING, auth);
	  InsertNode(info, childNode);
	  //printf("%s has auth %s\n", newClient->name, auth);
	}
      InsertNode(newClient, info);
    }

  // XXXX this sends updates in the wrong thread!
  PublishLocal(clientPath, newClient, 0);

#ifdef USE_THREADS
  if (okay)
    {
      int run = 1;
      printf("Starting client loop\n");
      do
	{
	  run = ServiceClient(socket, client);
	}
      while (run);
    }
#else
  AddSocket(socket);
#endif
  return 1;
}

void RemoveClient(ClientIDType *clientID)
{
  ClientIDType *childID;
  const char *clientPath;
  int i;

  printf("Remove %s\n", clientID->name);
  plogf("Removing client %s\n", clientID->name);
  
  RemoveClientTreeEffects(clientID);
  clientPath = AppendUNLPath("/clients", clientID->name);
  PublishLocal(clientPath, NULL, 0);
  for (i = 0; i < ArrayLength(g_clients); i++)
    {
      childID = *((ClientIDType **)ArrayNth(g_clients, i));
      if (childID == clientID)
	{
	  ArrayDeleteAt(g_clients, i);
	  break;
	}
    }

  for (i = 0; i < ArrayLength(clientID->auth); i++)
    free(*((char **)ArrayNth(clientID->auth, i)));
  ArrayFree(clientID->auth);
  free(clientID);
}

int ServiceClient(Socket *clientSock, ClientIDType *clientID)
{
  NodeType *rcvNode = NULL;
  char *buffer;
  size_t size;
  const char *command = NULL;
  const char *unl = NULL;
  unsigned char flags = 0;
  OptionSet options = NULL;
  plogf("Process client %s\n", clientID->name);
  do
    {
      if (!SocketRecvRecord(clientSock, &buffer, &size))
	break;
      if (size)
	{
	  if (!strcmp(buffer, "node"))
	    {
	      if (!ReceiveNode(clientSock, &rcvNode))
		{
		  if (SocketError(clientSock) && CLIENT_ERROR)
		    {
		      printf("Failed to receive a node from %s(%s): %s\n", 
			     clientID->name,
			     SocketName(clientSock),
			     strerror(SocketError(clientSock)));
		    }
		}
	    }
	  else
	    {
	      char *value = NULL;
	      size_t size;
	      ReadString(clientSock, &value, &size);
	      if (!strcmp(buffer, "command"))
		{
		  command = InstanceString(value);
		  free(value);
		}
	      else if (!strcmp(buffer, "unl"))
		{
		  unl = InstanceString(value);
		  free(value);
		}
	      else if (!strcmp(buffer, "links"))
		{
		  if (value[0] == 'f')
		    flags |= FOLLOW_SYMLINK | CROSS_SYMLINK;
		  if (value[0] == 'c')
		    flags |= CROSS_SYMLINK;
		  free(value);
		}
	      else
		{
		  if (!options)
		    options = NewOptionSet();
		  AddSetOption(options, InstanceString(buffer), value);
		  //printf("%s\n", buffer);
		}
	    }
	}
    }
  while (size);

  plogf("Enqueued %s %s %x request from %s\n", command, unl, flags,
	clientID->name);
  if (rcvNode)
    {
      logNode(rcvNode);
    }
  if (!command)
    {
      RemoveSocket(clientSock);
      EnqueueRequest(clientID, NULL, NULL, NULL, 0, NULL);
      return 0;
    }
  EnqueueRequest(clientID, rcvNode, command, unl, flags, options);
  return 1;
}

#ifndef USE_THREADS
int ServiceClients(fd_set *p_sockSet)
{
  int numServiced = 0;
  int i;
  ClientIDType *clientID;
  Socket *clientSock;

  if (!p_sockSet) return 0;

  //printf("Service\n");

  for (i = 0; i < ArrayLength(g_clients); i++)
    {
      clientID = *((ClientIDType **)ArrayNth(g_clients, i));
      
      if (clientID == serverID) continue;

      // Get the socket out of the node's local data
      clientSock = clientID->socket;
      assert(clientSock);
      
      //printf("Service %s\n", clientNode->name);
      if (!SocketIncluded(p_sockSet, clientSock)) 
	continue;
      ServiceClient(clientSock, clientID);
      numServiced++;
    }
  
  // Return number of clients serviced
  return numServiced;
}
#endif
