/*************************************************
 * File: request.c
 * Birth Date: Dec 03, 2003
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2003, 2004 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *************************************************/

#include <enchantment/types.h>
#include <enchantment/node.h>
#include <enchantment/unl.h>

#include "symbols.h"
#include "features.h"
#include "logging.h"
#include "clients.h"
#include "tree-manip.h"
#include "request.h"
#include "whiteboard-protocol.h"
#include <enchantment/whiteboard.h>

#include <string.h>
#include <stdlib.h>

#ifdef USE_THREADS
#include <pthread.h>
#include <semaphore.h>
#endif

DArray requestQueue = NULL;

#ifdef USE_THREADS
pthread_mutex_t requestLock;
pthread_cond_t requestPending;
#endif

typedef struct
{
  ClientIDType *clientID;
  NodeType *rcvNode;
  const char *command;
  OptionSet options;
  unsigned flags : 6;
  unsigned freeData : 1;

  NodeType *dataNode;
  int status;
  NodePathReturnType npReturn;
  const char **others;
} RequestType;

#define RequestHasFlag(request, flag) (!!((request)->flags & (flag)))

void InitRequestQueue() 
{
  requestQueue = ArrayNew(sizeof(RequestType), 0, NULL);
#ifdef USE_THREADS
  pthread_mutex_init(&requestLock, NULL);
  pthread_cond_init(&requestPending, NULL);
#endif
}

void EnqueueRequest(ClientIDType *clientID, NodeType *rcvNode, 
		    const char *command, const char *unl, 
		    unsigned char flags, OptionSet options)
{
  RequestType request;
  request.clientID = clientID;
  request.rcvNode = rcvNode;
  request.command = command;
  request.flags = flags;
  request.npReturn.UNL = unl;
  request.npReturn.cross_symlinks = !!(flags & CROSS_SYMLINK);
  request.npReturn.follow_symlinks = !!(flags & FOLLOW_SYMLINK);
  request.npReturn.drop_auth = 0;//!!(flags & DROP_AUTH);
  request.options = options;
  request.dataNode = NULL;
  request.others = NULL;
  request.freeData = 0;
  request.status = ENCHANT_ERROR; // if nothing happens...
#ifdef USE_THREADS
  pthread_mutex_lock(&requestLock);
  pthread_cond_signal(&requestPending);
#endif
  if (SHOW_QUEUING)
    printf("Enqueue from %s %s\n", request.clientID->name, request.command);
  ArrayAppend(requestQueue, &request);
#ifdef USE_THREADS
  pthread_mutex_unlock(&requestLock);
#endif
}

static void CompleteRequest(RequestType *request)
{
  SendResult(request->clientID, request->dataNode, 
	     request->command, request->npReturn.UNL,
	     request->status, request->others);
  if (request->rcvNode)
    FreeNode(request->rcvNode);
  if (request->freeData && request->dataNode)
    FreeNode(request->dataNode);
  if (request->options)
    {
      ArrayFree(request->options);
    }
  if (request->others)
    free(request->others);
}

static const char *GetRequestOption(RequestType *request, const char *key)
{
  return GetSetOption(request->options, key);
}

static int handleSymlink(RequestType *request)
{
  const char *symUNL;
  if (!request->npReturn.symlinkTarget)
    return 0;
  symUNL = AppendUNLPath(request->npReturn.symlinkTarget,
			 request->npReturn.symlinkCont);
  request->npReturn.UNL = symUNL;
  if (request->npReturn.controlled)
    {
      request->others = malloc(3 * sizeof(char *));
      request->others[0] = "controlled";
      request->others[1] = "t";
      request->others[2] = NULL;
    }
  return 1;
}

static int HandleGet(RequestType *request)
{
  request->npReturn.getSubscriptions = 0;
  GetNodePathInfo(request->clientID, &request->npReturn, 0);
  
  if (handleSymlink(request))
    return ENCHANT_SYMLINK;

  if (!request->npReturn.mayRead)
    return ENCHANT_NOT_EXIST;
  request->dataNode = request->npReturn.existNode;
  return request->npReturn.existNode ? ENCHANT_SUCCESS : ENCHANT_NOT_EXIST;
}

static int HandleGetLock(RequestType *request)
{
  ServerLockType *reportLock;

  request->npReturn.getSubscriptions = 0;
  GetNodePathInfo(request->clientID, &request->npReturn, 0);
  
  if (handleSymlink(request))
    return ENCHANT_SYMLINK;

  reportLock = request->npReturn.reportLock;
  if (!reportLock)
    return ENCHANT_NOT_EXIST;
  request->others = malloc(sizeof(char *) * 5);
  request->others[0] = K_GETLOCK_OWNER;
  request->others[1] = reportLock->owner->name;
  request->others[2] = K_LOCK_TYPE;
  request->others[3] =
    reportLock->remove ? "R" :
    reportLock->recurse ? "r" :
    "n";
  request->others[4] = NULL;
  return ENCHANT_SUCCESS;
}

static int HandleLock(RequestType *request)
{
  int result;
  ServerLockType clientLock;

  const char *locktype = 
    GetRequestOption(request, InstanceString(K_LOCK_TYPE));
  const char *updateflag =
    GetRequestOption(request, InstanceString(K_LOCK_UPDATE));

  request->npReturn.getSubscriptions = 0;
  GetNodePathInfo(request->clientID, &request->npReturn, 1);
  
  if (handleSymlink(request))
    return ENCHANT_SYMLINK;
  
  if (!request->npReturn.targetSchema)
    {
      return ENCHANT_NOT_EXIST;
    }

  if (locktype) // Lock operation
    {
      clientLock.owner = request->clientID;
      clientLock.remove = locktype[0] == 'R';
      clientLock.recurse = clientLock.remove || locktype[0] == 'r';

      if (clientLock.recurse && request->npReturn.existNode &&
	  !CheckDescendantPermission(request->npReturn.existNode,
				     request->clientID, 'l'))
	return ENCHANT_LOCKED;

      clientLock.update = !!updateflag;
	
      result = LockNodePath(request->npReturn.UNL, &clientLock);
    }
  else // Unlock operation
    {
      result = UnlockNodePath(request->npReturn.UNL, request->clientID);
    }
  return result ? ENCHANT_SUCCESS : ENCHANT_LOCKED;
}

static int HandlePublish(RequestType *request)
{
  const char *op;
  const char *opstr;
  int ret;

  NodeType *datanode;

  int is_hard = !!GetRequestOption(request, InstanceString(K_PUBLISH_HARD));
  int publish_if_locked = 
    !GetRequestOption(request, InstanceString(K_DROP_AUTH));

  datanode = request->rcvNode;

  opstr = GetRequestOption(request, InstanceString(K_PUBLISH_OPERATION));
  op = kv_replace;
  switch (opstr ? opstr[0] : '\0')
    {
    case 'm':
      op = kv_merge;
      break;
    case 'a':
      op = kv_addonly;
      break;
    }

  request->npReturn.getSubscriptions = 1;
  GetNodePathInfo(request->clientID, &request->npReturn, 0);
  
  if (handleSymlink(request))
    return ENCHANT_SYMLINK;

  ret = AddNodeToTree(request->clientID, datanode,
		      &request->npReturn, op, is_hard, publish_if_locked);
  if (ret == ENCHANT_SUCCESS)
    request->rcvNode = NULL;
  return ret;
}

static int HandleSubscribe(RequestType *request)
{
  int id;
  int self_subscribe = !!GetRequestOption(request, 
					  InstanceString(K_SUBSCRIBE_SELF));

  request->npReturn.getSubscriptions = 0;
  GetNodePathInfo(request->clientID, &request->npReturn, 1);
  
  if (handleSymlink(request))
    return ENCHANT_SYMLINK;
  
  id = Subscribe(request->clientID, &request->npReturn, self_subscribe);
  
  request->dataNode = request->npReturn.existNode;
  if (!id)
    {
      return ENCHANT_NO_EFFECT;
    }
  request->others = malloc(sizeof(char *) * 3);
  request->others[0] = "id";
  request->others[1] = InstanceNumber(id);
  request->others[2] = NULL;
  return ENCHANT_SUCCESS;
}

static int HandleUnsubscribe(RequestType *request)
{
  SubscriptionType *subscription;
  const char *option;

  request->npReturn.getSubscriptions = 0;
  option = GetRequestOption(request, InstanceString("id"));
  if (option)
    {
      int i = 0;
      int subID = atoi(option);
      if (request->clientID->subscribedNodes)
	for (i = 0; i < ArrayLength(request->clientID->subscribedNodes); i++)
	  {
	    SubscriptionType *sub = *(SubscriptionType **) 
	      ArrayNth(request->clientID->subscribedNodes, i);
	    if (sub->id == subID)
	      {
		subscription = sub;
		break;
	      }
	  }
    }
  else
    {
      GetNodePathInfo(request->clientID, &request->npReturn, 0);
      if (handleSymlink(request))
	return ENCHANT_SYMLINK;
      subscription = IsSubscribed(request->clientID, request->npReturn.UNL);
    }

  if (subscription)
    {
      request->others = malloc(sizeof(char *) * 3);
      request->others[0] = "id";
      request->others[1] = InstanceNumber(subscription->id);
      request->others[2] = NULL;
      //printf("Unsubscribing from %s\n", subscription->UNL);
      SendUpdate(subscription, NULL, NULL, NULL, NULL, NULL);
      ForgetSubscription(request->clientID, subscription);
      ForgetSubscriber(subscription);
      return ENCHANT_SUCCESS;
    }
  return ENCHANT_NO_EFFECT;
}

static int HandleGetMeta(RequestType *request)
{
  const char *metapath = GetRequestOption(request, InstanceString("metapath"));
  if (!metapath)
    return ENCHANT_ERROR;
  request->npReturn.getSubscriptions = 0;
  GetNodePathInfo(request->clientID, &request->npReturn, 1);
  request->dataNode = GetMetadataTree(request->npReturn.targetSchema,
				      metapath);
  request->freeData = 1;
  return ENCHANT_SUCCESS;
}

static int HandleSetMeta(RequestType *request)
{
  SchemaNode *schema;
  const char *metapath = GetRequestOption(request, InstanceString("metapath"));
  if (!metapath)
    return ENCHANT_ERROR;
  request->npReturn.getSubscriptions = 0;
  GetNodePathInfo(request->clientID, &request->npReturn, 1);
  schema = request->npReturn.targetSchema;
  if (!CheckPermission(schema, request->clientID, 'a'))
    return ENCHANT_LOCKED;
  return SetMetadataTree(schema, metapath, NULL, request->rcvNode);
}

static void ProcessRequest(RequestType *request)
{
  NodeType *rcvNode = request->rcvNode;
  const char *command;
  int status;
  if (!request->command)
    {
      plogf("Client %s done\n", request->clientID->name);
      SocketClose(request->clientID->socket);
      RemoveClient(request->clientID);
      return;
    }
  command = request->command;
  
  if (SHOW_REQUESTS)
    {
      printf("%s:\n", request->clientID->name);
      PrintNode(rcvNode, NULL);
    }
  plogf("Request from %s:\n", request->clientID->name);
  logTree();
  if (c_get == command)
    {
      status = HandleGet(request);
    }
  else if (c_getlock == command)
    {
      status = HandleGetLock(request);
    }
  else if (c_lock == command)
    {
      status = HandleLock(request);
    }
  else if (c_publish == command)
    {
      status = HandlePublish(request);
    }
  else if (c_subscribe == command)
    {
      status = HandleSubscribe(request);
    }
  else if (c_unsubscribe == command) 
    {
      status = HandleUnsubscribe(request);
    }
  else if (c_getmeta == command)
    {
      status = HandleGetMeta(request);
    }
  else if (c_setmeta == command)
    {
      status = HandleSetMeta(request);
    }
  else
    {
      if (CLIENT_ERROR) 
	printf("Unknown type of request '%s' received!\n", command);
      status = ENCHANT_ERROR;
    }
  request->status = status;
  CompleteRequest(request);
}

void ProcessRequests()
{
  int done = 0;
  //printf("%d requests enqueued\n", ArrayLength(requestQueue));
  while (!done)
    {
      RequestType request;
#ifdef USE_THREADS
      pthread_mutex_lock(&requestLock);
#endif
      if (ArrayLength(requestQueue))
	{
	  request = *(RequestType *)ArrayNth(requestQueue, 0);
	  ArrayDeleteAt(requestQueue, 0);
	}
      else
	done = 1;
#ifdef USE_THREADS
      if (SHOW_QUEUING)
	{
	  if (done)
	    printf("None\n");
	  else
	    printf("Dequeue from %s %s\n", request.clientID->name,
		   request.command);
	}
      pthread_mutex_unlock(&requestLock);
#endif
      if (!done)
	ProcessRequest(&request);
    }
}

#ifdef USE_THREADS
static void *TreeManagementLoop(void *data)
{
  do
    {
      if (SHOW_QUEUING)
	printf("Waiting for requests\n");
      pthread_mutex_lock(&requestLock);
      pthread_cond_wait(&requestPending, &requestLock);
      pthread_mutex_unlock(&requestLock);
      if (SHOW_QUEUING)
	printf("got requests\n");
      ProcessRequests();
    }
  while (1);
}

void StartRequestThread()
{
  pthread_t thread;
  pthread_create(&thread, NULL, TreeManagementLoop, NULL);
}
#endif
