/*************************************************
 * File: tree-manip.h
 * Birth Date: Dec 3, 2003
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2002 MIT Media Lab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *************************************************/

#ifndef _TREE_MANIP_H
#define _TREE_MANIP_H

#include <enchantment/node.h>
#include "features.h"
#include <enchantment/darray.h>
#include "socket.h"

typedef struct {
  const char *name;

  Socket *socket;
  DArray subscribedNodes;
  DArray locks;

  unsigned nextID;

  DArray auth;
} ClientIDType;

typedef struct {
  int self : 1;
  unsigned id;
  ClientIDType *subscriber;
  const char *UNL;
  struct SchemaNode *schema;
} SubscriptionType;

typedef struct
{
  ClientIDType *owner;
  unsigned recurse : 1;
  unsigned remove : 1; // when the client closes, remove the node; req. RECURSE
  unsigned update : 1; // send update and no error on publication
} ServerLockType;

typedef struct {
  ClientIDType *owner;
  unsigned short count;
} LockTrailType;

typedef struct 
{
  const char *name;
  struct SchemaNode *node;
} SchemaChild;

typedef struct 
{
  const char **tokens;
} ClientSelector;

typedef struct SchemaNode
{
  // Node currently here
  NodeType *currentNode;

  // Fixed type, if any
  TypeID type;

  // Symlink, if any
  const char *symlink;

  ClientSelector canRead;
  ClientSelector canWrite;
  ClientSelector canLock;
  ClientSelector canAdmin;

  ServerLockType lock;
  DArray lockTrails;

  DArray subscribers; // of SubscriptionType

  struct SchemaNode *parent;
  DArray children; /* For branch and such nodes */

  // Something that the server handles specially
  unsigned special : 1;

  // True if publications to this UNL should be redirected to children
  unsigned distinct : 1;

  // True if there is an implicit node currently here
  unsigned implicitNode : 1;

  // True if the node here is hard published
  unsigned hardNode : 1;

  // True if the schema should be maintained
  unsigned hardSchema : 1;
} SchemaNode;

typedef struct
{
  NodeType *treeParent;

  struct SchemaNode *schema;
} PathTreeDataType;

typedef struct 
{
  const char *UNL;

  // The node at the given UNL
  NodeType *existNode;

  // If there is a node at the given UNL, the parent of that node
  NodeType *existParent;

  // The closest ancestor of UNL which exists; parent if the node exists.
  const char *closestUNL;
  NodeType *closestAncestorNode;

  // The client which the permission information is valid for.
  ClientIDType *clientID;

  // The schema at the UNL
  SchemaNode *targetSchema;

  // The extant schema closest to the target UNL
  SchemaNode *closestSchema;
  // The UNL of the schema closest to the target UNL
  char *closestSchemaUNL;

  // True if existNode is controlled by the client.
  unsigned controlled : 1;

  unsigned cross_symlinks : 1;
  unsigned follow_symlinks : 1;
  unsigned drop_auth : 1;
  unsigned getSubscriptions : 1;

  // The lock to talk about
  ServerLockType *reportLock;

  unsigned mayCreateRemove : 1;
  unsigned mayModify : 1;
  unsigned mayRead : 1;

  // The list of subscriptions to existing ancestors of the UNL; that
  // is, the subscribers who should get updates but are not subscribed
  // to a UNL which will be linked/unlinked in a publication.
  DArray subscriptions;
  
  const char *symlinkTarget;
  const char *symlinkCont;
} NodePathReturnType;

int FitsSelector(ClientSelector *selector, ClientIDType *client);
void SelectorNone(ClientSelector *selector);
void SelectorAny(ClientSelector *selector);
void SelectorSome(ClientSelector *selector, ...);
void SetPermissions(SchemaNode *schema, char type, ClientSelector *selector);
void SetNodePermissions(NodeType *node, char type, ClientSelector *selector);

void SetFixedType(SchemaNode *schema, TypeID type);

int LocalSetPathPermissions(const char *path, char type, 
			    ClientSelector *selector);
int CheckPermission(SchemaNode *node, ClientIDType *client, char type);
int CheckNodePermission(NodeType *node, ClientIDType *client, char type);
int CheckDescendantPermission(NodeType *node, ClientIDType *client, char type);

int CheckReplace(NodePathReturnType *npReturn, NodeType *node);
int CheckRecursiveLock(NodePathReturnType *npReturn);

NodeType *GetMetadataTree(SchemaNode *node, const char *metapath);
int SetMetadataTree(SchemaNode *node, const char *metapath, char *op,
		    NodeType *value);

void GetNodePathInfo(ClientIDType *clientID,
		     NodePathReturnType *p_npReturn, 
		     int createSchema);

int AddNodeToTree(ClientIDType *clientID, NodeType *dataNode,
		  NodePathReturnType *npReturn, const char *op,
		  int hard, int publish_if_locked);

int LockNodePath(const char *path, ServerLockType *lock);
int UnlockNodePath(const char *path, ClientIDType *owner);

/*** Subscriptions ***/

/* Remove subscription from subscriber's list. */
void ForgetSubscription(ClientIDType *subscriber, SubscriptionType *sub);

/* Remove subscription from a subscribed node's list. Always frees the
 * subscription record. 
 */
void ForgetSubscriber(SubscriptionType *subscription);

/* Subscribe the given client to the given node. Returns FALSE if
 * the client was already subscribed.
 */
int Subscribe(ClientIDType *clientID, NodePathReturnType *npReturn, 
	      int SELF_SUBSCRIBE);

/* Returns the subscription, if any, that the given client has to the
 * given UNL. */
SubscriptionType *IsSubscribed(ClientIDType *clientID, const char *UNL);

void PublishLocal(const char *path, NodeType *insertNode, int hard);

void RemoveClientTreeEffects(ClientIDType *clientID);

extern NodeType *g_globalNode;

#define InTree(node) (!!(node)->local)
#define IsSpecial(node) (((PathTreeDataType *) (node)->local)->schema->special)
#define SchemaOf(node) (((PathTreeDataType *) (node)->local)->schema)

#define ParentOf(node) (((PathTreeDataType *) (node)->local)->treeParent)
#define IsImplicit(node) (((PathTreeDataType *) (node)->local)->schema->implicitNode)

#endif
