/*********************************************************
 * File: tree.h
 * Author: Jonathan Gips
 * Birth Date: 6/25/01
 *
 * Project: Enchantment IPC
 *
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 *
 * Copyright (C) 2002 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 ***********************************************************/

#ifndef ENCHANTMENT_NODE_H
#define ENCHANTMENT_NODE_H

#include <enchantment/types.h>
#include <stdio.h>

typedef struct 
{
  char *name;
  TypeID type;
  DataType data;

  void *local; // extra information about the node
} NodeType;

/* 
 * Returns a new node of the specified name and type, with the given
 * value. 
 */
NodeType *CreateTypedNode(const char *name, TypeID type, DataType *value);

/* Returns a new empty node with the given name and type. */
NodeType *CreateEmptyNode(const char *name, TypeID type);

/* 
 * Returns a new branch node with the given name and type, with no
 * initial contents.  
 */
NodeType *CreateBranchNode(const char *name, TypeID type);

/* 
 * The following functions return new nodes with the given name, type,
 * and value.  
 */
NodeType *CreateTimeNode(const char *name, TypeID type, struct timeval *value);
NodeType *CreateAddressNode(const char *name, TypeID type, 
			    NetSocketLocation *location);
NodeType *CreateStringNode(const char *name, TypeID type, char *value);
NodeType *CreateIntegerNode(const char *name, TypeID type, long value);
NodeType *CreateFloatNode(const char *name, TypeID type, double value);

/* Returns the double value of a T_NUMERIC node */
double NodeGetDouble(NodeType *node);

/* Sets the value of a T_STRING node, taking care of memory issues. */
void NodeSetString(NodeType *node, char *value);

/* Frees the given node, including all of the contents. */
void FreeNode(NodeType *node);

/* 
 * Returns the child of the given parent with the given name, or NULL
 * if there is no such child or the parent is not a branch.
 * NOTE: The name must be a dstring. 
 */
NodeType *GetChild(NodeType *parentNode, char *childName);

/* 
 * Returns the number of children that the node has. Returns 0 if it
 * is not a branch. 
 */
unsigned NumChildren(NodeType *node);

/* Returns the (0-based) indexed child of the given branch. Returns
 * NULL if there are fewer children or if the node is not a branch. */
NodeType *BranchIndex(NodeType *branch, unsigned index);

/* 
 * Inserts the given node into the given parent. Returns true if it
 * was successful, and false if parent is not a branch. If there is
 * already a child with the same name in the parent, it will be
 * freed. Returns true if the node is now in the parent. 
 */
int InsertNode(NodeType *parent, NodeType *node);

/* 
 * Removes the given node from the given parent. Returns true if the
 * parent previously contained it. 
 */
int RemoveNode(NodeType *parent, NodeType *node);

/* Returns a completely-independant copy of srcNode. */
NodeType *CopyNode(NodeType *srcNode);

/* 
 * Prints the given node to the given file. If the file is NULL,
 * prints to stdout. 
 */
void PrintNode(NodeType *node, FILE *file);

/* 
 * Prints the given node to the given file. If the file is NULL,
 * prints to stdout. Calls test with each node before printing it. 
 */
void PrintNodeFilter(NodeType *node, FILE *file, int (*test)(NodeType *node));

/* 
 * Prints just the value of the given node to the given file. If the
 * file is NULL, prints to stdout. 
 */
void PrintValue(NodeType *node, FILE *file);

/* 
 * TraverseTree recursively traverses the nodes in the given node,
 * calling pre before handling children and post after. Either
 * function may be NULL, in which case it has no effect. Both
 * functions are called for nodes which are not branches. If UNL is
 * provided, it is the UNL of node, and UNLs will be provided to the
 * called functions.
 *
 * TravTreeFN is called with the data value given to TraverseTree, the
 * node being traversed, the parent of the node being traversed, and
 * the UNL of the node being traversed (if applicable). The parent is
 * NULL for the node given to TraverseTree. If the return value is
 * false, TraverseTree aborts. If the return value is 2 (for pre), the
 * children of the node are skipped and post is not called, but other
 * children of the ancestors of the node will be traversed. 
 */
typedef int (*TravTreeFN)(void *data, NodeType *node, 
			  NodeType *parent, char *UNL);

int TraverseTree(NodeType *node, NodeType *parent, void *data, 
		 TravTreeFN pre, TravTreeFN post, char *UNL);

#endif
