/*************************************************
 * File: whiteboard.h
 * Author: Jonathan Gips
 * Birth Date: 6/27/01
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 *************************************************/

#ifndef ENCHANTMENT_WHITEBOARD_H
#define ENCHANTMENT_WHITEBOARD_H

#include <enchantment/types.h>
#include <enchantment/node.h>

#include <sys/types.h>

/*
 * Uses encryption for all future connections, when available. Note
 * that it does not affect existing connections, and that servers
 * which do not support encryption will still be accessed with an
 * insecure connection. Use SecureUNL() to determine whether a UNL
 * will be accessed securely. Returns false if encryption is not
 * available.
 */

int AttemptEncryption();

/*
 * Sets the server that will be used for UNLs without a server
 * portion. 
 */
void SetDefaultServer(char *server);

/*
 * Only looking at the server portion of the UNL, reports whether that
 * UNL is accessed over a secure connection. Note that, if there is a
 * symlink at the UNL, and it goes to a different server, the
 * connection to the other server may not be secure even if this
 * returns true. Returns false if no connection could be made.  
 */
int SecureUNL(char *unl);

/***************************************************************************
 * Whiteboard operations
 *
 * Whiteboard operations affect or query a whiteboard server's state
 * for a UNL. They may be used either with the default optional
 * behavior or with specified flags and extra return information.
 *
 * Calling an operation with NULL for the UNL and with an appropriate
 * info structure will fill out the default values for the
 * structure. This allows code to get the default behavior while
 * requesting the extra return information, and to specify only a
 * subset of the options.
 */

/* The common information for whiteboard operations. */
typedef struct
{
  // request information
  unsigned commonFlags : 4;    // flags with apply to multiple operations
  unsigned operationFlags : 4; // flags for a particular operation

  // response information
  char *actualUNL;             // the actual UNL affected
  char *errorCode;             // the server's response code
} WhiteboardInfo;

/* Flags for commonFlags */
#define ENCHANT_IGNORE_LINKS  0 // treat all symlinks as normal nodes
#define ENCHANT_CROSS_LINKS   1 // follow links in path, but not final node
#define ENCHANT_FOLLOW_LINKS  2 // follow all links

/****************************  GetNode  *******************************
 * Gets a node from the global tree by UNL. Returns NULL on error.
 *
 * The default flags are: ENCHANT_FOLLOW_LINKS.
 */
typedef struct 
{
  WhiteboardInfo common;
} GetNodeInfo;

NodeType *GetNode(char *UNL, GetNodeInfo *info);

/*************************  PublishNode  ******************************
 * Publishes a node to the specified UNL.
 *
 * The default flags are: ENCHANT_CROSS_LINKS,
 * ENCHANT_PUBLISH_REPLACE, and ENCHANT_PUBLISH_SOFT.
 */

#define ENCHANT_PUBLISH_REPLACE 0 // replace the contents entirely
#define ENCHANT_PUBLISH_MERGE   1 // add and replace, but don't remove
#define ENCHANT_PUBLISH_ADDONLY 2 // only add nodes

#define ENCHANT_PUBLISH_HARD    4 // publish to persistant state
#define ENCHANT_PUBLISH_SOFT    0 // publish to temporary state

typedef struct
{
  WhiteboardInfo common;
} PublishNodeInfo;

int PublishNode(char *UNL, NodeType *node, PublishNodeInfo *info);

/*************************  LockUNL  **********************************
 * Locks a UNL, such that the client has temporary control over what
 * appears there.
 *
 * The default flags are: ENCHANT_CROSS_LINKS and ENCHANT_LOCK_NORMAL.
 */
#define ENCHANT_LOCK_NONE      0 // unlock unl
#define ENCHANT_LOCK_NORMAL    1 // lock unl
#define ENCHANT_LOCK_RECURSIVE 2 // lock unl and children
#define ENCHANT_LOCK_REMOVE    3 // lock unl and children, unpublish on exit

typedef struct 
{
  WhiteboardInfo common;
} LockUNLInfo;

int LockUNL(char *UNL, LockUNLInfo *info);

/********************  GetUNLLock  *********************************
 * Returns information about the lock on the given UNL, if any.
 *
 * Default flags are ENCHANT_CROSS_LINKS.
 */

typedef struct
{
  WhiteboardInfo common;
  unsigned lockType : 2;
  char *lockOwner;
} GetUNLLockInfo;

int GetUNLLock(char *unl, GetUNLLockInfo *info);

/**********************  SubscribeUNL  *****************************
 * Subscribe to changes at a UNL.
 *
 * The default flags are: ENCHANT_FOLLOW_LINKS,
 * ENCHANT_SUBSCRIBE_OTHERS, and ENCHANT_SUBSCRIBE_NEW.
 */
typedef struct
{
  WhiteboardInfo common;
} SubscribeUNLInfo;

#define ENCHANT_SUBSCRIBE_OTHERS  0 // only get publications by others
#define ENCHANT_SUBSCRIBE_SELF    1 // also get publications by oneself

#define ENCHANT_SUBSCRIBE_NEW     0 // only get new publications
#define ENCHANT_SUBSCRIBE_INITIAL 2 // also get the current value

int SubscribeUNL(char *UNL, SubscribeUNLInfo *info);

/*******************  UnsubscribeUNL  ******************************
 * Stops the subscription at the UNL.
 *
 * The default flags are: ENCHANT_FOLLOW_LINKS.
 */
typedef struct
{
  WhiteboardInfo common;
} UnsubscribeUNLInfo;

int UnsubscribeUNL(char *UNL, UnsubscribeUNLInfo *info);

/******************  GetUpdate / WaitForUpdate  ********************
 * Returns the information about a publication that this client
 * subscribed to be notified of. The UNL is where the publication
 * happened (or where the publication began to affect the subscribed
 * subtree). The node is the current value of the UNL subscribed to
 * (NULL if it was unpublished). If a UNL is provided to the function,
 * it will only return updates for subscriptions to that UNL, skipping
 * over non-matching updates.
 *
 * GetUpdate will check for an update immediately available or already
 * arrived, while WaitForUpdate will block until an update arrives.
 *
 * WaitForInfo will accept NULL for info, and will return the
 * node. GetUpdate does not support this behavior, because it needs to
 * distinguish an unpublication from no update.
 */
typedef struct
{
  char *UNL;
  char *publisher;
  NodeType *node;
} GetUpdateInfo;

void GetUpdate(GetUpdateInfo *info, char *UNL);

NodeType *WaitForUpdate(GetUpdateInfo *info, char *UNL);

/*
 * Call when done with the client IPC. This function shuts down all
 * active server connections and frees up library memory. Note that
 * there is no reason to do this unless you want to close all of the
 * connections but have the client continue running.
 */
void ShutDownClientIPC();

/*
 * Call with the verbosity level of messages to stderr desired. 0
 * gives no output at all.
 */
void WhiteboardVerbosity(int level);

#endif
