/*************************************************
 * File: whiteboard.c
 * Author: Jonathan Gips
 * Birth Date: 7/02/01
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002, 2003, 2004 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 *************************************************/

#include "protocol.h"
#include "introduction.h"

#include <enchantment/whiteboard.h>
#include <enchantment/dstring.h>
#include <enchantment/unl.h>
#include "ipc.h"

#include "whiteboard-protocol.h"

#include "socket.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <enchantment/identity.h>

#define USE_THREADS yes

#ifdef USE_THREADS
#include <pthread.h>
#include <semaphore.h>
#define lock(name) pthread_mutex_lock(&(name))
#define unlock(name) pthread_mutex_unlock(&(name))
#else
#define lock(name)
#define unlock(name)
#endif

#define SERVER_ERROR     1
#define CLIENT_ERROR     1

#define SERVER_ENV_VAR "ENCHANTHOST"

#define FOLLOW_SYMLINK (1 << 5)

typedef struct
{
  char *callType;
  char *returnUNL;
  char *result;
  NodeType *dataNode;
  int flags;
} returnInfo;

typedef struct
{
  char *name;
#ifdef USE_THREADS
  pthread_mutex_t requestLock;
  pthread_mutex_t ioLock;
  pthread_cond_t statusWait;
#endif
  returnInfo status;
  Socket *socket;
} ServerInfo;

/* Locking requirements
 *
 *  Only one thread may alter the connection lists at a time.
 *
 *  Only one thread may make a request to a server at a time, and no
 *  other thread may make a request to that server until the return
 *  info is provided.
 *
 *  Only one thread may read from a server at a time, including getting
 *  updates.
 *
 * Locking design
 *
 *  There is a lock for the connection lists, required for all access to them.
 *
 *  There is a lock for the update queue, required for all access.
 *
 *  There is a lock for each connection, required to send data to it.
 *
 *  There is a second lock for each connection, required to read from it.
 *
 *  You may not acquire the write lock while holding the read lock.
 *
 *  You may not hold the connection list lock while taking another lock.
 *
 *  ???? Update/other?
 */

static int ipcInit = 0;

#ifdef USE_THREADS
static pthread_mutex_t connectionLock;
static pthread_mutex_t updateLock;
static pthread_mutex_t updateQueueLock;
#endif

static DArray servArray = NULL;
static DArray sockArray = NULL;

static unsigned updateSequence = 0;
static unsigned updatesSent = 0;
static DArray updateQueue = NULL; // of UpdateInfo

static int quiet_client_ipc = 0;

static int kUNLArrayInitSize = 10;
// Global String Constants
static char *s_error = NULL, *s_return = NULL,
  *sv_does_not_exist = NULL, *sv_bad_unl = NULL, *sv_bad_request = NULL,
  *k_unl = NULL, *s_update = NULL, *s_terminal = NULL;
static char *k_property_encrypt = NULL;
static char *defaultServer = NULL;

#define DEBUG_UPDATE_QUEUE 0

typedef struct
{
  char *publisher;
  char *server;
  char *path;
  NodeType *node;
} UpdateInfo;

static void initializeData()
{
  if (ipcInit)
    return;
  s_error = InstanceString(S_ERROR);
  sv_does_not_exist = InstanceString(SV_DOES_NOT_EXIST);
  sv_bad_unl = InstanceString(SV_BAD_UNL);
  sv_bad_request = InstanceString(SV_BAD_REQUEST);
  s_return = InstanceString(S_RETURN);
  k_unl = InstanceString(K_UNL);
  k_property_encrypt = InstanceString(K_PROPERTY_ENCRYPT);
  s_update = InstanceString(S_UPDATE);
  s_terminal = InstanceString(S_TERMINAL);
  sockArray = ArrayNew(sizeof(Socket*), 3, NULL);
  servArray = ArrayNew(sizeof(ServerInfo *), 3, NULL);
  updateQueue = ArrayNew(sizeof(UpdateInfo *), 5, NULL);
#ifdef USE_THREADS
  pthread_mutex_init(&updateQueueLock, NULL);
  pthread_mutex_init(&connectionLock, NULL);
#endif
  ipcInit = 1;
  if (!defaultServer)
    {
      defaultServer = getenv(SERVER_ENV_VAR);
      if (!defaultServer)
	defaultServer = "localhost";
    }
}

/**** Update queue management ****/

static int EnqueueUpdate(char *server, char *publisher, NodeType *node,
			 char *UNL)
{
  UpdateInfo *info = malloc(sizeof(UpdateInfo));
  info->server = server;
  info->path = UNL;
  info->node = node;
  info->publisher = publisher;
  if (DEBUG_UPDATE_QUEUE)
    printf("Enqueue %p at %s:%s\n", node, info->server, info->path);

  lock(updateQueueLock);
  updateSequence++;
  ArrayAppend(updateQueue, &info);
  unlock(updateQueueLock);

  return 1;
}

/**** Reading data from servers ****/

/* We need to get at least one message from this socket. 
 * Called with ioLock.
 */
static int doUnload(ServerInfo *server, returnInfo *sync)
{
  NodeType *rcvNode = NULL;
  Socket *socket = server->socket;
  char *buffer;
  size_t size;
  char *unl = NULL;
  char *publisher = NULL;
  char *type = NULL;
  char *command = NULL;
  char *result = NULL;

  if (sync)
    {
      if (server->status.result)
	{
	  *sync = server->status;
	  server->status.result = NULL;
	  return 1;
	}
    }
  else
    {
      sync = &server->status;
    }
  do 
    {
      if (!SocketRecvRecord(socket, &buffer, &size))
	{
	  fprintf(stderr, "Error reading record from server.\n");
	  exit(1);
	}
      if (size)
	{
	  if (!strcmp(buffer, "node"))
	    {
	      ReadNode(socket, &rcvNode, NULL, NULL);
	    }
	  else
	    {
	      char *value = NULL;
	      size_t size;
	      ReadString(socket, &value, &size);
	      if (!strcmp(buffer, "unl"))
		{
		  unl = value;
		}
	      else if (!strcmp(buffer, "publisher"))
		{
		  publisher = value;
		}
	      else if (!strcmp(buffer, "type"))
		{
		  type = InstanceString(value);
		  free(value);
		}
	      else if (!strcmp(buffer, "command"))
		{
		  command = InstanceString(value);
		  free(value);
		}
	      else if (!strcmp(buffer, "result"))
		{
		  result = InstanceString(value);
		  free(value);
		}
	    }
	}
    } 
  while (size);

  // Look at received node type and do appropriate action
  if (type == s_return)
    {
      sync->dataNode = rcvNode;
      sync->returnUNL = unl;
      sync->callType = command;
      sync->result = result;
    }
  else if (type == s_update)
    {
      EnqueueUpdate(server->name, publisher, rcvNode, unl);
    }
  else if (type == s_terminal || type == NULL)
    {
      return 0;
    }	      
  else
    {
      printf("doUnload::Unknown type of node received from server (%s).\n",
	     type);
      FreeNode(rcvNode);
    }
  return 1;
}

/* Called under requestLock */
static int UnloadResponse(ServerInfo *server, returnInfo *sync)
{
  sync->callType = NULL;
  while (1)
    {
      lock(server->ioLock);
      if (!doUnload(server, sync))
	{
	  unlock(server->ioLock);
	  return 0;
	}
      unlock(server->ioLock);
      if (sync->callType)
	return 1;
    }
}

/* If blocking is true, do not return until a new update is available. */
static void UnloadUpdates(int blocking, int updateSequenceSeen)
{
  int i;
  fd_set socketSet;

  // Only one thread may unload at a time
  lock(updateLock);

  // Another thread unloaded something.
  if (updateSequenceSeen != updateSequence)
    {
      unlock(updateLock);
      return;
    }

  // XXXX need to deal with connectionLock
  // XXXX need to account for the possibility of the list changing due to
  //  another thread.
  if (blocking)
    {
      SocketSelect(&socketSet, sockArray, NULL);
    }

  for (i = 0; i < ArrayLength(sockArray); i++)
    {
      ServerInfo *server = *((ServerInfo **) ArrayNth(servArray, i));
      
      // Only look at those fds that were set
      if (blocking && !SocketIncluded(&socketSet, server->socket))
	continue;

      lock(server->ioLock);
      // Loop until sock is empty
      while (SocketReadReady(server->socket, 0))
	{
	  doUnload(server, NULL);
	}
      unlock(server->ioLock);
    }
  unlock(updateLock);
}

// NULL on local failure or bad return value. Valid on other values.
static char *HandleSyncReturn(ServerInfo *server, char *syncCallName, 
			      char *UNL, 
			      NodeType **p_dataNode, int *flags)
{  
  returnInfo info;

  /*********** Get the Synchronous Return **************/
  
  if (!UnloadResponse(server, &info))
    return NULL;

  //printf("<- %s %s %s\n", info.callType, info.returnUNL, info.result);
  
  if (strcmp(info.callType, syncCallName))
    {
      if (SERVER_ERROR) 
	printf("Wrong type of call returned in result: %s should be %s.\n",
	       info.callType, syncCallName);
      
      return NULL;
    }
  if (!strcmp(info.result, SV_BAD_REQUEST))
    {
      if (CLIENT_ERROR)
	printf("Bad request returned for %s on %s.\n", syncCallName, UNL);
      return NULL;
    }
  
  if (p_dataNode)
    *p_dataNode = info.dataNode;
  else
    FreeNode(info.dataNode); //added 3/4/02
  
  if (info.returnUNL) 
    free(info.returnUNL);
  if (flags)
    *flags = info.flags;
  return info.result;
}

/* Called with requestLock */
static int SendRequest(ServerInfo *server, char *command, NodeType *request, 
		       char *unl, unsigned char flags)
{
  Socket *socket = server->socket;
  lock(server->ioLock);
  //printf("-> %s %s\n", command, unl);
  if (unl)
    {
      SocketSendRecord(socket, "unl", 3);
      WriteString(socket, unl, strlen(unl));
    }
  if (flags & F_LOWBIT0)
    {
      SocketSendRecord(socket, "lowbit-0", 8);
      WriteString(socket, "t", 1);
    }
  if (flags & F_LOWBIT1)
    {
      SocketSendRecord(socket, "lowbit-1", 8);
      WriteString(socket, "t", 1);
    }
  if (flags & F_LOWBIT2)
    {
      SocketSendRecord(socket, "lowbit-2", 8);
      WriteString(socket, "t", 1);
    }
  if (flags & F_LOWBIT3)
    {
      SocketSendRecord(socket, "lowbit-3", 8);
      WriteString(socket, "t", 1);
    }
  if (flags & FOLLOW_SYMLINK)
    {
      SocketSendRecord(socket, "follow-links", 12);
      WriteString(socket, "t", 1);
    }
  SocketSendRecord(socket, "command", 7);
  WriteString(socket, command, strlen(command));
  if (request)
    {
      SocketSendRecord(socket, "node", 4);
      WriteNode(socket, request, NULL, NULL);
    }
  SocketSendRecord(socket, NULL, 0);
  SocketFlush(socket);
  unlock(server->ioLock);
  return 1;
}

/* Runs under connectionLock */
static void DisconnectFromServer(int i)
{
  SocketClose((*((ServerInfo **) ArrayNth(servArray, i)))->socket);
  free(*((ServerInfo **) ArrayNth(servArray, i)));
  ArrayDeleteAt(servArray, i);
  ArrayDeleteAt(sockArray, i);
}

void SetDefaultServer(char *server)
{
  defaultServer = InstanceString(server);
}

void ShutDownClientIPC()
{
  // If we haven't initialized, we don't need to do anything.
  if (!ipcInit) 
    return;
  
  // Close down the socket connections and free memory
  // We leave this locked, since the lock will be reinitialized if needed.
  lock(connectionLock);
  if (servArray)
    {
      while (ArrayLength(servArray) > 0)
	{
	  DisconnectFromServer(0);
	}
      
      ArrayFree(servArray); // Deallocate the socket array
      ArrayFree(sockArray);
      servArray = NULL;
      sockArray = NULL;
    }
  
  // XXXX Free the update queue and its nodes

#if SOCK_STATS
  if (!quiet_client_ipc)
    PrintSockStats();
#endif 
  ipcInit = 0;
}

char *NicknameUNL(char *nickname)
{
  char *temp = strchr(nickname, '@');
  char *ret;
  char *branchName = InstanceStringN(nickname, temp - nickname);
  ret = CreateUNL(temp + 1, "/servers");
  ret = AppendUNLPath(ret, branchName);
  return ret;
}

char *GetServerName(char *UNL)
{
  char *ret = ExtractUNLServer(UNL);
  if (!ret)
    ret = defaultServer;
  return ret;
}

/**** Getting sockets ****/

static int ServerLocationByName(NetSocketLocation **locPtr, char *serverName)
{
  int port = DEFAULT_PORT;
  char *hostname;
  char *temp;
  NetSocketLocation *target = NULL;
  
  if ((temp = strchr(serverName, '@')))
    {
      port = atoi(serverName);
      if (port)
	hostname = temp + 1;
      else
	{
	  char *serverUNL = NicknameUNL(serverName);
	  NodeType *serverNode = GetNode(serverUNL, NULL);
	  if (serverNode && (serverNode->type == T_SERVER))
	    {
	      target = serverNode->data.netsock_val;
	    }
	  else
	    {
	      printf("Didn't like server name '%s'\n", serverName);
	      return 0;
	    }
	}
    }
  else
    hostname = serverName;
  if (!target)
    {
      target = CreateNetSocketLocation(hostname, port);
      if (!target)
	return 0;
    }
  *locPtr = target;
  return 1;
}

static int GetSeparateSocket(Socket **sockPtr, char *serverName)
{
  NetSocketLocation *target = NULL;
  int i;

  serverName = InstanceString(serverName);

  lock(connectionLock);
  for (i = 0; i < ArrayLength(servArray); i++) 
    {
      ServerInfo *server = *((ServerInfo **) ArrayNth(servArray, i));
      if (server->name == serverName)
	{
	  target = SocketRemote(server->socket);
	}
    }
  unlock(connectionLock);

  if (!target)
    {
      if (!ServerLocationByName(&target, serverName))
	return 0;
    }
  *sockPtr = SocketConnect(target, 1);
  if (0) // XXXX
    free(target);

  return (!!*sockPtr);
}

int GetSignalProxySocket(Socket **sockPtr, char *UNL)
{
  Socket *sock;
  char *server;

  if (!ipcInit)
    initializeData();

  server = GetServerName(UNL);
  if (!GetSeparateSocket(&sock, server))
    return 0;
  *sockPtr = sock;
  return 1;
}

static void negotiateClient(Socket *socket)
{
  IntroMessage message;
  int ret;
  int did_type = 0;
  int done;
  int use_ssl = 0;
  int using_ssl = 0;
  do
    {
      done = 1;
      if (!did_type)
	{
	  message.format = &protocolInteraction;
	  message.value.string_val = PROTOCOL_TYPE_WHITEBOARD;
	  IntroSend(socket, &message);
#ifdef WANT_SSL
	  message.format = &protocolEncrypt;
	  message.value.string_val = PROTOCOL_ENCRYPT_SSL;
	  IntroSend(socket, &message);
#endif
	  did_type = 1;
	  done = 0;
	}
      
      IntroDone(socket);

      if (!done)
	{
	  done = 1;

	  do
	    {
	      ret = IntroRecv(socket, serverCommonProperties, &message);
	      if (ret == -1)
		;
	      else if (message.format == &protocolEncrypt)
		{
#ifdef WANT_SSL
		  // In this case, we asked for it previously
		  use_ssl = 1;
#endif
		}
	      if (ret)
		{
		  done = 0;
		}
	      IntroFree(&message);
	    }
	  while (ret > 0);
	}
      if (use_ssl && !using_ssl)
	{
	  SocketSecure(socket, NULL, 0);
	  using_ssl = 1;
	}
    }
  while (!done);
}

/* Must be called under the connection lock. */
/* Currently, we get new connections for different names for the same
 * server. */
static int LookupServer(ServerInfo **server_p, char *serverName,
			NetSocketLocation *target)
{
  int i;
  ServerInfo *server;
  for (i = ArrayLength(servArray) - 1; i >= 0; i--)
    {
      server = *((ServerInfo **)ArrayNth(servArray, i));
      if ((server)->name == serverName)
	{
	  if (server_p)
	    *server_p = server;
	  return 1;
	}
    }
  return 0;
}

static int PrepareSocket(ServerInfo **server_p, char *serverName)
{
  Socket *sock;
  ServerInfo *server;
  NetSocketLocation *target = NULL;
    
  if (!serverName) return 0;

  serverName = InstanceString(serverName);

  // Check if we already have a connection
  lock(connectionLock);
  if (LookupServer(server_p, serverName, NULL))
    {
      unlock(connectionLock);
      return 1;
    }
  unlock(connectionLock);

  // Find the location; must not hold the connection lock, because it can
  // recurse.
  if (!ServerLocationByName(&target, serverName))
    return 0;

  // Recheck if we already have a connection, because we may have been bypassed
  lock(connectionLock);
  if (LookupServer(server_p, serverName, target))
    {
      unlock(connectionLock);
      return 1;
    }

  //printf("Must connect\n");
  sock = SocketConnect(target, 1);
  if (!sock)
    {
      unlock(connectionLock);
      printf("Failed to connect to %s at ", serverName);
      PrintLocation(target);
      printf("\n");
      free(target);
      return 0;
    }

  server = malloc(sizeof(ServerInfo));
  server->name = serverName;
  server->socket = sock;
  server->status.result = NULL;
#ifdef USE_THREADS
  pthread_mutex_init(&server->requestLock, NULL);
  pthread_mutex_init(&server->ioLock, NULL);
#endif
  // Initialize the connection
  negotiateClient(sock);

  //printf("Connected\n");

  ArrayAppend(servArray, &server);
  ArrayAppend(sockArray, &sock);
  unlock(connectionLock);
  *server_p = server;
  free(target);
  return 1;
}

int AttemptEncryption()
{
#ifdef USE_SSL
  return 1;
#else
  return 0;
#endif
}

static int GetLibraryInfo(char *path, NodeType **p_retNode)
{
  int i;
  NodeType *ret = NULL;
  NodeType *sub = NULL;
  char *headName;
  headName = GetUNLHead(path);

  //printf("Library asked for '%s'\n", headName);
  if (!strcmp(headName, ""))
    {
      ret = CreateBranchNode("local", T_BRANCH);
      if (GetLibraryInfo("/Info", &sub))
	InsertNode(ret, sub);
      if (GetLibraryInfo("/Servers", &sub))
	InsertNode(ret, sub);
    }
  else if (!strcmp(headName, "Servers"))
    {
      ret = CreateBranchNode("Servers", T_BRANCH);
      lock(connectionLock);
      for (i = 0; i < ArrayLength(servArray); i++)
	{
	  ServerInfo *info = *((ServerInfo **) ArrayNth(servArray, i));
	  NetSocketLocation *remote, *copy;
	  NodeType *server = 
	    CreateBranchNode(info->name, T_BRANCH);
	  remote = SocketRemote(info->socket);
	  copy = CopyNetSocketLocation(remote);
	  sub = CreateAddressNode("Address", T_NETSOCK, copy);
	  InsertNode(server, sub);
				  
	  //printf("Adding %d: %s\n", i, sub->name);
	  InsertNode(ret, server);
	}
      unlock(connectionLock);
    }
  *p_retNode = ret;
  return !!ret;
}

static DArray InitUNLArray(char *UNL)
{
  DArray newArray = NULL;

  newArray = ArrayNew(sizeof(char *), kUNLArrayInitSize, NULL);
  if (newArray)
    {
      ArrayAppend(newArray, &UNL);
    }

  return newArray;
}

static void ApproveNewUNL(DArray visitedUNLs, char *newUNL, int *CYCLIC)
{
  if (ArraySearch(visitedUNLs, &newUNL, ArrayEq, 0, 0) == NOT_FOUND)
    {
      ArrayAppend(visitedUNLs, &newUNL);
      *CYCLIC = 0;
    }
  else 
    {
      *CYCLIC = 1;
    }
}

static char *FollowSymlinks(char *UNL, char *command, NodeType *sendNode,
			    NodeType **p_retNode, char **p_actualUNL,
			    int checkOwned, unsigned char flags,
			    int *p_respFlags)
{
  int CYCLIC;
  char *server;
  ServerInfo *serverInfo;
  Socket *socket;
  DArray visitedUNLs = NULL;
  char *result;
  NodeType *retNode;
  int respFlags;

  do
    {
      server = GetServerName(UNL);
      if (!*server)
	{
	  // Talking to the library...
	  //printf("Local\n");
	  *p_actualUNL = UNL;
	  result = "Local";
	  break;
	}
      if (!PrepareSocket(&serverInfo, server))
	{
	  printf("Couldn't prepare\n");
	  result = NULL;
	  break;
	}
      socket = serverInfo->socket;
      lock(serverInfo->requestLock);
      if (!SendRequest(serverInfo, command, sendNode, UNL, flags))
	{
	  printf("Couldn't send\n");
	  result = NULL;
	  break;
	}
      result = HandleSyncReturn(serverInfo, command, UNL, 
				&retNode, &respFlags);
      unlock(serverInfo->requestLock);
      if (p_respFlags)
	*p_respFlags = respFlags;
      if (!result)
	{
	  printf("No result\n");
	  break;
	}
      if (!strcmp(result, SV_SUCCESS))
	{
	  if (p_actualUNL)
	    *p_actualUNL = UNL;
	  break;
	}
      else if (!strcmp(result, SV_SYMLINK))
	{
	  if (checkOwned)
	    {
	      if (!(respFlags & F_SYM_LOCKED))
		{
		  // XXXX Fix this...
		  flags |= F_PUBLISH_SYMUNTRUSTED;
		}
	    }
	  UNL = InstanceString(retNode->data.string_val);
	  
	  FreeNode(retNode);
	  retNode = NULL;

	  if (!visitedUNLs)
	    visitedUNLs = InitUNLArray(UNL);
	  else
	    {
	      ApproveNewUNL(visitedUNLs, UNL, &CYCLIC);
	      if (CYCLIC)
		{
		  printf("Cycle formed by symlinks\n");
		  result = NULL;
		  break;
		}
	    }
	}
      else
	{
	  printf("result: %s\n", result);
	  break;
	}
    }
  while (1);
  if (visitedUNLs)
    ArrayFree(visitedUNLs);
  if (p_retNode)
    *p_retNode = retNode;
  return result;
}

/*************************** Public functions ************************/

int SecureUNL(char *UNL)
{
  // NOTE: does not follow symlinks to determine whether the eventual
  // connection is secure.

  ServerInfo *serverInfo;
  char *server;

  if (!ipcInit)
    initializeData();

  server = GetServerName(UNL);
  if (!*server)
    return 1; // unsuprisingly, we have a secure "connection" to the library
  if (!PrepareSocket(&serverInfo, server))
    return 0; // No connection at all...
  
  return SocketIsSecure(serverInfo->socket);
}

NodeType *GetNode(char *UNL, GetNodeInfo *info)
{
  char *result;
  unsigned char ipcFlags = 0;
  GetNodeInfo localInfo;
  NodeType *ret;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_FOLLOW_LINKS;
    }
  if (!UNL)
    {
      return NULL;
    }
  
  if (info->common.commonFlags & ENCHANT_FOLLOW_LINKS)
    ipcFlags |= FOLLOW_SYMLINK;

  result = 
    FollowSymlinks(UNL, C_GET, NULL, &ret, &UNL, 0, ipcFlags, NULL);
  if (!result)
    return NULL;
  if (!strcmp(result, SV_SUCCESS))
    {
      info->common.actualUNL = UNL;
      return ret;
    }
  if (!strcmp(result, "Local"))
    {
      char *path;
      info->common.actualUNL = UNL;
      path = ExtractUNLPath(UNL);
      GetLibraryInfo(path, &ret);
      return ret;
    }
  return NULL;
}

int LockUNL(char *UNL, LockUNLInfo *info)
{
  char *callType, *result = NULL;
  unsigned char ipcFlags = 0;
  LockUNLInfo lockInfo;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &lockInfo;
      info->common.operationFlags = ENCHANT_LOCK_NORMAL;
      info->common.commonFlags = ENCHANT_CROSS_LINKS;
    }
  if (!UNL)
    return 0;
  
  callType = ((info->common.operationFlags & 3) == 0) ? C_UNLOCK : C_LOCK;

  // Package up the call in a node to send

  if (info->common.operationFlags & ENCHANT_LOCK_RECURSIVE)
    ipcFlags |= F_LOCK_RECURSE;
  if (info->common.commonFlags & ENCHANT_FOLLOW_LINKS)
    ipcFlags |= FOLLOW_SYMLINK;

  // ???? should be separate...
  if ((info->common.operationFlags & ENCHANT_LOCK_REMOVE) == 
      ENCHANT_LOCK_REMOVE)
    ipcFlags |= F_LOCK_REMOVE;
    
  result =
    FollowSymlinks(UNL, callType, NULL, NULL, &UNL, 0, ipcFlags, NULL);
  if (!result)
    return 0;

  if (!strcmp(result, SV_SUCCESS))
    {
      info->common.actualUNL = UNL;
      return 1;
    }

  return 0;
}

static void UnpackGetNodeAttrReturn(NodeType *attrNode,
				    GetUNLLockInfo *info)
{
  NodeType *node;
  char *lockType;

  node = GetChild(attrNode, S_ATTR_LOCK);

  if (node)
    {
      info->lockOwner = 
	InstanceString(GetChild(node, S_ATTR_LOCK_OWNER)->data.string_val);
      lockType = GetChild(node, S_ATTR_LOCK_TYPE)->data.string_val;
      if (!strcmp(lockType, KV_ATTR_LOCK_TYPE_REMOVE))
	{
	  info->lockType = ENCHANT_LOCK_REMOVE;
	}
      else if (!strcmp(lockType, KV_ATTR_LOCK_TYPE_RECURSE))
	{
	  info->lockType = ENCHANT_LOCK_RECURSIVE;
	}
      else
	{
	  info->lockType = ENCHANT_LOCK_NORMAL;
	}
    }
  else
    {
      info->lockOwner = NULL;
      info->lockType = ENCHANT_LOCK_NONE;
    }
}

int GetUNLLock(char *UNL, GetUNLLockInfo *info)
{
  char *result = NULL;
  unsigned char ipcFlags = 0;
  GetUNLLockInfo localInfo;
  NodeType *attrNode;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_CROSS_LINKS;
    }
  if (!UNL)
    return 0;

  result = 
    FollowSymlinks(UNL, C_GETATTR, NULL, &attrNode, &UNL, 0, ipcFlags,
		   NULL);
  if (!result)
    return 0;

  if (!strcmp(result, SV_SUCCESS))
    {
      UnpackGetNodeAttrReturn(attrNode, info);
      info->common.actualUNL = UNL;
      return 1;
    }

  return 0;
}

int PublishNode(char *UNL, NodeType *node, PublishNodeInfo *info)
{
  char *result = NULL;
  unsigned char ipcFlags = 0;
  PublishNodeInfo localInfo;
  
  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_CROSS_LINKS;
      info->common.operationFlags = 
	ENCHANT_PUBLISH_REPLACE | ENCHANT_PUBLISH_SOFT;
    }
  if (!UNL)
    return 0;
  
  if (info->common.operationFlags & ENCHANT_PUBLISH_HARD)
    ipcFlags |= F_PUBLISH_HARD;
  if (info->common.commonFlags & ENCHANT_FOLLOW_LINKS)
    ipcFlags |= FOLLOW_SYMLINK;

  if ((info->common.operationFlags & 3) == ENCHANT_PUBLISH_REPLACE)
    ipcFlags |= F_PUBLISH_REPLACE;
  if ((info->common.operationFlags & 3) == ENCHANT_PUBLISH_ADDONLY)
    ipcFlags |= F_PUBLISH_PRESERVE;

  result = 
    FollowSymlinks(UNL, C_PUBLISH, node, NULL, &UNL, 1, ipcFlags, NULL);

  if (!result)
    return 0;

  if (!strcmp(result, SV_SUCCESS))
    {
      info->common.actualUNL = UNL;
      return 1;
    }

  return 0;
}
 
int SubscribeUNL(char *UNL, SubscribeUNLInfo *info)
{
  NodeType *dataNode = NULL;
  char *result = NULL;
  unsigned char ipcFlags = 0;
  SubscribeUNLInfo localInfo;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_FOLLOW_LINKS;
      info->common.operationFlags =
	ENCHANT_SUBSCRIBE_OTHERS | ENCHANT_SUBSCRIBE_NEW;
    }
  if (!UNL) return 0;

  if (info->common.operationFlags & ENCHANT_SUBSCRIBE_SELF)
    ipcFlags |= F_SUBSCRIBE_SELF;
  if (info->common.commonFlags & ENCHANT_FOLLOW_LINKS)
    ipcFlags |= FOLLOW_SYMLINK;

  result = FollowSymlinks(UNL, C_SUBSCRIBE, NULL, &dataNode, &UNL, 0, 
			  ipcFlags, NULL);
  if (!result)
    return 0;

  if (!strcmp(result, SV_SUCCESS))
    {
      if (dataNode)
	{
	  if (info->common.operationFlags & ENCHANT_SUBSCRIBE_INITIAL)
	    {
	      char *path = ExtractUNLPath(UNL);
	      // Act like we got it in an update
	      EnqueueUpdate(GetServerName(UNL), NULL, dataNode, path);
	    }
	  else
	    FreeNode(dataNode);
	}
      info->common.actualUNL = UNL;
      return 1;
    }

  return 0;
}

int UnsubscribeUNL(char *UNL, UnsubscribeUNLInfo *info)
{
  char *result = NULL;
  unsigned char ipcFlags = 0;
  UnsubscribeUNLInfo localInfo;

  if (!ipcInit)
    initializeData();
  if (!UNL || !info)
    {
      if (!info)
	info = &localInfo;
      info->common.commonFlags = ENCHANT_FOLLOW_LINKS;
    }
  if (!UNL) return 0;

  if (info->common.commonFlags & ENCHANT_FOLLOW_LINKS)
    ipcFlags |= FOLLOW_SYMLINK;

  result = FollowSymlinks(UNL, C_UNSUBSCRIBE, NULL, NULL, &UNL, 
			  0, ipcFlags, NULL);
  if (!result)
    return 0;

  if (!strcmp(result, SV_SUCCESS)) 
    {
      info->common.actualUNL = UNL;
      return 1;
    }

  return 0;
}

/* The core of GetUpdate and WaitForUpdate. */
static void scanForUpdate(GetUpdateInfo *getInfo, char *target, int blocking)
{
  UpdateInfo *info;
  UpdateInfo *result = NULL;
  int found = -1;
  int idx = 0;
  int updatesSentSeen;
  int updateSequenceSeen;
  int any;
  char *targetServer, *targetPath;

  if (!ipcInit)
    initializeData();
  
  if (target)
    {
      targetServer = GetServerName(target);
      targetPath = ExtractUNLPath(target);
      any = 0;
    }
  else
    {
      any = 1;
    }

  getInfo->UNL = NULL;
  getInfo->node = NULL;
  getInfo->publisher = NULL;

  if (!blocking)
    UnloadUpdates(0, updateSequence); // need a single check
  
  // Wait until there is something in the queue
  do
    {
      lock(updateQueueLock);
      updatesSentSeen = updatesSent;
      updateSequenceSeen = updateSequence;
      while (ArrayLength(updateQueue) > idx)
	{
	  info = *((UpdateInfo **)ArrayNth(updateQueue, idx));
	  if (any || (!strcmp(targetServer, info->server) &&
		      (!strcmp(targetPath, info->path) || 
		       DescendsFrom(targetPath, info->path))))
	    {
	      found = idx;
	      result = info;
	      break;
	    }
	  else
	    {
	      //printf("Skip %s %s\n", info->server, info->path);
	    }
	  idx++;
	}
      unlock(updateQueueLock);

      if (found >= 0 || !blocking)
	break;

      UnloadUpdates(1, updateSequenceSeen);

      // Move back to account for delivered updates Note that, if
      // something gets added to the end and delivered to another
      // thread, we'll back up inappropriately. We could fall off the
      // start, or we could look at a skipped update again; both are
      // fine.
      idx -= (updatesSent - updatesSentSeen);
      // Catch falling off the start
      if (idx < 0)
	idx = 0;
    }
  while (1);

  if (result)
    {      
      // Delete the pointer to this node and move indexes.
      ArrayDeleteAt(updateQueue, found);
      updatesSent++;
      getInfo->node = result->node;
      getInfo->UNL = CreateUNL(result->server, result->path);
      getInfo->publisher = result->publisher;
      free(result);
    }
}

void GetUpdate(GetUpdateInfo *info, char *UNL)
{
  scanForUpdate(info, UNL, 0);
}

NodeType *WaitForUpdate(GetUpdateInfo *info, char *UNL)
{
  GetUpdateInfo localInfo;
  if (!info)
    info = &localInfo;
  scanForUpdate(info, UNL, 1);
  return info->node;
}

/* Multiplexer for admin operations:
 * if (settype), sets the fixed type (or unsets it if type is NULL)
 * if (operation), sets the permissions for that operation to auth
 */
static int Admin(char *UNL, char operation, char **auth,
		 long *type, int settype)
{
  int i;
  static NodeType *sendNode = NULL;
  static NodeType *opNode = NULL;
  static NodeType *typeNode = NULL;
  NodeType *authNode = NULL, *tokenNode;
  char *result = NULL;

  if (!ipcInit)
    initializeData();
  if (!UNL) return 0;

  if (!sendNode)
    {
      sendNode = CreateBranchNode(C_ADMIN, T_BRANCH);
      opNode = CreateIntegerNode(K_OPERATION, T_CHARACTER, 0);
      typeNode = CreateEmptyNode(K_TYPE, T_EMPTY);
      
      InsertNode(sendNode, opNode);
    }

  if (settype)
    {
      if (type)
	{
	  typeNode->type = T_TYPE;
	  typeNode->data.type_val = *type;
	}
      else
	{
	  typeNode->type = T_EMPTY;
	}
      InsertNode(sendNode, typeNode);
    }

  if (operation)
    {
      if (auth)
	{
	  authNode = CreateBranchNode(K_AUTH, T_BRANCH);
	  for (i = 0; auth[i]; i++)
	    {
	      tokenNode = CreateStringNode(InstanceNumber(i), T_STRING,
					   auth[i]);
	      InsertNode(authNode, tokenNode);
	    }
	  InsertNode(sendNode, authNode);
	}
      
      opNode->data.int_val = operation;
    }
  else
    {
      RemoveNode(sendNode, opNode);
    }

  result = FollowSymlinks(UNL, C_ADMIN, sendNode, NULL, NULL, 0, 0, NULL);

  if (!operation)
    {
      InsertNode(sendNode, opNode);
    }

  if (settype)
    {
      RemoveNode(sendNode, typeNode);
    }

  if (authNode)
    {
      RemoveNode(sendNode, authNode);
      FreeNode(authNode);
    }

  return result && !strcmp(result, SV_SUCCESS);  
}

int SetUNLFixedType(char *UNL, long type)
{
  return Admin(UNL, 0, NULL, &type, 1);
}

int ClearUNLFixedType(char *UNL)
{
  return Admin(UNL, 0, NULL, NULL, 1);
}

int SetUNLPermissions(char *UNL, char operation, char **auth)
{
  return Admin(UNL, operation, auth, NULL, 0);
}

void WhiteboardVerbosity(int level)
{
  if (level == 0)
    quiet_client_ipc = 1;
  else
    quiet_client_ipc = 0;
}
