/*************************************************
 * File: request.c
 * Birth Date: Dec 03, 2003
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2003, 2004 MIT Media Lab
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *************************************************/

#include <enchantment/types.h>
#include <enchantment/node.h>
#include <enchantment/unl.h>

#include "symbols.h"
#include "features.h"
#include "logging.h"
#include "clients.h"
#include "tree-manip.h"
#include "request.h"
#include "whiteboard-protocol.h"

#include <string.h>
#include <stdlib.h>

#ifdef USE_THREADS
#include <pthread.h>
#include <semaphore.h>
#endif

DArray requestQueue = NULL;

#ifdef USE_THREADS
pthread_mutex_t requestLock;
pthread_cond_t requestPending;
#endif

typedef struct
{
  ClientIDType *clientID;
  NodeType *rcvNode;
  char *command;
  unsigned flags : 6;

  NodeType *dataNode;
  char *status;
  NodePathReturnType npReturn;
} RequestType;

#define RequestHasFlag(request, flag) (!!((request)->flags & (flag)))

void InitRequestQueue() 
{
  requestQueue = ArrayNew(sizeof(RequestType), 0, NULL);
#ifdef USE_THREADS
  pthread_mutex_init(&requestLock, NULL);
  pthread_cond_init(&requestPending, NULL);
#endif
}


void EnqueueRequest(ClientIDType *clientID, NodeType *rcvNode, char *command,
		    char *unl, unsigned char flags)
{
  RequestType request;
  request.clientID = clientID;
  request.rcvNode = rcvNode;
  request.command = command;
  request.npReturn.UNL = unl;
  request.flags = flags;
  request.dataNode = NULL;
  request.status = NULL;
#ifdef USE_THREADS
  pthread_mutex_lock(&requestLock);
  pthread_cond_signal(&requestPending);
#endif
  if (SHOW_QUEUING)
    printf("Enqueue from %s %s\n", request.clientID->name, request.command);
  ArrayAppend(requestQueue, &request);
#ifdef USE_THREADS
  pthread_mutex_unlock(&requestLock);
#endif
}

static void CompleteRequest(RequestType *request)
{
  SendResultNode(request->clientID, request->dataNode, 
		 request->command, request->npReturn.UNL,
		 request->status);
  if (request->rcvNode)
    FreeNode(request->rcvNode);
  request->rcvNode = NULL;
}

static char *PrepareSymlinkReturn(NodeType **p_dstNode, 
				  NodePathReturnType *npReturn, 
				  ClientIDType *clientID)
{
  char *symUNL = AppendUNLPath(npReturn->existNode->data.string_val, 
			       npReturn->symlinkCont);

  *p_dstNode = CreateStringNode(S_SYMLINK_VAL, T_STRING, symUNL);
  
  return SV_SYMLINK;
}

static char *HandleSymlink(RequestType *request)
{
  return PrepareSymlinkReturn(&request->dataNode, &request->npReturn, 
			      request->clientID);
}

static char *HandleGet(RequestType *request)
{
  int derefSymlinks = RequestHasFlag(request, FOLLOW_SYMLINK);

  GetNodePathInfo(request->clientID, request->npReturn.UNL,
		  &request->npReturn, 0, derefSymlinks, 0);
  
  if (derefSymlinks && request->npReturn.existNode && 
      request->npReturn.existNode->type == T_SYMLINK)
    {
      return HandleSymlink(request);
    }

  if (!request->npReturn.mayRead)
    return SV_DOES_NOT_EXIST;
  request->dataNode = request->npReturn.existNode;
  return request->npReturn.existNode ? SV_SUCCESS : SV_DOES_NOT_EXIST;
}

/*
static NodeType *CreateFreshHandleNode(ClientIDType *clientID, char *name, 
				       TypeID type)
{
  static int id = 1;
  NodeType *ret = CreateNode(name, type, sizeof(long), NULL, 0);
  ret->data.handle_val = id++;
  return ret;
}
*/

/*
static char *HandleCreate(RequestType *request)
{
  NodeType *reqNode = request->rcvNode;
  TypeID type = reqNode->data.type_val;
  char *result = SV_BAD_REQUEST;
  switch (GetFormat(type))
    {
    case T_HANDLE:
      request->dataNode =
	CreateFreshHandleNode(request->clientID, "created", type);
      result = SV_SUCCESS;
      break;
    }
  return result;
}
*/

static char *HandleGetNodeAttr(RequestType *request)
{
  static NodeType *rootNode = NULL, *lockNode = NULL, *permNode = NULL, 
    *ownerNode = NULL, *typeNode = NULL;
  int derefSymlinks = RequestHasFlag(request, FOLLOW_SYMLINK);
  ServerLockType *reportLock;

  GetNodePathInfo(request->clientID, request->npReturn.UNL,
		  &request->npReturn, 0, derefSymlinks, 0);
  
  if (derefSymlinks && request->npReturn.existNode && 
      request->npReturn.existNode->type == T_SYMLINK)
    {
      return HandleSymlink(request);
    }

  reportLock = request->npReturn.reportLock;

  if (!rootNode)
    {
      rootNode = CreateBranchNode(S_ATTR, T_BRANCH);
    }
  else
    {
      RemoveNode(rootNode, lockNode);
      RemoveNode(rootNode, permNode);
    }

  if (!request->npReturn.targetSchema)
    return SV_DOES_NOT_EXIST;

  if (reportLock)
    {
      // Static initialization done here, once.
      if (!lockNode)  
	{
	  lockNode = CreateBranchNode(S_ATTR_LOCK, T_BRANCH);
	  ownerNode = CreateStringNode(S_ATTR_LOCK_OWNER, T_STRING, "");
	  typeNode = CreateStringNode(S_ATTR_LOCK_TYPE, T_STRING, "");
	  InsertNode(lockNode, ownerNode);
	  InsertNode(lockNode, typeNode);
	}
      InsertNode(rootNode, lockNode);

      NodeSetString(ownerNode, reportLock->owner->name);

      if (reportLock->REMOVE)
	{
	  NodeSetString(typeNode, KV_ATTR_LOCK_TYPE_REMOVE);
	}
      else if (reportLock->RECURSE)
	{
	  NodeSetString(typeNode, KV_ATTR_LOCK_TYPE_RECURSE);
	}
      else
	{
	  NodeSetString(typeNode, KV_ATTR_LOCK_TYPE_NORMAL);
	}
      
      /*
      if (reportLock->RECURSE)
	LowBitSet(rootNode, F_ATTR_LOCK_RECURSE);
      else 
	LowBitClear(rootNode, F_ATTR_LOCK_RECURSE);

      if (reportLock->REMOVE)
	LowBitSet(rootNode, F_ATTR_LOCK_REMOVE);
      else 
	LowBitClear(rootNode, F_ATTR_LOCK_REMOVE);
      */
    }

  if (1)
    {
      if (!permNode)
	{
	  permNode = CreateStringNode(S_ATTR_PERM, T_STRING, "    ");
	}
      InsertNode(rootNode, permNode);
      permNode->data.string_val[0] = 
	CheckPermission(request->npReturn.targetSchema, 
			request->clientID, 'r') ? 'r' : ' ';
      permNode->data.string_val[1] = 
	CheckPermission(request->npReturn.targetSchema, 
			request->clientID, 'w') ? 'w' : ' ';
      permNode->data.string_val[2] = 
	CheckPermission(request->npReturn.targetSchema, 
			request->clientID, 'l') ? 'l' : ' ';
      permNode->data.string_val[3] = 
	CheckPermission(request->npReturn.targetSchema, 
			request->clientID, 'a') ? 'a' : ' ';
    }
  
  request->dataNode = rootNode;
  return SV_SUCCESS;
}

static char *HandleLockOps(RequestType *request)
{
  int result;
  ServerLockType clientLock;

  int is_recurse = RequestHasFlag(request, F_LOCK_RECURSE);
  int is_remove = RequestHasFlag(request, F_LOCK_REMOVE);
  int derefSymlinks = RequestHasFlag(request, FOLLOW_SYMLINK);

  GetNodePathInfo(request->clientID, request->npReturn.UNL,
		  &request->npReturn, 0, derefSymlinks, 1);
  
  if (derefSymlinks && request->npReturn.existNode && 
      request->npReturn.existNode->type == T_SYMLINK)
    {
      return HandleSymlink(request);
    }
  
  if (!request->npReturn.targetSchema)
    {
      return SV_DOES_NOT_EXIST; 
    }

  if (request->command == c_lock) // Lock operation
    {
      clientLock.owner = request->clientID;
      clientLock.RECURSE = is_recurse;
      clientLock.REMOVE = is_remove;

      if (is_recurse && request->npReturn.existNode &&
	  !CheckDescendantPermission(request->npReturn.existNode,
				     request->clientID, 'l'))
	return sv_locked;
	
      result = LockNodePath(request->npReturn.UNL, &clientLock);
    }
  else // Unlock operation
    {
      result = UnlockNodePath(request->npReturn.UNL, request->clientID);
    }
  return result ? SV_SUCCESS : SV_LOCKED;
}

static char *HandlePublish(RequestType *request)
{
  char *op;
  char *ret;

  NodeType *datanode;

  int is_hard = RequestHasFlag(request, F_PUBLISH_HARD);
  int publish_if_locked = !RequestHasFlag(request, F_PUBLISH_SYMUNTRUSTED);
  int derefSymlinks = RequestHasFlag(request, FOLLOW_SYMLINK);

  //UnpackPublish(request->rcvNode, &dataNode);
  if (!*request->npReturn.UNL)
    {
      if (CLIENT_ERROR) 
	printf("Failed to publish node because of bad UNL.\n");
      return SV_BAD_UNL;
    }

  datanode = request->rcvNode;

  if (RequestHasFlag(request, F_PUBLISH_REPLACE))
    {
      if (datanode)
	op = kv_replace;
      else
	op = kv_delete;
    }
  else
    {
      if (RequestHasFlag(request, F_PUBLISH_PRESERVE))
	op = kv_addonly;
      else
	op = kv_merge;
    }
  
  GetNodePathInfo(request->clientID, request->npReturn.UNL, 
		  &request->npReturn, 1, derefSymlinks, 0);
  
  if (derefSymlinks && request->npReturn.existNode && 
      request->npReturn.existNode->type == T_SYMLINK)
    {
      // We want to follow this link, not replace it
      return HandleSymlink(request);
    }

  // The name of the node to be merged in must be the
  // same as the existing node to prevent clients from
  // changing the names of node that other clients might
  // be subscribed to.
  
  if (request->npReturn.existNode && datanode) 
    datanode->name = request->npReturn.existNode->name;

  ret = AddNodeToTree(request->clientID, datanode,
		      &request->npReturn, op, is_hard, publish_if_locked);
  if (ret == sv_success)
    request->rcvNode = NULL;
  return ret;
}

static char *HandleAdmin(RequestType *request)
{
  NodeType *reqNode = request->rcvNode;
  NodeType *eltNode, *tokenNode;
  NodeType *typeNode = NULL;
  int derefSymlinks = 0;
  ClientSelector selector;
  char operation = '\0';
  unsigned index;

  selector.tokens = NULL;

  for (index = 0; (eltNode = BranchIndex(reqNode, index)); index++)
    {
      if (eltNode->name == k_type)
	{
	  typeNode = eltNode;
	}
      if (eltNode->name == k_auth)
	{
	  int i = 0;
	  unsigned tokenIndex;
	  selector.tokens = malloc(sizeof(char *) * 
				   (ArrayLength(eltNode->data.branch_val) + 1));
	  for (tokenIndex = 0; (tokenNode = BranchIndex(eltNode, tokenIndex));
	       tokenIndex++)
	    {
	      selector.tokens[i++] = 
		InstanceString(tokenNode->data.string_val);
	    }
	  selector.tokens[i] = NULL;
	}
      if (eltNode->name == k_operation)
	{
	  operation = eltNode->data.int_val;
	}
    }
  if (!request->npReturn.UNL)
    return SV_BAD_REQUEST;
  GetNodePathInfo(request->clientID, request->npReturn.UNL,
		  &request->npReturn, 0, derefSymlinks, 1);
  if (!request->npReturn.targetSchema)
    {
      // This should only happen if it could not be created (?)
      if (selector.tokens)
	free(selector.tokens);
      return SV_DOES_NOT_EXIST;
    }
  if (!CheckPermission(request->npReturn.targetSchema, request->clientID, 'a'))
    {
      if (selector.tokens)
	free(selector.tokens);
      return SV_LOCKED;
    }
  if (operation == 'a' && !FitsSelector(&selector, request->clientID))
    {
      if (selector.tokens)
	free(selector.tokens);
      return SV_LOCKED;
    }
  if (operation)
    SetPermissions(request->npReturn.targetSchema, operation, &selector);
  if (typeNode)
    {
      if (typeNode->type == T_TYPE)
	{
	  SetFixedType(request->npReturn.targetSchema, 
		       typeNode->data.type_val);
	  /*
	  if (request->npReturn.existNode)
	    request->npReturn.existNode->type &= F_FIXEDTYPE_MASK;
	  */
	}
      else
	{
	  SetVariableType(request->npReturn.targetSchema);
	}
    }
  return SV_SUCCESS;
}

static char *HandleSubscribe(RequestType *request)
{
  int self_subscribe = RequestHasFlag(request, F_SUBSCRIBE_SELF);
  int derefSymlinks = RequestHasFlag(request, FOLLOW_SYMLINK);
  
  GetNodePathInfo(request->clientID, request->npReturn.UNL,
		  &request->npReturn, 0, derefSymlinks, 1);
  
  if (derefSymlinks && request->npReturn.existNode && 
      request->npReturn.existNode->type == T_SYMLINK)
    {
      return HandleSymlink(request);
    }
  
  // If the node exists, than we are fine.
  Subscribe(request->clientID, &request->npReturn, self_subscribe);
  request->dataNode = request->npReturn.existNode;
  return SV_SUCCESS;
}

static char *HandleUnsubscribe(RequestType *request)
{
  SubscriptionType *subscription;
  int derefSymlinks = RequestHasFlag(request, FOLLOW_SYMLINK);

  GetNodePathInfo(request->clientID, request->npReturn.UNL,
		  &request->npReturn, 0, derefSymlinks, 0);
  
  if (derefSymlinks && request->npReturn.existNode && 
      request->npReturn.existNode->type == T_SYMLINK)
    {
      return HandleSymlink(request);
    }
  
  if ((subscription = IsSubscribed(request->clientID, request->npReturn.UNL)))
    {
      //printf("Unsubscribing from %s\n", subscription->UNL);
      ForgetSubscription(request->clientID, subscription);
      if (request->npReturn.existNode)
	ForgetSubscriber(request->npReturn.existNode, subscription);
      else
	ForgetSubscriber(request->npReturn.closestAncestorNode, subscription);
    }

  return SV_SUCCESS;
}

static void ProcessRequest(RequestType *request)
{
  NodeType *rcvNode = request->rcvNode;
  char *rootName;
  char *status;
  if (!request->command)
    {
      plogf("Client %s done\n", request->clientID->name);
      SocketClose(request->clientID->socket);
      RemoveClient(request->clientID);
      return;
    }
  rootName = request->command;
  
  // Every root node should be interpretable, so do check
  if (SHOW_REQUESTS)
    {
      printf("%s:\n", request->clientID->name);
      PrintNode(rcvNode, NULL);
    }
  plogf("Request from %s:\n", request->clientID->name);
  logTree();
  if (c_get == rootName)
    {
      status = HandleGet(request);
    }
  else if (c_getattr == rootName)
    {
      status = HandleGetNodeAttr(request);
    }
  else if (c_lock == rootName || c_unlock == rootName)
    {
      status = HandleLockOps(request);
    }
  else if (c_publish == rootName)
    {
      status = HandlePublish(request);
    }
  else if (c_subscribe == rootName)
    {
      status = HandleSubscribe(request);
    }
  else if (c_unsubscribe == rootName) 
    {
      status = HandleUnsubscribe(request);
    }
  else if (c_admin == rootName)
    {
      status = HandleAdmin(request);
    }
  else
    {
      //!!!! ERROR case, so should return S_ERROR node??
      if (CLIENT_ERROR) 
	printf("Unknown type of root node '%s' received!\n", rootName);
      status = SV_BAD_REQUEST;
    }
  request->status = status;
  CompleteRequest(request);
}

void ProcessRequests()
{
  int done = 0;
  //printf("%d requests enqueued\n", ArrayLength(requestQueue));
  while (!done)
    {
      RequestType request;
#ifdef USE_THREADS
      pthread_mutex_lock(&requestLock);
#endif
      if (ArrayLength(requestQueue))
	{
	  request = *(RequestType *)ArrayNth(requestQueue, 0);
	  ArrayDeleteAt(requestQueue, 0);
	}
      else
	done = 1;
#ifdef USE_THREADS
      if (SHOW_QUEUING)
	{
	  if (done)
	    printf("None\n");
	  else
	    printf("Dequeue from %s %s\n", request.clientID->name,
		   request.command);
	}
      pthread_mutex_unlock(&requestLock);
#endif
      if (!done)
	ProcessRequest(&request);
    }
}

#ifdef USE_THREADS
static void *TreeManagementLoop(void *data)
{
  do
    {
      if (SHOW_QUEUING)
	printf("Waiting for requests\n");
      pthread_mutex_lock(&requestLock);
      pthread_cond_wait(&requestPending, &requestLock);
      pthread_mutex_unlock(&requestLock);
      if (SHOW_QUEUING)
	printf("got requests\n");
      ProcessRequests();
    }
  while (1);
}

void StartRequestThread()
{
  pthread_t thread;
  pthread_create(&thread, NULL, TreeManagementLoop, NULL);
}
#endif
