/*************************************************
 * File: signal-proxy.c
 * Birth Date: Aug 13, 2004
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2004 Aware Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *************************************************/

#include <enchantment/node.h>
#include <enchantment/signals.h>
#include <enchantment/unl.h>
#include "symbols.h"
#include "whiteboard-protocol.h"
#include "tree-manip.h"
#include "main.h"
#include "logging.h"

#include "introduction.h"
#include "protocol.h"
#include "clients.h"

#include <stdlib.h>

DArray g_signalProxies = NULL;

typedef struct
{
  Signal *signalCon;
  Signal *signalGen;
} SignalProxyInfo;

Socket *IntSignalGetSocket(Signal *signal);
int SignalGenConnected(Signal *signal, Socket *socket);

int HandleNewSignalProxy(Socket *socket)
{
#ifndef USE_THREADS
  Socket *remoteSock;
#endif
  char *unl = NULL;
  int okay = 1;
  int idx;
  int index;
  IntroMessage message;
  int done;
  char *path;
  char *result;
  NodePathReturnType npReturn;
  NodeType *dataNode = NULL;
  int i;

  SignalProxyInfo *info = NULL;

  printf("Got a signal proxy client.\n");

  do
    {
      index = IntroRecv(socket, receiverProperties, &message);
      if (index <= 0)
	;
      else if (message.format == &signalUNL)
	{
	  unl = message.value.string_val;
	}
      else
	{
	  printf("Other message %s?\n", message.format->key);
	}
      if (index)
	done = 0;
    }
  while (index > 0 && !unl);

  if (!unl || !okay)
    {
      SocketClose(socket);
      return 0;
    }

  printf("Request to proxy a signal at %s\n", unl);
  path = ExtractUNLPath(unl);
  if (!path)
    {
      result = SV_BAD_UNL;
    }
  else 
    {
      GetNodePathInfo(serverID, path, &npReturn, 0, 0, 0);
      if (npReturn.existNode)
	{
	  dataNode = npReturn.existNode;
	  if (dataNode->type == T_SIGNAL)
	    result = SV_SUCCESS;
	  else
	    result = SV_FAILURE;
	}
      else 
	result = SV_DOES_NOT_EXIST;
    }
  if (result != SV_SUCCESS)
    {
      printf("Nope.\n");
      IntroDone(socket);
      SocketClose(socket);
      return 0;
    }
  printf("Found signal\n");
  if (!g_signalProxies)
    {
      g_signalProxies = ArrayNew(sizeof(SignalProxyInfo), 0, NULL);
    }
  for (i = 0; i < ArrayLength(g_signalProxies); i++)
    {
      SignalProxyInfo *testInfo = 
	(SignalProxyInfo *)ArrayNth(g_signalProxies, i);
      NetSocketLocation *remote = 
	SignalSource(testInfo->signalCon);
      if (remote->ip == dataNode->data.netsock_val->ip &&
	  remote->port == dataNode->data.netsock_val->port)
	{
	  info = testInfo;
	}
    }
  if (!info)
    {
      // will be put in the array
      SignalProxyInfo newInfo;
      printf("Connecting to proxied signal.\n");
      newInfo.signalCon = NewSignalNode(dataNode);
      SignalConnectWait(newInfo.signalCon);
      if (!SignalConnected(newInfo.signalCon))
	{
	  SocketClose(socket);
	  SignalClose(newInfo.signalCon);
	  return 0;
	}
      newInfo.signalGen = NewSignalGenNone();
      printf("Connected to proxied signal.\n");
      SignalCopyProperties(newInfo.signalCon, newInfo.signalGen);
      
#ifndef USE_THREADS
      remoteSock = IntSignalGetSocket(newInfo.signalCon);
      AddSocket(remoteSock);
#endif
      ArrayAppend(g_signalProxies, &newInfo);
      idx = ArrayLength(g_signalProxies) - 1;
      //RemoveSocket(socket);
      info = ArrayNth(g_signalProxies, idx);
    }
  printf("Negotiating gen.\n");
  SignalGenConnected(info->signalGen, socket);
  printf("Gen negotiated.\n");
  
#ifdef USE_THREADS
  printf("Looping on serving signal...\n");
  while (ServiceSignal(info))
    ;
  printf("Deleted\n");
  ArrayDeleteAt(g_signalProxies, idx);
  return 0;
#endif
  return 1;
}

int ServiceSignal(SignalProxyInfo *info)
{
  char *buffer = malloc(SignalBufferSize(info->signalCon));
  struct timeval time;
  int ret = 1;
  plogf("Service signal... ");
  SignalRecv(info->signalCon, &time, buffer);
  if (!SignalConnected(info->signalCon))
    {
      printf("Proxied signal closed\n");
      SignalClose(info->signalGen);
      ret = 0;
    }
  else
    {
      SignalSend(info->signalGen, time, buffer);
      if (!SignalClientCount(info->signalGen))
	{
	  printf("Nobody listening\n");
	  SignalClose(info->signalGen);
	  SignalClose(info->signalCon);
	  ret = 0;
	}
    }
  free(buffer);
  plogf("done\n");
  return ret;
}

void ServiceSignals(fd_set *p_sockSet)
{
  int i = 0;
  if (!g_signalProxies)
    return;
  while (i < ArrayLength(g_signalProxies))
    {
      SignalProxyInfo *info = (SignalProxyInfo *)ArrayNth(g_signalProxies, i);
      Socket *socket = IntSignalGetSocket(info->signalCon);
      if (!SocketIncluded(p_sockSet, socket))
	{
	  i++;
	  continue;
	}
      if (ServiceSignal(info))
	i++;
      else
	{
	  RemoveSocket(socket);
	  ArrayDeleteAt(g_signalProxies, i);
	}
    }
}
