/*************************************************
 * File: tree-manip.c
 * Birth Date: 12/3/2003
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 * 
 * Copyright (C) 2002 MIT Media Lab,
 *               2004 AWare Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *************************************************/

/* This file implements manipulations of the server's tree state. */

#include "tree-manip.h"
#include "logging.h"

#include <enchantment/unl.h>
#include "symbols.h"

#include "clients.h"
#include "journal.h"

#include <stdlib.h>
#include <string.h>

#include <stdarg.h>

#include <assert.h>

NodeType *g_globalNode = NULL;

extern ClientIDType *serverID;

static void MoveSubscription(SubscriptionType *sub, NodeType *target);

int FitsSelector(ClientSelector *selector, ClientIDType *client)
{
  int i, j;
  if (!selector->tokens)
    {
      if (SHOW_AUTH && client != serverID)
	printf("Any\n");
      return 1;
    }
  for (i = 0; selector->tokens[i]; i++)
    {
      for (j = 0; j < ArrayLength(client->auth); j++)
	{
	  if (!strcmp(selector->tokens[i],
		      *((char **)ArrayNth(client->auth, j))))
	    {
	      if (SHOW_AUTH && client != serverID)
		printf("Matched %s\n", selector->tokens[i]);
	      return 1;
	    }
	  if (SHOW_AUTH && client != serverID)
	    printf("%s != %s\n", selector->tokens[i],
		   *((char **)ArrayNth(client->auth, j)));
	}
    }
  if (SHOW_AUTH && client != serverID)
    {
      if (selector->tokens)
	{
	  int i;
	  printf("[");
	  for (i = 0; selector->tokens[i]; i++)
	    {
	      if (i)
		printf(", ");
	      printf("%s", selector->tokens[i]);
	    }
	  printf("]\n");
	}
      else
	printf("(any)\n");
    }
  return 0;
}

int CheckPermission(SchemaNode *schema, ClientIDType *client, char type)
{
  int ret;
  ClientSelector *selector;
  if (SHOW_AUTH && client != serverID)
    printf("Checking permission %c on %p for %s\n", type, schema, 
	   client->name);
  switch (type)
    {
    case 'w':
      // Locks prevent writes to a node
      if (schema->lock.owner && schema->lock.owner != client)
	return 0;
      selector = &schema->canWrite;
      break;
    case 'r':
      selector = &schema->canRead;
      break;
    case 'l':
      selector = &schema->canLock;
      break;
    case 'a':
      selector = &schema->canAdmin;
      break;
    default:
      return 0;
    }
  ret = FitsSelector(selector, client);
  if (SHOW_AUTH && client != serverID)
    printf("%s\n", ret ? "Allowed" : "Denied");
  return ret;
}

void SelectorNone(ClientSelector *selector)
{
  static char **none = NULL;
  if (!none)
    {
      none = (char **)malloc(sizeof(char *));
      none[0] = NULL;
    }
  selector->tokens = none;
}

void SelectorAny(ClientSelector *selector)
{
  selector->tokens = NULL;
}

void SelectorSome(ClientSelector *selector, ...)
{
  va_list args;
  char *auth;
  int i;
  DArray auths = ArrayNew(sizeof(char *), 0, NULL);
  va_start(args, selector);
  do
    {
      auth = va_arg(args, char *);
      ArrayAppend(auths, &auth);
    }
  while (auth);
  va_end(args);
  selector->tokens =
    (char **)malloc(sizeof(char *) * ArrayLength(auths));
  for (i = 0; i < ArrayLength(auths); i++)
    selector->tokens[i] = *((char **)ArrayNth(auths, i));
  ArrayFree(auths);
}

static SchemaNode *SchemaDefault(SchemaNode *parent, char *name)
{
  SchemaNode *schema = (SchemaNode *)malloc(sizeof(SchemaNode));
  schema->lock.owner = NULL;
  schema->lockTrails = NULL;
  schema->subscribers = NULL;
  schema->hardNode = 0;
  schema->fixed_type = 0;
  schema->hardSchema = 0;
  schema->parent = parent;
  schema->children = ArrayNew(sizeof(SchemaChild), 0, NULL);
  if (parent)
    {
      SchemaChild child;
      child.name = name;
      child.node = schema;
      ArrayAppend(parent->children, &child);
      schema->canRead = parent->canRead;
      schema->canWrite = parent->canWrite;
      schema->canLock = parent->canLock;
      schema->canAdmin = parent->canAdmin;
    }
  else
    {
      SelectorAny(&schema->canRead);
      SelectorAny(&schema->canWrite);
      SelectorAny(&schema->canLock);
      SelectorAny(&schema->canAdmin);
    }
  return schema;
}

SchemaNode *GetSchemaChild(SchemaNode *parent, char *name)
{
  int i;
  for (i = 0; i < ArrayLength(parent->children); i++)
    {
      SchemaChild *child =
	(SchemaChild *)ArrayNth(parent->children, i);
      if (!strcmp(child->name, name))
	return child->node;
    }
  return NULL;
}

int CheckNodePermission(NodeType *node, ClientIDType *client, char type)
{
  if (SHOW_AUTH && client != serverID)
    printf("For %s, ", node->name);
  return CheckPermission(SchemaOf(node), client, type);
}

void SetPermissions(SchemaNode *schema, char type, ClientSelector *selector)
{
  ClientSelector *target;
  switch (type)
    {
    case 'w':
      target = &schema->canWrite;
      break;
    case 'r':
      target = &schema->canRead;
      break;
    case 'l':
      target = &schema->canLock;
      break;
    case 'a':
      target = &schema->canAdmin;
      break;
    default:
      return;
    }
  *target = *selector;
  if (SHOW_AUTH)
    {
      printf("Set %c permission on %p to ", type, schema);
      if (selector->tokens)
	{
	  int i;
	  printf("[");
	  for (i = 0; selector->tokens[i]; i++)
	    {
	      if (i)
		printf(", ");
	      printf("%s", selector->tokens[i]);
	    }
	  printf("]\n");
	}
      else
	printf("(any)\n");
    }
}

void SetFixedType(SchemaNode *schema, int type)
{
  schema->type = type;
  schema->fixed_type = 1;
}

void SetVariableType(SchemaNode *schema)
{
  schema->fixed_type = 0;
}

void SetNodePermissions(NodeType *node, char type, ClientSelector *selector)
{
  SetPermissions(SchemaOf(node), type, selector);
}

struct checkPermData
{
  ClientIDType *clientID;
  char perm;
};

static int checkPermPre(struct checkPermData *data, NodeType *node,
			NodeType *parent, char *path)
{
  if (!CheckNodePermission(node, data->clientID, data->perm))
    return 0;
  return 1;
}

int CheckDescendantPermission(NodeType *node, ClientIDType *clientID,
			      char type)
{
  struct checkPermData data;
  data.perm = type;
  data.clientID = clientID;
  return TraverseTree(node, NULL, &data, (TravTreeFN) checkPermPre, NULL, 
		      NULL);
}

static int mayAddRec(ClientIDType *clientID, NodeType *node, 
		     SchemaNode *schema)
{
  //printf("Checking on %s\n", node->name);
  if (!CheckPermission(schema, clientID, 'w'))
    return 0;
  //printf("Checking on %s\n", node->name);
  if (GetFormat(node->type) == T_BRANCH)
    {
      DArray nodes = node->data.branch_val;
      int index = 0;
      NodeType *child;
      SchemaNode *schemaChild;
      while (index < ArrayLength(nodes))
	{
	  child = *((NodeType **)ArrayNth(node->data.branch_val, index));
	  schemaChild = GetSchemaChild(schema, child->name);
	  if (schemaChild)
	    {
	      if (!mayAddRec(clientID, child, schemaChild))
		return 0;
	    }
	  index++;
	}
    }
  return 1;
}


/* This tests in advance if we are permitted to do the
 * operation. Note: only needed for replace/delete, since the others
 * can simply do nothing if not permitted. Operation is therefore
 * determined by whether dataNode is NULL.  
 */
static char *mayPublish(ClientIDType *clientID, NodeType *dataNode,
			NodePathReturnType *npReturn, int hard,
			int publish_if_locked)
{
  if (dataNode)
    {
      // We need to traverse the schema and dataNode and determine:

      if (!npReturn->targetSchema)
	return sv_success;
      if (!mayAddRec(clientID, dataNode, npReturn->targetSchema))
	return sv_locked;
      // whether we will be allowed to publish everything in dataNode
      // to the UNLs they will get.

      // whether we are trying to publish the parent of a required
      // child without publishing the child.
    }
  else
    {
      // If there's already nothing there, fine.
      if (!npReturn->existNode)
	return sv_success;

      // We need to make sure we can write to every existing node in
      // the current value.
      if (!CheckDescendantPermission(npReturn->existNode, clientID, 'w'))
	return sv_locked;

      // We need to make sure we aren't removing a required child at
      // the root.
    }
  return sv_success;
}

/**** Operations on tree structure ****/

#define LINK_HARD (1 << 0)
#define LINK_IMPLICIT (1 << 1)

struct unlinkInfo
{
  ClientIDType *clientID;
  // The deepest ancestor which will remain
  NodeType *lastLeft;
};

static int UnlinkNodePost(struct unlinkInfo *info, NodeType *node, 
			  NodeType *parent, char *UNL)
{
  DArray subscribers;
  SubscriptionType *p_subscriber;
  int iLen, i;
  if (!parent)
    return 1; // We're replacing this node; don't unlink

  if (!InTree(node))
    {
      printf("Node already unlinked!\n");
      return 1;
    }

  // We need to move all of the subscriptions
  if ((subscribers = SchemaOf(node)->subscribers))
    {
      iLen = ArrayLength(subscribers);
      for (i = 0; i < iLen; i++)
	{
	  p_subscriber = 
	    *((SubscriptionType **) ArrayNth(subscribers, i));
	  //printf("%s is %s\n", node->name, UNL);
	  if (info->clientID != p_subscriber->subscriber ||
	      p_subscriber->SELF_SUBSCRIBE)
	    SendUpdateNode(p_subscriber->subscriber, NULL, UNL, 
			   info->clientID->name);
	  MoveSubscription(p_subscriber, info->lastLeft);
	}
      ArrayFree(subscribers);
      SchemaOf(node)->subscribers = NULL;
    }
  
  if ((subscribers = MissingSubs(node)))
    {
      iLen = ArrayLength(subscribers);
      for (i = 0; i < iLen; i++)
	{
	  p_subscriber =
	    *((SubscriptionType **) ArrayNth(subscribers, i));
	  MoveSubscription(p_subscriber, info->lastLeft);
	}
      ArrayFree(subscribers);
      MissingSubs(node) = NULL;
    }
  
  if (SchemaOf(node)->lock.owner)
    {
      // NOTE: unpublishing a node unlocks it
      SchemaOf(node)->lock.owner = NULL;
    }
  free(node->local);
  node->local = NULL;

  RemoveNode(parent, node);
  FreeNode(node);
  return 1;
}

static void UnlinkNode(NodeType *parent, NodeType *node, char *UNL, 
		       ClientIDType *clientID)
{
  // Anyone subscribed to this node or a descendant will be notified in
  // transit, when the subscribed node is unlinked.
  struct unlinkInfo info = { clientID, parent };
  TraverseTree(node, parent, &info, NULL, (TravTreeFN) UnlinkNodePost, UNL);
}

struct linkInfo
{
  ClientIDType *clientID;
  int flags;
};

static int LinkNodePre(struct linkInfo *info, NodeType *node, 
		       NodeType *parent, char *UNL)
{
  SchemaNode *schema;
  SchemaNode *schemaParent;
  assert(node->name);

  if (InTree(node)) // just need the children here.
    return 1;

  schema = NULL;
  if (parent)
    {
      schemaParent = SchemaOf(parent);
      schema = GetSchemaChild(schemaParent, node->name);
    }
  else
    schemaParent = NULL;

  if (!schema)
    {
      schema = SchemaDefault(schemaParent, node->name);
      schema->hardNode = info->flags & LINK_HARD;
      schema->hardSchema = schema->hardNode;
    }

  /*
  if (HasFixedTypeFlag(node->type))
    {
      schema->fixed_type = 1;
      schema->type = node->type;
    }
  */

  node->local = malloc(sizeof(PathTreeDataType));
  memset(node->local, 0, sizeof(PathTreeDataType));
  SchemaOf(node) = schema;
  ParentOf(node) = parent;
  if (parent && MissingSubs(parent))
    {
      DArray missing = MissingSubs(parent);
      int i = 0;
      while (i < ArrayLength(missing))
	{
	  SubscriptionType *sub = *((SubscriptionType **)ArrayNth(missing, i));
	  //printf("Missing subscription found for %s when publishing %s\n",
	  //sub->UNL, UNL);
	  if (UNL == sub->UNL)
	    {
	      //printf("Missing subscription found for linked node\n");
	      sub->missing = 0;
	      if (!SchemaOf(node)->subscribers)
		SchemaOf(node)->subscribers =
		  ArrayNew(sizeof(SubscriptionType *), 0, NULL);
	      sub->node = node;
	      ArrayAppend(SchemaOf(node)->subscribers, &sub);
	      if (!(info->flags & LINK_IMPLICIT) && 
		  (info->clientID != sub->subscriber ||
		   sub->SELF_SUBSCRIBE))
		SendUpdateNode(sub->subscriber, sub->node, UNL, 
			       info->clientID->name);
	      ArrayDeleteAt(missing, i);
	    }
	  else if (DescendsFrom(UNL, sub->UNL))
	    {
	      //printf("Missing subscription found over node\n");
	      if (!MissingSubs(node))
		MissingSubs(node) =
		  ArrayNew(sizeof(SubscriptionType *), 0, NULL);
	      sub->node = node;
	      ArrayAppend(MissingSubs(node), &sub);
	      ArrayDeleteAt(missing, i);
	    }
	  else
	    i++;
	}
    }
  return 1;
}

/* Can fail if parent isn't a branch */
static int LinkNode(NodeType *parent, NodeType *node, char *UNL, 
		    ClientIDType *clientID, int flags)
{
  struct linkInfo info;
  info.flags = flags;
  info.clientID = clientID;
  if (parent && !InsertNode(parent, node))
    return 0;
  TraverseTree(node, parent, (void *)&info, (TravTreeFN) LinkNodePre, 
	       NULL, UNL);
  ParentOf(node) = parent;
  return 1;
}

static void SwapNodes(NodeType *node1, NodeType *node2)
{
  DataType data = node1->data;
  char *name = node1->name;
  long type = node1->type;

  node1->data = node2->data;
  node1->name = node2->name;
  node1->type = node2->type;

  node2->data = data;
  node2->name = name;
  node2->type = type;
}

static void ReplaceNode(NodeType *old, NodeType *new, char *UNL, 
			ClientIDType *clientID, int hard)
{
  struct unlinkInfo info = { clientID, old };
  SubscriptionType *p_subscriber;
  int iLen, i;
  DArray subscribers;
  NodeType *parent = ParentOf(old);

  struct linkInfo link;
  link.clientID = clientID;
  link.flags = hard ? LINK_HARD : 0;
  // Unlink all of the descendants of old, but not old itself
  // NULL parent tells UnlinkNodePost not to do it.
  TraverseTree(old, NULL, &info, NULL, (TravTreeFN) UnlinkNodePost, UNL);
  SwapNodes(old, new);
  FreeNode(new);
  TraverseTree(old, parent, &link, (TravTreeFN) LinkNodePre, NULL, UNL);
  if ((subscribers = SchemaOf(old)->subscribers))
    {
      iLen = ArrayLength(subscribers);
      for (i = 0; i < iLen; i++)
	{
	  p_subscriber = *((SubscriptionType **) ArrayNth(subscribers, i));
	  if (clientID != p_subscriber->subscriber ||
	      p_subscriber->SELF_SUBSCRIBE)
	    {
	      SendUpdateNode(p_subscriber->subscriber, old, UNL, 
			     clientID->name);
	    }
	}
    }
  SchemaOf(old)->hardNode = hard;
  if (hard)
    SchemaOf(old)->hardSchema = 1;
}

static int MergeNode(char *path, NodeType *intoNode, NodeType *fromNode,
		     ClientIDType *permittedOwner, int DESTRUCTIVE, int hard)
{
  int mayWrite;
  /*
  printf("merge\n");
  PrintNode(fromNode, NULL);
  printf("into\n");
  PrintNode(intoNode, NULL);
  */

  // A recursive lock we don't own stops things right away.
  if (SchemaOf(intoNode)->lock.owner &&
      SchemaOf(intoNode)->lock.RECURSE &&
      SchemaOf(intoNode)->lock.owner != permittedOwner)
    return 1;
  mayWrite = CheckNodePermission(intoNode, permittedOwner, 'w');
  if (GetFormat(intoNode->type) == T_BRANCH && 
      GetFormat(fromNode->type) == T_BRANCH)
    {
      // They're both branches; we're not just replacing
      NodeType *mergeChild = NULL;
      SubscriptionType *p_subscriber;
      int iLen, i;
      unsigned index;
      DArray subscribers;
      if ((DESTRUCTIVE ||
	   (hard && !SchemaOf(intoNode)->hardNode)) && 
	  mayWrite)
	{
	  // We want to change things and we're allowed to. So we
	  // replace information local to the branch node.
	  intoNode->type = fromNode->type;
	  // It can't be implicit after this
	  IsImplicit(intoNode) = 0;
	}
      // We can't do anything with the children of a node we can't read
      if (!CheckNodePermission(intoNode, permittedOwner, 'r'))
	return 1;
      for (index = 0; (mergeChild = BranchIndex(fromNode, index)); index++)
	{
	  NodeType *existChild = GetChild(intoNode, mergeChild->name);
	  if (existChild)
	    {
	      char *newPath = AppendUNLPath(path, mergeChild->name);
	      MergeNode(newPath, existChild, mergeChild, permittedOwner, 
			DESTRUCTIVE, hard);
	    }
	  else
	    {
	      if (mayWrite)
		{
		  NodeType *copyNode = NULL;
		  char *newPath;
		  SchemaNode *childSchema =
		    GetSchemaChild(SchemaOf(intoNode), mergeChild->name);
		  if (childSchema && 
		      !CheckPermission(childSchema, permittedOwner, 'w'))
		    return 1;

		  copyNode = CopyNode(mergeChild);
		  newPath = AppendUNLPath(path, copyNode->name);
		  //printf("With UNL %s:\n", newPath);
		  //PrintNode(copyNode, NULL);
		  LinkNode(intoNode, copyNode, newPath,
			   permittedOwner, hard ? LINK_HARD : 0);
		}
	    }
	}
      //Send updates to subscribed nodes
      if (InTree(intoNode) && (subscribers = SchemaOf(intoNode)->subscribers))
	{
	  iLen = ArrayLength(subscribers);
	  for (i = 0; i < iLen; i++)
	    {
	      p_subscriber = *((SubscriptionType **) ArrayNth(subscribers, i));
	      //printf("%s is %s\n", node->name, UNL);
	      if (permittedOwner != p_subscriber->subscriber ||
		  p_subscriber->SELF_SUBSCRIBE)
		{
		  SendUpdateNode(p_subscriber->subscriber, intoNode, path, 
				 permittedOwner->name);
		}
	    }
	}
      return 1;
    }
  else
    {
      NodeType *copyNode;
      // We are going to try to replace the node; stop if the
      // operation is add-only or if we can't write here.
      if (!DESTRUCTIVE)
	return 1;
      if (!mayWrite)
	return 1;
      // Branches with locked children are protected, even against the
      // owner
      if (GetFormat(intoNode->type) == T_BRANCH && 
	  SchemaOf(intoNode)->lockTrails && 
	  ArrayLength(SchemaOf(intoNode)->lockTrails) > 0)
	return 1;
      // Can't replace a persistant node with a non-persistant one
      if (!hard && SchemaOf(intoNode)->hardNode)
	return 1;
      copyNode = CopyNode(fromNode);
      ReplaceNode(intoNode, copyNode, path, permittedOwner, hard);
      return 1;
    }
}

/**** Returning information about paths ****/

// Concatenate two subscription lists for npReturn
static void
ConcatSubscriptions(DArray *p_subscriptArray, NodeType *node, char *UNL)
{
  int i, len;
  SubscriptionType *p_subscriber = NULL;

  DArray subscriberArray = SchemaOf(node)->subscribers;

  if (!subscriberArray || !(len = ArrayLength(subscriberArray)))
    return;
  
  if (!*p_subscriptArray) 
    *p_subscriptArray = ArrayNew(sizeof(SubscriptionType *), len, NULL);
      
  for (i = 0; i < len; i++)
    {
      p_subscriber = *((SubscriptionType **) ArrayNth(subscriberArray, i));
      ArrayAppend(*p_subscriptArray, &p_subscriber);
    } 
}

static void applySchema(SchemaNode *current, NodePathReturnType *npReturn)
{
  ServerLockType *currLock = &current->lock;
  // A recursive lock has different effects from lacking write permission:
  // on any ancestor it prevents changes, not just on a few relevant nodes.
  if (currLock->owner && currLock->RECURSE &&
      currLock->owner != npReturn->clientID)
    {
      npReturn->mayCreateRemove = 0;
      npReturn->mayModify = 0;
    }

  if (currLock->owner == npReturn->clientID)
    {
      // We own a lock on this node, either a recursive lock on an
      // ancestor, or a lock on the node itself.
      if (currLock->RECURSE || current == npReturn->targetSchema)
	{
	  npReturn->controlled = 1;
	}
    }

  if (!CheckPermission(current, npReturn->clientID, 'r'))
    {
      npReturn->mayRead = 0;
      if (current != npReturn->targetSchema)
	{
	  npReturn->mayCreateRemove = 0;
	  npReturn->mayModify = 0;
	}
    }

  if (!CheckPermission(current, npReturn->clientID, 'w'))
    {
      if (current == npReturn->targetSchema)
	{
	  // Can't write to node
	  npReturn->mayCreateRemove = 0;
	  npReturn->mayModify = 0;
	}
      if (current == npReturn->closestSchema)
	{
	  // Can't write to the ancestor which has the list of children which
	  // would change to include this UNL.
	  npReturn->mayCreateRemove = 0;
	}
    }
}

static void GetSchemaRec(SchemaNode *current, char *path,
			 NodePathReturnType *npReturn, int createSchema)
{
  if (!path)
    {
      npReturn->targetSchema = current;
      if (current->lock.owner)
	npReturn->reportLock = &current->lock;
    }
  else
    {
      char *head;
      char *tail;
      SchemaNode *child;
      
      head = GetUNLHead(path);
      tail = GetUNLTail(path);

      child = GetSchemaChild(current, head);
      if (!child && createSchema)
	child = SchemaDefault(current, head);
      if (child)
	{
	  GetSchemaRec(child, tail, npReturn, createSchema);
	}
      if (!npReturn->closestSchema)
	{
	  npReturn->closestSchema = current;
	  //npReturn->closestSchemaUNL = npReturn->UNL - path;
	}
    }
  applySchema(current, npReturn);
}

static void GetNodePathInfoRec(NodeType *currNode, char *path,
			       NodePathReturnType *npReturn,
			       int derefSymlinks, int createSchema)
{
  char *currUNL;
  char *headName, *tailName;
  NodeType *pathNode;
   
  if (!path || !path[0])
    {
      // We're at the end of the path.
      if (currNode->type == T_SYMLINK && derefSymlinks)
	{
	  // The UNL is of the symlink.
	  npReturn->existNode = currNode;

	  // Do we have it locked?
	  if (SchemaOf(currNode)->lock.owner == npReturn->clientID)
	    npReturn->controlled = 1;

	  // Will find these in the caller
	  npReturn->existParent = NULL;
	  npReturn->closestUNL = NULL;
	  npReturn->closestAncestorNode = NULL;

	  // Nothing more
	  npReturn->symlinkCont = NULL;

	  npReturn->getSubscriptions = 0;
	  // Don't bother with the code at the bottom
	  return;
	}
      // Found the node that the UNL specified, so set it.
      npReturn->existNode = currNode;
      npReturn->targetSchema = SchemaOf(currNode);
    }
  else
    {
      headName = GetUNLHead(path);
      tailName = GetUNLTail(path);
  
      // UNL of the current node
      currUNL = InstanceStringN(npReturn->UNL, 
				strlen(npReturn->UNL) - 
				strlen(path));
  
      // On the way down....
      pathNode = GetChild(currNode, headName);
      
      if (!pathNode)
	{
	  // We've come to the end of the tree without coming to the
	  // end of the UNL. This is the closest ancestor, and the
	  // node doesn't exist.
	  
	  npReturn->existNode = NULL;
	  npReturn->existParent = NULL;
	  npReturn->closestUNL = currUNL;
	  npReturn->closestAncestorNode = currNode;
	  GetSchemaRec(SchemaOf(currNode), path, npReturn, createSchema);
	}
      else if (pathNode->type == T_SYMLINK && derefSymlinks)
	{
	  // The UNL is descended from a symlink
	  npReturn->existNode = pathNode;
	  
	  // Will pick this up in the caller
	  npReturn->existParent = NULL;
	  
	  npReturn->closestUNL = currUNL;
	  npReturn->closestAncestorNode = currNode;
	  npReturn->symlinkCont = tailName;
	  
	  npReturn->getSubscriptions = 0;
	}
      else 
	{
	  // Got a valid path to follow, so recurse downwards
	  GetNodePathInfoRec(pathNode, tailName, npReturn, derefSymlinks, 
			     createSchema);
	  
	  // If the existNode was set and the existParent is still NULL, then
	  // the current current node must be the existParent.
	  if (npReturn->existNode && !npReturn->existParent)
	    {
	      // The node exists, and its parent hasn't been found yet;
	      // therefore, this node is that parent.
	      npReturn->existParent = currNode;
	      
	      // Also, it must be the closest existing ancestor, since it
	      // exists and is the closest possible ancestor.
	      npReturn->closestAncestorNode = currNode;
	      npReturn->closestUNL = currUNL;

	      npReturn->closestSchema = SchemaOf(currNode);
	    }
	}
    }
  applySchema(SchemaOf(currNode), npReturn);
  // We don't list subscribers to the target node, because updates for
  // the target node, missing ancestors of the target node, and
  // descendants of the target node are handled separately.
  if (npReturn->existNode != currNode)
    {
      if (npReturn->getSubscriptions &&
	  SchemaOf(currNode)->subscribers)
	{
	  ConcatSubscriptions(&npReturn->subscriptions, currNode, currUNL);
	}
    }
}

void GetNodePathInfo(ClientIDType *clientID, char *UNL, 
		     NodePathReturnType *npReturn, int getSubscriptions, 
		     int derefSymlinks, int createSchema)
{
  char *path = ExtractUNLPath(UNL);

  memset(npReturn, 0, sizeof(NodePathReturnType));
  npReturn->clientID = clientID;
  npReturn->UNL = path;
  npReturn->getSubscriptions = getSubscriptions;

  npReturn->mayRead = 1;
  npReturn->mayCreateRemove = 1;
  npReturn->mayModify = 1;

  GetNodePathInfoRec(g_globalNode, path, npReturn, derefSymlinks, 
		     createSchema);

  if (npReturn->existNode && 
      SchemaOf(npReturn->existNode)->lock.owner)
    {
      npReturn->reportLock = &SchemaOf(npReturn->existNode)->lock;
    }
}

void GetSchemaInfo(ClientIDType *clientID, char *UNL, 
		   NodePathReturnType *npReturn)
{
}

/**** Subscriptions ****/

// Move subscriptions around, in response to change of what nodes exist
static void MoveSubscription(SubscriptionType *sub, NodeType *target)
{
  if (!MissingSubs(target))
    MissingSubs(target) = ArrayNew(sizeof(SubscriptionType *), 0, NULL);
  sub->node = target;
  sub->missing = 1;
  ArrayAppend(MissingSubs(target), &sub);
}

// Send messages to subscribers
static void InformSubscribers(ClientIDType *clientID, 
			      NodePathReturnType *npReturn)
{
  if (npReturn->subscriptions)
    {
      int i, len = ArrayLength(npReturn->subscriptions);
      SubscriptionType *p_subscription = NULL;
      for (i = 0; i < len; i++)
	{
	  p_subscription = *((SubscriptionType **) 
			     ArrayNth(npReturn->subscriptions, i));
	  plogf("Inform %s\n", p_subscription->subscriber);
	  if (p_subscription->missing)
	    continue;
	  if (p_subscription->SELF_SUBSCRIBE ||
	      p_subscription->subscriber != clientID)
	    SendUpdateNode(p_subscription->subscriber, p_subscription->node,
			   npReturn->UNL, clientID->name);
	}
      ArrayFree(npReturn->subscriptions);
      npReturn->subscriptions = NULL;
    }
}

/* Note that removing subscriptions is split, because subscriptions
 * are often removed when a node or a client is also removed, in which
 * case only one function needs to be called, since the whole other
 * array will be deleted together afterwards. 
 */

/* Remove subscription from subscriber's list */
void ForgetSubscription(ClientIDType *subscriber, SubscriptionType *sub)
{
  DArray subscribedNodes = subscriber->subscribedNodes;
  int j, jLen = ArrayLength(subscribedNodes);
  for(j = 0; j < jLen; j++)
    {
      SubscriptionType *subscription = 
	*((SubscriptionType **) ArrayNth(subscribedNodes, j));
      if(subscription == sub)
	{
	  if (jLen == 1)
	    {
	      ArrayFree(subscribedNodes);
	      subscriber->subscribedNodes = NULL;
	    }
	  else
	    ArrayDeleteAt(subscribedNodes, j);
	  return;
	}
    }
  printf("Didn't find subscription!\n");
}

/* Remove subscription from a subscribed node's list. */
void ForgetSubscriber(NodeType *node, SubscriptionType *subscription)
{
  DArray subscribers = SchemaOf(node)->subscribers;
  int j, jLen;
  if (subscribers)
    {
      jLen = ArrayLength(subscribers);
      for (j = 0; j < jLen; j++)
	{
	  SubscriptionType *sub = 
	    *((SubscriptionType **) ArrayNth(subscribers, j));
	  if (sub == subscription)
	    {
	      free(sub);
	      if (jLen == 1)
		{
		  ArrayFree(subscribers);
		  SchemaOf(node)->subscribers = NULL;
		}
	      else 
		ArrayDeleteAt(subscribers, j);
	      return;
	    }
	}
    }
  subscribers = MissingSubs(node);
  if (subscribers)
    {
      jLen = ArrayLength(subscribers);
      for (j = 0; j < jLen; j++)
	{
	  SubscriptionType *sub = 
	    *((SubscriptionType **) ArrayNth(subscribers, j));
	  if (sub == subscription)
	    {
	      free(sub);
	      if (jLen == 1)
		{
		  ArrayFree(subscribers);
		  MissingSubs(node) = NULL;
		}
	      else 
		ArrayDeleteAt(subscribers, j);
	      return;
	    }
	}
    }
  printf("Didn't find subscriber!\n");
}

/* Returns 0 to indicate that it didn't do anything, in which case
 * the operation was not needed. */
int Subscribe(ClientIDType *clientID, NodePathReturnType *npReturn, 
	      int SELF_SUBSCRIBE)
{
  int i, iLen;
  DArray subscribers, subscribedNodes;
  SubscriptionType *newSubscriber;
  NodeType *dataNode;
  if (npReturn->existNode)
    {
      dataNode = npReturn->existNode;
      subscribers = SchemaOf(dataNode)->subscribers;
      if (!subscribers)
	{
	  SchemaOf(dataNode)->subscribers = 
	    ArrayNew(sizeof(SubscriptionType *), 1, NULL);
	  subscribers = SchemaOf(dataNode)->subscribers;
	}
    }
  else
    {
      dataNode = npReturn->closestAncestorNode;
      subscribers = MissingSubs(dataNode);
      if (!subscribers)
	{
	  MissingSubs(dataNode) =
	    ArrayNew(sizeof(SubscriptionType *), 1, NULL);
	  subscribers = MissingSubs(dataNode);
	}
      //plogf("sub to missing %s\n", npReturn->UNL);
    }
  subscribedNodes = clientID->subscribedNodes;
  if (!subscribedNodes)
    {
      clientID->subscribedNodes = 
	ArrayNew(sizeof(SubscriptionType *), 1, NULL);
      subscribedNodes = clientID->subscribedNodes;
    }
  else
    {
      // Only need to check for duplicates if there were subscriptions before
      iLen = ArrayLength(subscribedNodes);
      for (i = 0; i < iLen; i++)
	{
	  SubscriptionType *sNode =
	    *(SubscriptionType **) ArrayNth(subscribedNodes, i);
	  if (sNode->UNL == npReturn->UNL)
	    {
	      sNode->SELF_SUBSCRIBE = SELF_SUBSCRIBE;
	      return 0;
	    }
	}
    }
  newSubscriber = malloc(sizeof(SubscriptionType));
  newSubscriber->subscriber = clientID;
  newSubscriber->SELF_SUBSCRIBE = SELF_SUBSCRIBE;
  newSubscriber->UNL = npReturn->UNL;
  newSubscriber->missing = 0; // ???? 
  newSubscriber->node = dataNode;
  ArrayAppend(subscribers, &newSubscriber);
  ArrayAppend(subscribedNodes, &newSubscriber);
  return 1;
}

SubscriptionType *IsSubscribed(ClientIDType *clientID, char *UNL)
{
  DArray subscribedNodes = clientID->subscribedNodes;
  if (subscribedNodes)
    {
      int i, iLen = ArrayLength(subscribedNodes);
      for (i = 0; i < iLen; i++)
	{
	  SubscriptionType *sNode = 
	    *(SubscriptionType **) ArrayNth(subscribedNodes, i);
	  //printf("Looking for sub: %s ?= %s\n", sNode->UNL, UNL);
	  if (sNode->UNL == UNL)
	    return sNode;
	}
    }
  return NULL;
}

/**** Locks ****/

static int ContainsUnownedLockTrails(NodeType *node, ClientIDType *owner)
{
  int len;
  DArray lockTrails;

  lockTrails = SchemaOf(node)->lockTrails;

  if (!lockTrails || (len = ArrayLength(lockTrails)) == 0)
    {
      return 0;
    }
  else if (len > 1)
    {
      // Can't both be the owner's
      return 1;
    }
  else
    {
      LockTrailType *p_lockTrail = (LockTrailType *) ArrayNth(lockTrails, 0);
      
      return (p_lockTrail->owner != owner);
    }	     
}

static void RecurseRemoveLockTrail(NodeType *node, char *path, 
				   LockTrailType *p_endLockTrail)
{
  DArray lockTrails = NULL;
  int i, len;
  char *head, *tail;
  NodeType *recurseNode = NULL;
  
  lockTrails = SchemaOf(node)->lockTrails;

  len = ArrayLength(lockTrails);

  for (i = 0; i < len; i++)
    {
      LockTrailType *p_trail = (LockTrailType *) ArrayNth(lockTrails, i);
      
      if (p_trail->owner == p_endLockTrail->owner)
	{
	  assert(p_trail->count >= p_endLockTrail->count);
	  p_trail->count -= p_endLockTrail->count;
	  if (p_trail->count == 0)
	    ArrayDeleteAt(lockTrails, i);
	  
	  if (p_trail == p_endLockTrail) 
	    return;
	  break;
	}
    }
     
  head = GetUNLHead(path);
  tail = GetUNLTail(path);
      
  // Get the next node on the path
  recurseNode = GetChild(node, head);
      
  RecurseRemoveLockTrail(recurseNode, tail, p_endLockTrail);
}

// Whereas RecurseUnlockNodePath will unlock a node at a designated
// point on a path, EliminateLockTrail will totally nuke any trail
// that is formed by the locked nodes owned by removeOwner below the
// passed in node at the give path.  This allows the LockTrail
// invariant to be maintained without requiring the entire subtree of
// node to be recursed over for lock removal.

static void EliminateLockTrail(NodeType *endNode, char *path, 
			       ClientIDType *removeOwner)
{
  int len;
  LockTrailType *p_lockTrail = NULL;
  DArray lockTrails = NULL;

  if (!GetFormat(endNode->type) == T_BRANCH)
    return; //leaf nodes cannot have lock trails
  
  lockTrails = SchemaOf(endNode)->lockTrails;
  if (!lockTrails) return; // never had any
   
  len = ArrayLength(lockTrails);
  if (len == 0) return; // none currently
  assert(len == 1); // Must not be other owners

  p_lockTrail = (LockTrailType *) ArrayNth(lockTrails, 0);

  RecurseRemoveLockTrail(g_globalNode, path, p_lockTrail);
}

static int RemoveLocksPre(ClientIDType *clientID, NodeType *node, 
			  NodeType *parent, char *UNL)
{
  int i, len, ret;
  DArray lockTrails;

  if (SchemaOf(node)->lock.owner == clientID)
    {
      if (SchemaOf(node)->lock.REMOVE)
	{
	  NodePathReturnType npReturn;
	  GetNodePathInfo(clientID, UNL, &npReturn, 1, 0, 0);
	  // We need to remove the node, not just the lock
	  if (mayPublish(clientID, NULL, &npReturn, 0, 1) ==
	      sv_success)
	    {
	      UnlinkNode(parent, node, UNL, clientID);
	      InformSubscribers(clientID, &npReturn);
	      return 2; // Don't need to recurse if you unlinked it...
	    }
	}
      SchemaOf(node)->lock.owner = NULL;
    }

  lockTrails = SchemaOf(node)->lockTrails;
      
  if (!lockTrails) 
    return 2;

  len = ArrayLength(lockTrails);

  ret = 2;
  for (i = len - 1; i >= 0; i--)
    {
      LockTrailType *p_lockTrail = (LockTrailType *) ArrayNth(lockTrails, i);
	  
      if (p_lockTrail->owner == clientID)
	{
	  ArrayDeleteAt(lockTrails, i);
	  ret = 1;
	  break;
	}
    }

  return ret;
}

static int RemoveLocksPost(ClientIDType *clientID, NodeType *node, 
			   NodeType *parent, char *UNL)
{
  if (IsImplicit(node) && !ArrayLength(node->data.branch_val))
    UnlinkNode(parent, node, UNL, clientID);
  return 1;
}

static int RemoveLocksByClient(ClientIDType *clientID)
{
  int i;
  TraverseTree(g_globalNode, NULL, clientID, 
	       (TravTreeFN) RemoveLocksPre, 
	       (TravTreeFN) RemoveLocksPost, "");
  if (clientID->locks)
    {
      for (i = 0; i < ArrayLength(clientID->locks); i++)
	{
	  ServerLockType *lock =
	    *((ServerLockType **)ArrayNth(clientID->locks, i));
	  lock->owner = NULL;
	}
      ArrayFree(clientID->locks);
    }
  return 1;
}

static void DecrementLockTrail(SchemaNode *node, ClientIDType *owner)
{
  int i, len;
  DArray lockTrails = node->lockTrails;
  len = ArrayLength(lockTrails);

  for (i = 0; i < len; i++)
    {
      LockTrailType *p_lockTrail = (LockTrailType *) ArrayNth(lockTrails, i);
      if (p_lockTrail->owner == owner) 
	{
	  if (p_lockTrail->count > 1) 
	    p_lockTrail->count--;
	  else
	    ArrayDeleteAt(lockTrails, i);
	  return;
	}
    }
  printf("Couldn't find the lock trail for the removed lock!\n");
}

static void IncrementLockTrail(SchemaNode *node, ClientIDType *owner)
{
  int i, len;
  LockTrailType newLockTrail;
  DArray *p_lockTrails = &node->lockTrails;

  // Lazy allocate child Lock Owners array for the parent node
  if (!(*p_lockTrails))
    {
      *p_lockTrails = ArrayNew(sizeof(LockTrailType), 1, NULL);
      len = 0;
    }
  else
    len = ArrayLength(*p_lockTrails);

  for (i = 0; i < len; i++)
    {
      LockTrailType *p_lockTrail = 
	(LockTrailType *) ArrayNth(*p_lockTrails, i);
      if (p_lockTrail->owner == owner) 
	{
	  p_lockTrail->count++;
	  return;
	}
    }

  newLockTrail.owner = owner;
  newLockTrail.count = 1;
  
    // Put the owner into this path node
  ArrayAppend(*p_lockTrails, &newLockTrail); 
}

static int RecurseLockNodePath(SchemaNode *parent, char *path, 
			       ServerLockType *p_newLock)
{
  ServerLockType *p_existLock = &parent->lock;
  char *head, *tail;
  SchemaNode *child;
  
  if (!path)
    {
      // We've gotten to the node we want to lock

      // Already locked (by us or someone else)
      if (p_existLock->owner)
	return 0;

      // We're not allowed to lock this node
      if (!CheckPermission(parent, p_newLock->owner, 'l'))
	return 0;

      *p_existLock = *p_newLock;
      ArrayAppend(p_existLock->owner->locks, &p_existLock);
      return 1;
    }

  // Recursive check: The path is not null, so the node we want to
  // lock is farther down the path.  However, before we can
  // descend, we need to make sure that the current parent node
  // does not have any recursive locks on it.
  
  if (p_existLock->owner && p_existLock->RECURSE)
    {
      // There's a recursive lock here
      return 0;
    }

  // Not allowed to read the children
  if (!CheckPermission(parent, p_newLock->owner, 'r'))
    return 0;
      
  head = GetUNLHead(path);
  tail = GetUNLTail(path);
      
  // Get the next node on the path
  child = GetSchemaChild(parent, head);
      
  if (!child)
    {
      // There's no such node...
      return 0;
    }
      
  // Recursive case: We need to recurse down the path and get the
  // result so that the proper changes can be made to this node
  // upon success.
  if (!RecurseLockNodePath(child, tail, p_newLock))
    {
      // Problems further down
      return 0;
    }
      
  // SUCCESSFUL lock: we need to leave the lock trail in this parent.
  IncrementLockTrail(parent, p_newLock->owner);
  return 1;
}

static int RecurseUnlockNodePath(SchemaNode *schema, char *path,
				 ClientIDType *owner)
{
  char *head, *tail;
  int i;
  
  if (path) // This indicates that the desired node is below this one
    { 
      SchemaNode *child;
      
      head = GetUNLHead(path);
      tail = GetUNLTail(path);
      
      // Get the next node on path
      child = GetSchemaChild(schema, head);

      if (!child)
	{
	  // No such node
	  printf("No such node %s\n", head);
	  return 0;
	}

      // Path is good so far, so keep on descending
      if (!RecurseUnlockNodePath(child, tail, owner))
	{
	  // Problems below
	  return 0;
	}

      // Since the node was successfully unlocked below, we
      // should be able to find and decrement a LockTrailType
      // entry from this node above.
      
      DecrementLockTrail(schema, owner);
      return 1;
    }

  // Test this end node to see if passed in owner owns it
  if (schema->lock.owner != owner)
    {
      printf("Didn't own it\n");
      return 0;
    }

  // The owner does indeed own this node, so unlock it and indicate
  // this in the result. 
  for (i = 0; i < ArrayLength(schema->lock.owner->locks); i++)
    {
      if (*((ServerLockType **)ArrayNth(schema->lock.owner->locks, i)) ==
	  &schema->lock)
	{
	  ArrayDeleteAt(schema->lock.owner->locks, i);
	  break;
	}
    }
  schema->lock.owner = NULL;
  return 1;
}

int LockNodePath(char *path, ServerLockType *lock)
{
  if (!lock->owner->locks)
    lock->owner->locks = ArrayNew(sizeof(ServerLockType *), 0, NULL);
  return RecurseLockNodePath(SchemaOf(g_globalNode), path, lock);
}

int UnlockNodePath(char *path, ClientIDType *owner)
{
  return RecurseUnlockNodePath(SchemaOf(g_globalNode), path, owner);
}

/**** Utilities for generating automatically generated nodes ****/

// Returns a name not used in parent
static void DistinctName(char **name_p, NodeType *parent)
{
  static int id = 0; 
  char buffer[20];
  // XXXX this needs to take into account hard published children on a restart
  sprintf(buffer, "%d", id);
  id++;
  *name_p = InstanceString(buffer);
}

static void CreateAncestors(NodePathReturnType *npReturn, 
			    ClientIDType *clientID)
{
  char *path = DescendsFrom(npReturn->closestUNL, npReturn->UNL);
  char *headName;
  char *tailName;
  char *UNL;
  NodeType *branch, *parent;

  parent = npReturn->closestAncestorNode;

  headName = GetUNLHead(path);
  tailName = GetUNLTail(path);

  while (tailName)
    {
      branch = CreateBranchNode(headName, T_BRANCH);
      // Never hard
      UNL = AppendUNLPath(npReturn->closestUNL, headName);
      LinkNode(npReturn->closestAncestorNode, branch, UNL, clientID, 
	       LINK_IMPLICIT);
      ConcatSubscriptions(&npReturn->subscriptions, branch, UNL);
      IsImplicit(branch) = 1;
      npReturn->closestAncestorNode = branch;
      npReturn->closestUNL = UNL;
      path = tailName;
      headName = GetUNLHead(path);
      tailName = GetUNLTail(path);
    }
}

// Given a path and a node, returns a created branch which puts the
// node at the path.
static NodeType *EmbedNode(char *path, NodeType *node)
{
  char *headName = NULL, *tailName = NULL;
  NodeType *branch;

  //printf("Working on %s\n", path);
  headName = GetUNLHead(path);
  tailName = GetUNLTail(path);
  //printf("Head %s tail %s\n", headName, tailName);

  if (!tailName)
    {
      if (headName)
	node->name = headName;
      return node;
    }
  branch = CreateBranchNode(headName, T_BRANCH);
  InsertNode(branch, EmbedNode(tailName, node));
  //PrintNode(branch, NULL);
  return branch;
}

static char *doPublish(ClientIDType *clientID, NodeType *dataNode,
		       NodePathReturnType *npReturn, char *op,
		       int hard, int publish_if_locked)
{
  char *ret;
  if (!dataNode != (op == kv_delete))
    {
      if (dataNode)
	printf("Delete with node\n");
      else
	printf("Non-delete without node\n");
      // Iff you don't have a node, you must be deleting.
      return sv_failure;
    }

  if (!publish_if_locked && npReturn->controlled)
    {
      printf("Publish from untrusted symlink failed on locked node.\n");
      return sv_locked;
    }

  if (npReturn->existNode)
    {
      if (!npReturn->mayModify)
	{
	  printf("Was locked\n");
	  return sv_locked;
	}
  
      if (npReturn->targetSchema->fixed_type && dataNode &&
	  !(dataNode->type == npReturn->targetSchema->type))
	{
	  // Tried to change type of node with fixed type
	  /* This is disabled, because it doesn't work for subscriptions to
	     ancestors of the node.
	  if (IsType(npReturn->existNode->type, T_REQUESTED))
	    return sv_success; // only sends updates
	  */
	  printf("Fixed type\n");
	  return sv_failure;
	}
      if (npReturn->existNode->type == T_DISTINCT && op != kv_delete)
	{
	  DistinctName(&dataNode->name, npReturn->existNode);
	  npReturn->UNL = AppendUNLPath(npReturn->UNL, dataNode->name);
	  ConcatSubscriptions(&npReturn->subscriptions, 
			      npReturn->existNode, npReturn->UNL);
	  LinkNode(npReturn->existNode, dataNode, npReturn->UNL, 
		   clientID, hard ? LINK_HARD : 0);
	  if (hard)
	    {
	      HardPublish(dataNode, npReturn->UNL);
	    }
	  return sv_success;
	}
      if (!hard && SchemaOf(npReturn->existNode)->hardNode)
	{
	  printf("Hard published\n");
	  return sv_locked;
	}
      if (op == kv_merge || op == kv_addonly)
	{
	  if (MergeNode(npReturn->UNL, npReturn->existNode, dataNode, 
			clientID, op == kv_merge, hard))
	    {
	      if (hard)
		{
		  HardPublish(dataNode, npReturn->UNL);
		}
	      FreeNode(dataNode);
	      return sv_success;
	    }
	  return sv_failure;
	}
      // Now we're down to replacement, which needs to check that
      // children are modifiable.
      ret = mayPublish(clientID, dataNode, npReturn, hard, publish_if_locked);
      if (ret != sv_success)
	return ret;

      if (ContainsUnownedLockTrails(npReturn->existNode, clientID))
	{
	  printf("Contains locked children\n");
	  // There are locked children, no replace or delete
	  return sv_locked;
	}
      EliminateLockTrail(npReturn->existNode, npReturn->UNL, clientID);
      if (op == kv_replace)
	{
	  if (hard)
	    {
	      HardPublish(dataNode, npReturn->UNL);
	    }
	  ReplaceNode(npReturn->existNode, dataNode, npReturn->UNL, 
		      clientID, hard);

	  return sv_success;
	}
      // KV_DELETE
      if (!npReturn->mayCreateRemove)
	{
	  // Parent is locked, couldn't change child list
	  printf("parent locked\n");
	  return sv_locked;
	}
      // if the node isn't hard published, don't hard unpublish it
      // ???? hard unpublication of soft node with hard descendants?
      if (!SchemaOf(npReturn->existNode)->hardNode)
	hard = 0;
      // Go up through implicit parents
      while (IsImplicit(npReturn->existParent) &&
	     ArrayLength(npReturn->existParent->data.branch_val) == 1)
	{
	  npReturn->UNL = ExtractParentUNL(npReturn->UNL);
	  npReturn->existNode = npReturn->existParent;
	  npReturn->existParent = ParentOf(npReturn->existParent);
	}
      UnlinkNode(npReturn->existParent, npReturn->existNode, 
		 npReturn->UNL, clientID);

      // Redo unlink...
      if (hard)
	HardPublish(NULL, npReturn->UNL);
      return sv_success;
    }
  else
    {
      // Node doesn't exist.

      ret = mayPublish(clientID, dataNode, npReturn, hard, publish_if_locked);
      if (ret != sv_success)
	return ret;

      if (op == kv_delete)
	{
	  // It's not there, as you requested...
	  return sv_success;
	}

      // Trying to publish directly inside a T_DISTINCT node, or such
      // that an implicitly-created ancestor would be added to a
      // T_DISTINCT node.
      if (npReturn->closestAncestorNode &&
	  npReturn->closestAncestorNode->type == T_DISTINCT)
	{
	  return sv_failure;
	}
      // Otherwise, operation doesn't matter, since it's new

      // The closest existing ancestor must be modified; it must not
      // be locked, and its ancestors must not be locked recursively.
      if (!npReturn->mayCreateRemove)
	{
	  // Path was locked
	  return sv_locked;
	}

      CreateAncestors(npReturn, clientID);
      dataNode->name = 
	GetUNLHead(DescendsFrom(npReturn->closestUNL, npReturn->UNL));
      if (!LinkNode(npReturn->closestAncestorNode, dataNode, npReturn->UNL, 
		    clientID, hard ? LINK_HARD : 0))
	{
	  // npReturn->closestAncestorNode isn't a branch.
	  printf("Error\n");
	  return sv_error;
	}

      // XXXX mark created nodes as implicit

      if (hard)
	{
	  HardPublish(dataNode, npReturn->UNL);
	}
      return sv_success;
    }
}

char *AddNodeToTree(ClientIDType *clientID, NodeType *dataNode,
		    NodePathReturnType *npReturn, char *op,
		    int hard, int publish_if_locked)
{
  char *ret;
  ret = doPublish(clientID, dataNode, npReturn, op,
		  hard, publish_if_locked);
  if (ret == sv_success)
    {
      InformSubscribers(clientID, npReturn);
    }
  return ret;
}

void PublishLocal(char *path, NodeType *insertNode, int hard)
{
  /* This uses MERGE mode */
  /* Does not dereference symlinks */
  NodePathReturnType npReturn;
  if (!path && !g_globalNode)
    {
      g_globalNode = insertNode;
      LinkNode(NULL, g_globalNode, NULL, serverID, 0);
      return;
    }
  GetNodePathInfo(serverID, path, &npReturn, 1, 0, 0);
  if (AddNodeToTree(serverID, insertNode, &npReturn, 
		    insertNode ? kv_merge : kv_delete, hard, 1) !=
      sv_success)
    {
      printf("Local %s failed!\n", insertNode ? "Publish" : "Unpublish");
      FreeNode(insertNode);
      return;
    }
}

int LocalSetPathPermissions(char *path, char type, ClientSelector *selector)
{
  NodePathReturnType npReturn;
  GetNodePathInfo(serverID, path, &npReturn, 1, 0, 1);
  if (!npReturn.targetSchema)
    return 0;
  SetPermissions(npReturn.targetSchema, type, selector);
  return 1;
}

void RemoveClientTreeEffects(ClientIDType *clientID)
{
  SubscriptionType *subscription;
  int i, iLen;
  if (clientID->subscribedNodes)
    {
      iLen = ArrayLength(clientID->subscribedNodes);
      for(i = 0; i < iLen; i++)
	{
	  subscription = 
	    *((SubscriptionType **) ArrayNth(clientID->subscribedNodes, i));
	  ForgetSubscriber(subscription->node, subscription);
	}
      ArrayFree(clientID->subscribedNodes);
    }
  RemoveLocksByClient(clientID);
}
