/*************************************************
 * File: test-utils.c
 * Birth Date: Sep 02, 2004
 *
 * Project: Enchantment IPC
 * 
 * Based on Enchantment IPC Whitepaper
 *   by Edward Keyes.
 *
 * Copyright (C) 2004 Aware Technologies, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *************************************************/

#include "test-utils.h"

#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>

void vfail(char *format, va_list args)
{
  vfprintf(stderr, format, args);
  done();
  exit(1);
}

void fail(char *format, ...)
{
  va_list args;
  va_start(args, format);
  vfail(format, args);
  va_end(args);
}

static void waitturn()
{
  char buffer[256];
  int sz = read(subclient, buffer, 256);
  if (sz == 0)
    exit(0);
  if (sz < 0)
    {
      fprintf(stderr, "Subclient error reading from main client: %s\n",
	      strerror(errno));
      exit(-1);
    }
  if (strcmp(buffer, "turn"))
    {
      fprintf(stderr, "Subclient is confused: got %d chars, '%s'\n", sz,
	      buffer);
      exit(-1);
    }
}

void check(int cond, char *format, ...)
{
  va_list args;
  va_start(args, format);
  vcheck(cond, format, args);
  va_end(args);
}

void vcheck(int cond, char *format, va_list args)
{
  ssize_t size;
  if (issubclient)
    {
      size = write(subclient, cond ? "P" : "F", 1);
      size = write(subclient, format, strlen(format) + 1);
      waitturn();
    }
  else
    {
      if (cond)
	{
	  fprintf(stderr, "[pass] ");
	}
      else
	{
	  failures++;
	  fprintf(stderr, "[FAIL] ");
	}
      vfprintf(stderr, format, args);
    }
}

void require(int cond, char *format, ...)
{
  va_list args;
  va_start(args, format);
  vcheck(cond, format, args);
  if (issubclient && !cond)
    {
      done();
      exit(1);
    }
  va_end(args);
}

void vrequire(int cond, char *format, va_list args)
{
  vcheck(cond, format, args);
  if (issubclient && !cond)
    {
      done();
      exit(1);
    }
}

void info(char *format, ...)
{
  char buffer[256];
  int ret;
  va_list args;
  va_start(args, format);
  if (!issubclient)
    {
      fprintf(stderr, "       ");
      vfprintf(stderr, format, args);
      write(subclient, "turn", 5);
      ret = read(subclient, buffer, 256);
      if (!ret)
	{
	  fail("Subclient has gone away.\n");
	}
      else if (ret == -1)
	{
	  fail("Error reading from subclient: %s\n", strerror(errno));
	}
      if (strcmp(buffer, format))
	{
	  fail("Got out of sync with subclient: '%s', not '%s'.\n",
	       buffer, format);
	}
    }
  else
    {
      write(subclient, format, strlen(format) + 1);
      waitturn();
    }
  va_end(args);
}

void turn(char *format, ...)
{
  char buffer[256];
  int cond;
  int stop;
  ssize_t size;
  size_t total = 0;
  va_list args;
  va_start(args, format);
  write(subclient, "turn", 5);
  do
    {
      size = read(subclient, buffer + total, strlen(format) + 2 - total);
      if (size < 0)
	{
	  fail("Error reading from subclient: %s\n", strerror(errno));
	  return;
	}
      total += size;
    }
  while (total < strlen(format) + 2);

  if (strcmp(buffer + 1, format))
    {
      fail("Subclient didn't sync:\n sub: %smain: %s", buffer + 1, format);
    }
  switch (buffer[0])
    {
    case 'P':
      cond = 1;
      stop = 0;
      break;
    case 'F':
      cond = 0;
      stop = 0;
      break;
    case 'p':
      cond = 1;
      stop = 1;
      break;
    case 'f':
      cond = 0;
      stop = 1;
      break;
    default:
      cond = 0;
      stop = 1;
    }
  if (cond)
    {
      fprintf(stderr, "<pass> ");
    }
  else
    {
      fprintf(stderr, "<FAIL> ");
      failures++;
    }
  vfprintf(stderr, format, args);
  va_end(args);
  if (stop && !cond)
    {
      done();
      exit(1);
    }
}

void beSubclient(char *cases)
{
  int i = 0;
  while (tests[i].name)
    {
      if (!cases || strstr(cases, tests[i].name))
	(*tests[i].sub)();
      i++;
    }
}

void splitSubclient(char *cases)
{
  int sv[2];
  int val;
  if (socketpair(AF_LOCAL, SOCK_STREAM, 0, sv))
    {
      printf("couldn't set up a socket pair: %s\n", strerror(errno));
      exit(-1);
    }
  val = fork();
  if (val == -1)
    {
      printf("couldn't fork: %s\n", strerror(errno));
      exit(-1);
    }
  if (!val)
    {
      close(sv[1]);
      subclient = sv[0];
      issubclient = 1;
      waitturn();
      beSubclient(cases);
      done();
      exit(0);
    }
  close(sv[0]);
  subclient = sv[1];
}
